#include "services.h"
/*************************************************************************
 * This file contains functions for enhancing or limiting ones access to *
 * the network. Though it contained specific AKILL only functions before *
 * they have been rewritten to now support: AKILL, Services Ignore,      *
 * AKILL Exemption and timed jinxes. These functions are all covered by  *
 * the generic term of 'access modifiers'.       -- Remmy 06/03/2002     *
 *************************************************************************/ 
static struct access_mod *first_access_mod = NULL;

void auto_remove_access_mod(char *mask_and_type);
int remove_covered_exempts(char *nick, char *user, char *host);

void list_access_mod(char *from, char type, char *masker)
{
    struct access_mod *ac;
    char mask[NICKLEN + USERLEN + HOSTLEN + 2];
    char length[16];
    struct tm *t;
    char date[20];
    int found = 0;

    if (first_access_mod) {
	for (ac = first_access_mod; ac; ac = ac->next) {
	    if (ac->type == type) {
		if (!found)
		    sSend(":%s NOTICE %s :%-10s %-14s %-8s %-20s %-10s %s",
			  OperServ, from, "ID", "SetTime", "Length", "Mask",
			  "Set By", "Reason");
		found++;
		strcpy(mask, ac->nick);
		strcat(mask, "!");
		strcat(mask, ac->user);
		strcat(mask, "@");
		strcat(mask, ac->host);
		t = localtime(&(ac->set));
		strftime(date, 20, "%m/%d/%y-%I:%M%p", t);

		if (ac->unset)
		    sprintf(length, "%luh",
			    (long) (ac->unset - ac->set) / 3600);
		else
		    strcpy(length, "forever");

		if (masker) {
		    if (!match(masker, mask) || atoi(&masker[1]) == ac->id)
			sSend(":%s NOTICE %s :#%-9d %s %-8s %-20s %-10s %s",
			      OperServ, from, ac->id, date, length, mask,
			      ac->setby, ac->reason);
		} else
                    sSend(":%s NOTICE %s :#%-9d %s %-8s %-20s %-10s %s", 
                          OperServ, from, ac->id, date, length, mask, 
                          ac->setby, ac->reason);
	    }
	}
    }

    if (found)
	sSend(":%s NOTICE %s :End of list. (%d total)", OperServ, from, found);
}

struct access_mod *isAKilled(char *nick, char *user, char *host)
{
    struct access_mod *ac;

    for (ac = first_access_mod; ac; ac = ac->next) {
	if (!match(nick, ac->nick) && !match(user, ac->user)
	    && !match(host, ac->host) && (ac->type == A_AKILL)) {
	    if (!isExempted(nick, user, host))
		return ac;
	    else
		return NULL;
	}
    }
    return NULL;
}

struct access_mod *isExempted(char *nick, char *user, char *host)
{
    struct access_mod *ac;

    for (ac = first_access_mod; ac; ac = ac->next) {
	if (!match(nick, ac->nick) && !match(user, ac->user)
	    && !match(host, ac->host) && (ac->type == A_EXEMPT))
	    return ac;
    }
    return NULL;
}

struct access_mod *isHJinxed(char *nick, char *user, char *host)
{
    struct access_mod *ac;

    for (ac = first_access_mod; ac; ac = ac->next) {
	if (!match(nick, ac->nick) && !match(user, ac->user)
	    && !match(host, ac->host) && (ac->type == A_HOSTJINX))
	    return ac;
    }
    return NULL;
}

struct access_mod *isIgnored(char *nick, char *user, char *host)
{
    struct access_mod *ac;

    for (ac = first_access_mod; ac; ac = ac->next) {
	if (!match(nick, ac->nick) && !match(user, ac->user)
	    && !match(host, ac->host) && (ac->type == A_IGNORE))
	    return ac;
    }
    return NULL;
}

void add_access_mod(long length, char *mask, char *by, char type, char *reason)
{
    UserList *from;
    struct access_mod *ac;
    char *acmask;
    char *temp1;
    char *user1 = NULL, *host1 = NULL;
    char type_and_mask[NICKLEN + USERLEN + HOSTLEN + 4];

    if (strlen(mask) < 5)
	return;

    if (!check_mask( mask, 1))
	return;

    ac = smalloc(sizeof(struct access_mod));

    acmask = strdup(mask);

    for (temp1 = acmask; *temp1; temp1++) {
	if (*temp1 == '!') {
	    *temp1 = 0;
	    user1 = temp1 + 1;
	}
	if (*temp1 == '@') {
	    *temp1 = 0;
	    host1 = temp1 + 1;
	}
    }

    if ((!user1) || (!host1)) {
	sfree(acmask);
	return;
    }

    strcpy(ac->nick, acmask);
    strcpy(ac->user, user1);
    strcpy(ac->host, host1);
    ac->id = ++ac_mod_counter;

    if (strlen(ac->user) == 1 && strlen(ac->host) == 1) {
	from = getNickData(by);
    }

    if (first_access_mod)
	first_access_mod->prev = ac;

    ac->next = first_access_mod;
    ac->prev = NULL;
    first_access_mod = ac;

    strcpy(ac->setby, by);

    ac->set = time(NULL);

    if (length == 0)
	ac->unset = 0;
    else
	ac->unset = ac->set + length;

    ac->type = type;
    strcpy(ac->reason, reason);

    switch (type) {
	case A_AKILL:
            sSend(":%s GLOBOPS :%s set akill for %s for %s [#%ld: %s]", OperServ,
                  by, mask, expand_time(length), ac->id, reason);
	    break;
	case A_EXEMPT:
	    sSend(":%s GLOBOPS :%s set akill exemption for %s for %s [#%d: %s]",
                  OperServ, by, mask, expand_time(length), ac->id, reason);
	    break;
	case A_HOSTJINX:
	    sSend(":%s GLOBOPS :%s set hostjinx for %s for %s [%s]", OperServ, by,
	          mask, expand_time(length), reason);
	    JinxJinxedHost( mask, by, reason);
	    break;
	case A_IGNORE:
            sSend(":%s GLOBOPS :%s set ignore for %s for %s [%s]", OperServ,
	          by, mask, expand_time(length), reason);
	    break;
    }

    if (type == A_AKILL && !strcmp(ac->nick, "*"))
	sSend(":%s AKILL %s %s :K-lined for %s [#%d: %s]", myname, ac->host,
	      ac->user, expand_time(length), ac_mod_counter, ac->reason);
    else if (type == A_EXEMPT && !strcmp(ac->nick, "*"))
	sSend(":%s AKILLEX %s %s :Exempted for %s [%s]", myname, ac->host,
	      ac->user, expand_time(length), ac->reason);
    else if (type == A_HOSTJINX && match(ac->nick, "*"))
	AddJinx(ac->nick, by, ac->reason);

    if (ac->unset) {
	type_and_mask[0] = type;
	strcpy(&type_and_mask[1], mask);
	ac->tid = timer(length, auto_remove_access_mod, "auto_remove_access_mod", type_and_mask);
    }
    else
	ac->tid = 0;

    sfree(acmask);
}

void remove_access_mod(char *from, char type, char *mask)
{
    char *temp1;
    struct access_mod *ac;
    char *nick = NULL, *user = NULL, *host = NULL, *oldmask = NULL;
    char tempmask[NICKLEN + USERLEN + HOSTLEN + 2];
    time_t temp;
    int removed, id = 0;

    if (mask[0] == '#') {
        id = atoi(&mask[1]);
        nick = mask;
        user = mask;
        host = mask;
    } else {
       if (strlen(mask) < 5)
           return;

       if (!check_mask( mask, 1))
           return;

       oldmask = strdup(mask);

       nick = oldmask;
    
    
       for (temp1 = oldmask; *temp1; temp1++) {
	   if (*temp1 == '!') {
	       *temp1 = 0;
	       user = temp1 + 1;
	   }
	   if (*temp1 == '@') {
	       *temp1 = 0;
	       host = temp1 + 1;
	   }
       }

       if ((!user) || (!host))
           return;
    }
    
    for (ac = first_access_mod; ac; ac = ac->next) {

	if ((!strcmp(host, ac->host) && !strcmp(user, ac->user)
	    && !strcmp(nick, ac->nick) && ac->type == type) ||
            (ac->id == id && ac->type == type)) {

	    if (id > 0) {
		strcpy(tempmask, ac->nick);
		strcat(tempmask, "!");
		strcat(tempmask, ac->user);
		strcat(tempmask, "@");
		strcat(tempmask, ac->host);

                mask = tempmask;
            }
            
	    if (ac->type == A_AKILL) {
                sSend(":%s GLOBOPS :%s is removing akill #%d %s.", OperServ,
		      from, ac->id, mask);
		logDump(operlog, "%s removed akill #%d %s", from, ac->id, mask);
	    } else if (ac->type == A_EXEMPT) {
		sSend(":%s GLOBOPS :%s is removing exemption #%d %s.", 
                      OperServ, from, ac->id, mask);
		logDump(operlog, "%s removed exemption #%d %s", from, ac->id, 
                        mask);
	    } else if (ac->type == A_HOSTJINX) {
		sSend(":%s GLOBOPS :%s is removing hostjinx #%d %s.", OperServ,
                      from, ac->id, mask);
		logDump(operlog, "%s removed hostjinx #%d for %s", from, 
                        ac->id, mask);
	    } else {
		sSend(":%s GLOBOPS :%s is removing ignore #%d %s.", 
                      OperServ, from, ac->id, mask);
		logDump(operlog, "%s removed ignore #%d %s", from, ac->id, 
                        mask);
	    }

	    temp = time(NULL);

	    if (ac->prev)
		ac->prev->next = ac->next;
	    if (ac->next)
		ac->next->prev = ac->prev;
	    if (ac == first_access_mod)
		first_access_mod = ac->next;

	    if (ac->type == A_AKILL && !strcmp(ac->nick, "*"))
		sSend(":%s RAKILL %s %s", myname, ac->host, ac->user);
	    else if (ac->type == A_EXEMPT && !strcmp(ac->nick, "*"))
		sSend(":%s RAKILLEX %s %s", myname, ac->host, ac->user);
/*
 * Fun fact:  If OperServ is removing the autokill, the timer is up, and
 * it shouldn't be canceled, or two timers will be cancelled...
 */
	    if (strcmp(from, OperServ)) {
		if (ac->type == A_AKILL) 
		    sSend(":%s NOTICE %s :AutoKill %s removed", OperServ, from, mask);
		else if (ac->type == A_EXEMPT)
		    sSend(":%s NOTICE %s :Exempt %s removed", OperServ, from, mask);
		else if (ac->type == A_HOSTJINX)
		    sSend(":%s NOTICE %s :Hostjinx %s removed", OperServ, from, mask);
		else
		    sSend(":%s NOTICE %s :Ignore %s removed", OperServ, from, mask);
		cancel_timer(ac->tid);
	    }

            if (ac->type == A_AKILL) {
                removed = remove_covered_exempts(ac->nick, ac->user, ac->host);
                if (removed > 0) {
                    sSend(":%s GLOBOPS :Auto removed %d matching exempt(s).", 
                          OperServ, removed);
                }
            }

	    /* A place to check for jinxed users from a HOSTJINX */
	    if (ac->type == A_HOSTJINX)
	        FreeJinxedHost(mask, from);

	    sfree(ac);
	    sfree(oldmask);
	    return;
	}
    }

    if (strcmp(from, OperServ))
	sSend(":%s NOTICE %s :Access modifier %s not found", OperServ,
	      from, mask);

    sfree(oldmask);
}

/* This function removes exempts when an akill is removed. Its purpose is
 * to keep the access modifier list clean from exempts that serve no longer
 * any purpose. The number of exempts removed is returned to the calling 
 * function. -- Remmy 20020619 */
int remove_covered_exempts(char *nick, char *user, char *host)
{
    struct access_mod *ac, *ac_tmp;
    int counter = 0;

    ac = first_access_mod;
    while (ac) {
        if (!match(ac->host, host) && !match(ac->user, user)
            && !match(ac->nick, nick) && ac->type == A_EXEMPT) {
            if (ac->prev)
                ac->prev->next = ac->next;
            if (ac->next)
                ac->next->prev = ac->prev;
            if (ac == first_access_mod)
                first_access_mod = ac->next;

            ac_tmp = ac;
            ac = ac->next;
            
            sfree(ac_tmp);
            counter++;
        } else {
            ac = ac->next;
        }
    }
    
    return counter;
}

void send_matching_exempts(char *nick, char *user, char *host) {

    struct access_mod *ex;
    
    for (ex = first_access_mod; ex; ex = ex->next)
        if (!match(ex->host, host) && !match(ex->user, user)
            && !match(ex->nick, nick) && ex->type == A_EXEMPT)
            sSend(":%s AKILLEX %s %s :Exempted for %s [%s]", myname,
                  ex->host, ex->user, expand_time(ex->unset), ex->reason);

    return;
}

int find_access_mod(char type, char *mask)
{
    char *temp1;
    struct access_mod *ac;
    char *nick, *user = NULL, *host = NULL, *oldmask;

    oldmask = strdup(mask);

    nick = oldmask;

    for (temp1 = oldmask; *temp1; temp1++) {
	if (*temp1 == '!') {
	    *temp1 = 0;
	    user = temp1 + 1;
	}
	if (*temp1 == '@') {
	    *temp1 = 0;
	    host = temp1 + 1;
	}
    }

    for (ac = first_access_mod; ac; ac = ac->next) {
	if (!strcmp(host, ac->host) && !strcmp(user, ac->user)
	    && !strcmp(nick, ac->nick) && (type == ac->type)) {
	    sfree(oldmask);
	    return 1;
	}
    }
    sfree(oldmask);
    return 0;
}

/* This function only gets called by the timer function when an access mode
 * is to expire. As the timer can only be set with a single argument to the
 * function it is to call I had to put both the access mode type and the
 * mask to apply it to into a single string to pass as argument. I know it
 * is not pretty, but it will have to do for now. -- Remmy 17/03/2002 */
void auto_remove_access_mod(char *mask_and_type)
{
    char *mask;
    char type;
    
    /* Extract access mode type and mask to apply it to */
    type = mask_and_type[0];
    mask = mask_and_type + 1;
    
    remove_access_mod(OperServ, type, mask);
}

void save_access_mod()
{
    FILE *acsave;
    struct access_mod *ac;

    acsave = fopen(DBDIR "/access_mod.db", "w");
    if (acsave == NULL) {
	Debug((DEBUG_ERROR, "Failed to open acces_mod db"));
	return;
    }

    for (ac = first_access_mod; ac; ac = ac->next) {
	fputs("-\n", acsave);
	fputs(ac->nick, acsave);
	fputs("\n", acsave);
	fputs(ac->user, acsave);
	fputs("\n", acsave);
	fputs(ac->host, acsave);
	fputs("\n", acsave);
	fprintf(acsave, "%lu %lu\n", (long) ac->set, (long) ac->unset);
	fputs(ac->setby, acsave);
	fputs("\n", acsave);
	fputs(ac->reason, acsave);
	fputs("\n", acsave);
	fprintf(acsave, "%d\n", ac->type);
	fputs("\n", acsave);
	fprintf(acsave, "%d\n", ac->id);
    }
    fclose(acsave);
}

void load_access_mod()
{
    FILE *acload;
    struct access_mod *ac;
    char buffer[256], mask[NICKLEN + USERLEN + HOSTLEN + 3];
    char type_and_mask[NICKLEN + USERLEN + HOSTLEN + 4];
    int i = 1, t, expired = 0, count = 0;
    long length;
    time_t start;

    first_access_mod = NULL;

    start = time(NULL);

    acload = fopen(DBDIR "/access_mod.db", "r");

    if (!acload) {
	return;
    }

    while (sfgets(buffer, 256, acload)) {
/*
 * Check for sync 
 */
	if (strcmp(buffer, "-") != 0) {
	    sSend(":%s GLOBOPS :Sync error in access_mod.db near line %d",
		  OperServ, i);
	    fprintf(corelog, "Sync error in access_mod.db near line %d\n", i);
	    fclose(acload);
	    return;
	}

	ac = smalloc(sizeof(struct access_mod));

	fscanf(acload, "%s\n", ac->nick);
	fscanf(acload, "%s\n", ac->user);
	fscanf(acload, "%s\n", ac->host);
	fscanf(acload, "%lu %lu\n", (long *) &(ac->set),
	       (long *) &(ac->unset));
	sfgets(ac->setby, NICKLEN, acload);
	sfgets(ac->reason, 80, acload);
	fscanf(acload, "%d\n", &t);
	ac->type = (char) t;
	fscanf(acload, "%d\n", &t);
	ac->id = t;
        if (ac->id > ac_mod_counter)
            ac_mod_counter = ac->id;
	ac->tid = 0;

	strcpy(mask, ac->nick);
	strcat(mask, "!");
	strcat(mask, ac->user);
	strcat(mask, "@");
	strcat(mask, ac->host);

	count++;
	length = 0;

	if (ac->unset)
	    length = ac->unset - ac->set;

	if ((start < ac->unset) || (ac->unset == 0)) {
	    if (ac->unset) {
	        type_and_mask[0] = ac->type;
	        strcpy(&type_and_mask[1], mask);
	        ac->tid = timer(length, auto_remove_access_mod, 
                                "auto_remove_access_mod", type_and_mask);
            }

	    if ((ac->type == A_AKILL) && (!strcmp(ac->nick, "*")))
		sSend(":%s AKILL %s %s :K-lined for %s [%s]", myname,
		      ac->host, ac->user, expand_time(length), ac->reason);

	    ac->next = first_access_mod;
	    ac->prev = NULL;
	    if (ac->next)
		ac->next->prev = ac;
	    first_access_mod = ac;
	} else {		/* IT'S EXPIRED, REALLY! */
	    if (ac->type == A_AKILL)
		sSend(":%s RAKILL %s %s", myname, ac->host, ac->user);
	    expired++;
	    sfree(ac);
	}
    }
    fclose(acload);
}
