#include "services.h"
#include <getopt.h>
#include <dirent.h>
#include <sys/types.h>

char buffer[4096];
int pdr, pds, brcv, bsnd;

void checkTusers(void);
void readConf(void);

struct option long_opts[] = {
	{"datadir", 1, 0, 'd'},
	{"help", 0, 0, 'h'},
	{"version", 0, 0, 'V'},
	{0, 0, 0, 0}
};

void usage(void)
{
	printf(PACKAGE " v" VERSION " \n");
	printf("Usage: services [options]\n");
	printf("\t-d, --datadir\t\tDirectory with our data\n");
	printf("\t-h, --help\t\tPrint this usage message\n");
	printf("\t-V, --version\t\tPrint the release version\n");
}


int main(int argc, char *argv[])
{
	fd_set readme;
	FILE *fp;
	struct timeval tv;
	int i;
	int o;
	time_t timestart;
	char backup[124];
	int languages;

	
#ifdef FORCE_CORE
    struct rlimit corelim;
#endif
    int pid = 0;

    antiloop = 0;
    timerbuff[0] = '\0';
    debugbuff[0] = '\0';

    startup = time(NULL);

    /* use some very neato FreeBSD malloc() stuff... */
 //   setenv("MALLOC_OPTIONS", "AR", 1);

#ifdef FORCE_CORE
    corelim.rlim_cur = corelim.rlim_max = RLIM_INFINITY;
    setrlimit(RLIMIT_CORE, &corelim);
#endif

	while ((o = getopt_long(argc, argv, "d:hV", &long_opts[0], NULL)) != EOF) {
		switch (o) {
			case 'd':
				strncpy(datadir, optarg, 512);
				printf("Data directory is `%s`\n", datadir);
				break;
			case 'V':
				printf(PACKAGE " v" VERSION "\n");
				return 0;
			case 'h':
				usage();
				return 0;
			default:
			case '?':
				printf("Try `services --help` for more information.\n");
				return 1;
		}
	}

	if (getuid() == 0 || geteuid() == 0) {
		fprintf(stderr, "Please do not run services as root!\n");
		exit(1);
	}
	
	if (strcmp(datadir, ".") == 0) {
		fprintf(stderr, "Please specify a full pathname to the data directory.\n");
		exit(1);
	}

	if (strcmp(datadir, "") == 0) {
		strcpy(datadir, getenv("HOME"));
		strcat(datadir, "/services");
	}

	if (chdir(datadir) == -1) {
		fprintf(stderr, "Invalid data directory. [%s]\n", datadir);
		exit(1);
	}	

	fp = fopen(PIDFILE, "r");
	if (fp != NULL) {
		fgets(buffer, 512, fp);
		pid = atoi(buffer);
		if (pid != 0) {
			if(kill(pid, 0) == 0)
				exit(1);
		}
	
		fclose(fp);
	}
	
// Make this smarker about missing directories
//	if (opendir(datadir))
//	
	cslock = open(DBDIR "/chanserv.db", O_RDONLY);
	i = flock(cslock, LOCK_EX | LOCK_NB);
	if (!i)
		nslock = open(DBDIR "/nickserv.db", O_RDONLY);
		i = flock(nslock, LOCK_EX | LOCK_NB);
	if (!i)
		mslock = open(DBDIR "/memoserv.db", O_RDONLY);
		i = flock(mslock, LOCK_EX | LOCK_NB);
	if (i) {
		sshutdown(-1);
	} else
		printf("Database files locked\n");

	timestart = time(NULL);
	strftime(backup, 120, COREDIR "/core.%d%b.%H%M",
	localtime(&timestart));
	if (rename(COREFILE, backup) == 0) {
		printf("Saving core file '%s' -> '%s'\n", COREFILE, backup);
	}

	firstRegNick = lastRegNick = NULL;
	firstUser = lastUser = NULL;
	firstReg = lastReg = NULL;
	firstChan = lastChan = NULL;
	firstRegChan = lastRegChan = NULL;
#ifdef SERVICESOPER
	firstOper = lastOper = NULL;
#endif
	pdr = pds = 0;
	brcv = bsnd = 0;

	printf("Loading Serenity-IRC " NETWORK " Services...\n");

    /* logging files */
    rootlog = fopen(LOGDIR "/root.log", "a");
    if (rootlog == NULL) {
	return 201;
    }
    operlog = fopen(LOGDIR "/operserv.log", "a");
    if (operlog == NULL) {
	printf("Failed to open operserv log\n");
	return 201;
    }
    nicklog = fopen(LOGDIR "/nickserv.log", "a");
    if (nicklog == NULL) {
	printf("Failed to open nickserv log\n");
	return 201;
    }
    chanlog = fopen(LOGDIR "/chanserv.log", "a");
    if (chanlog == NULL) {
	printf("Failed to open chanserv log\n");
	return 201;
    }
    corelog = fopen(LOGDIR "/core.log", "a");
    if (corelog == NULL) {
	printf("Failed to open core log\n");
	return 201;
    }
    printf("Logfiles opened\n");

    /*
     * clear out our hash lists 
     */
    cleanNickHash();
    cleanChanHash();
    printf("Hashes cleaned\n");

    AllocCalled = FreeCalled = 0;
    printf("Reading Databases/configuration...\n");
    printf("NickServ...");
    readNickData();
    printf("done.\n");
    printf("ChanServ...");
    readChanData();
    printf("done.\n");
    printf("MemoServ...");
	 readConf();
	 printf("done.\n");
    printf("Complete...\n");
    printf("My name          : %s\n", myname);
    printf("Primary host/port: %s %i\n", hostname[0], port[0]);
    printf("Op/Akick Limits  : %i/%i\n", OpLimit, AkickLimit);
    printf("Nick/Chan Limits : %i/%i\n", NickLimit, ChanLimit);
    printf("Access Item Limit: %i\n", AccessLimit);

    fp = fopen("services.totals", "r");

	if (fp) {
		sfgets(buffer, 256, fp);
		mostusers = atoi(buffer);
		sfgets(buffer, 256, fp);
		mostnicks = atoi(buffer);
		sfgets(buffer, 256, fp);
		mostchans = atoi(buffer);
		sfgets(buffer, 256, fp);
		mostmemos = atoi(buffer);
		sfgets(buffer, 256, fp);
		ac_mod_counter = atoi(buffer);
		fclose(fp);
   }
	printf("Saved Totals: Users: %lu Nicks: %lu Chans: %lu Memos: %lu\n",
		mostusers, mostnicks, mostchans, mostmemos);
	totalusers = 3;

#ifndef WIN32
	signal(SIGPIPE, segvdie);
	signal(SIGSEGV, segvdie);
	signal(SIGHUP, SIG_IGN);
#ifdef DEBUG
	signal(SIGINT, segvdie);
#else
	signal(SIGINT, SIG_IGN);
#endif
	signal(SIGBUS, segvdie);
	signal(SIGTERM, segvdie);

	printf("Added signal handlers\n");
#endif
	OperServ = services[0].name;
	NickServ = services[1].name;
	ChanServ = services[2].name;
	MemoServ = services[3].name;
	WebServ = services[4].name;
	Serenity = services[5].name;

    /* Load the language-tables here */

	languages = LoadTables();

	server = -1;
	for (i = 0; i < 10 && hostname[i][0]; i++) {
		server = ConnectToServer(hostname[i], port[i]);
		if (server >= 1)
			break;
	}

	if (server < 0) {
		printf("Could not connect to server!, exiting now.\n");
		exit(0);
	}
	
	synched = 0;
	sSend("PASS %s", mypass);
	sSend("SERVER %s 1 Star5._cap1 :%s", myname, SERVERMLINE);

	/* This could be done with SJOIN or SNICK */

	addUser(services[0].name, services[0].uname, services[0].host, services[0].rname, services[0].mode);
	addUser(services[1].name, services[1].uname, services[1].host, services[1].rname, services[1].mode);
	addUser(services[2].name, services[2].uname, services[2].host, services[2].rname, services[2].mode);
	addUser(services[3].name, services[3].uname, services[3].host, services[3].rname, services[3].mode);
	addUser(services[4].name, services[4].uname, services[4].host, services[4].rname, services[4].mode);
	addUser(services[5].name, services[5].uname, services[5].host, services[5].rname, services[5].mode);

	printf("Connected to server %s/%i\n", hostname[i], port[i]);
	flushLogs("void");
	expireChans("void");
	expireNicks("void");
	sync_cfg("4");
	load_access_mod();
	loadpermignores();
	loaddynamicekills();
#ifdef BACKUPDATABASE
	do_remaining_backup(NULL);
#endif
	printf("Initialized in %lu seconds\n\n", time(NULL) - startup);
#if !defined(DEBUG) && !defined(NOFORK)
	pid = fork();
	if (pid) {
		printf("Forking to background. PID: %i\n", pid);

		flock(pidlock, LOCK_UN);
		fp = fopen(PIDFILE, "w");
		if (fp == NULL) {
			printf("Warning: Failed to create pid file!");
			return(0);
		}
		fprintf(fp, "%i", pid);
		fclose(fp);
		pidlock = open(PIDFILE, O_RDONLY);
		return 0;
	}
#endif
	sSend(":Serenity GLOBOPS :Serenity-IRC Network Services (%s) are now online.", PACKAGE " v" VERSION);

	/* Add our jupes AFTER we sync, so that existing servers don't prevent us from connecting if they are in our JupeList */
	timer(2, sync_jupes, "sync_jupes", "void");
//	sync_jupes();
	
	while (1) {			/* loop forever...supposedly... */
		tv.tv_sec = 0;
		tv.tv_usec = 500;
		FD_ZERO(&readme);
		FD_SET(server, &readme);
		bzero(buffer, 4096);
/*		fflush(servout);*/
#ifdef TELNET
	checkTelnet();
#endif

		FromSendQ();		/* And here goes the sendQ */

		if (select(server + 1, &readme, NULL, NULL, &tv)) {
			recv(server, buffer, 4096, 0);

			ToRecvQ(buffer);	/* All data read goes right into the recvQ */
		}

		FromRecvQ();		/* And here it comes out again, looks so stupid without threads :) */

		checktimers();
	}
	return(0);
}

void flushLogs(char *placeholder)
{
    fflush(nicklog);
    fflush(chanlog);
    fflush(operlog);
    fflush(rootlog);
    timer(60, flushLogs, "flushLogs", "void");
}

void srestart()
{
    FILE *fp = fopen("services.totals", "w");

    if (fp) {
	fprintf(fp, "%lu\n%lu\n%lu\n%lu\n%lu\n", mostusers, mostnicks,
		mostchans, mostmemos, ac_mod_counter);
	fflush(fp);
	fclose(fp);
    }

    unlink(PIDFILE);
    save_access_mod();
    savepermignores();
    savedynamicekills();
    saveChanData(firstRegChan);
    saveNickData(firstRegNick);
    saveMemoData(firstRegNick);

    flock(cslock, LOCK_UN);
    flock(nslock, LOCK_UN);
    flock(mslock, LOCK_UN);
    flock(pidlock, LOCK_UN);
    sSend(":%s SQUIT %s :Services %s restarted.", myname, hostservername, VERSION);
    FromSendQ();		/* Let's empty our SendQ before splitting */
/*    fflush(servout);
    fclose(servout);*/
    close(server);
    (void) execv("services", NULL);
    exit(0);
}

void sshutdown(int type)
{
    FILE *fp = fopen("services.totals", "w");

	if (fp) {
		fprintf(fp, "%lu\n%lu\n%lu\n%lu\n%lu\n", mostusers, mostnicks, mostchans, mostmemos, ac_mod_counter);
		fflush(fp);
		fclose(fp);
	}

	if (type != (-1) && antiloop < 2) {
		antiloop = 1;
		save_access_mod();
		savepermignores();
		savedynamicekills();
		saveChanData(firstRegChan);
		saveNickData(firstRegNick);
		saveMemoData(firstRegNick);
   }
	
	unlink(PIDFILE);

	if (type == 1) {
		sSend(":%s SQUIT %s :Services %s fall down go BOOM!", myname, hostservername, VERSION);
		FromSendQ();		/* Let's empty our SendQ before splitting */
/*		fflush(servout);
		fclose(servout);*/
    close(server);
		fclose(corelog);
		kill(getpid(), 3);
		kill(getpid(), 9);
		exit(1);
    } else if (type != -1) {
		sSend(":%s SQUIT %s :Services %s shutdown.", myname, hostservername, VERSION);
		FromSendQ();	/* Let's empty our SendQ before splitting */
/*		fflush(servout);
		fclose(servout);*/
		fclose(corelog);
    close(server);
	}
	exit(0);
}

void WriteCoreLog(void)
{
	char temp[512];

	strftime(temp, 80, "%a %d/%m/%Y %T %Z", localtime(&timertime));
	logDump(corelog,
	    "<- Core Info ->");
	if (coreBuffer[0])
		logDump(corelog, "Last parsed command: %s", coreBuffer);
	else
		logDump(corelog, "Last parsed command: No commands parsed.");
	if (timerbuff[0])
		logDump(corelog, "Last used timer    : %s at %s", timerbuff, temp);
	else
		logDump(corelog, "Last used timer    : No timers called.");
}

#ifndef WIN32
void segvdie(int doot)
{
    FILE *fp;			/* File Pointer for the totals. -GZ */

    switch (doot) {
    case SIGPIPE:
    case SIGINT:
    case SIGTERM:
	sshutdown(0);
	break;
    case SIGSEGV:
    case SIGBUS:
	if (!antiloop) {
	    antiloop++;

	    fp = fopen("services.totals", "w");

	    if (fp) {
		fprintf(fp, "%lu\n%lu\n%lu\n%lu\n%lu\n", mostusers, mostnicks, mostchans, mostmemos, ac_mod_counter);
		fflush(fp);
		fclose(fp);
	    }

	    sSend(":%s GLOBOPS :\002PANICK!\002 Services has detected a fatal error and will shut itself down.", Serenity);
	    WriteCoreLog();
	    sshutdown(1);
	} else {

	    fclose(corelog);
	    close(server);
	    kill(getpid(), 3);
	    kill(getpid(), 9);
	}
	break;
    default:
	sSend(":%s GLOBOPS :Received unidentified signal %i, ignoring.",
	      myname, doot);
	break;
    }
    return;
}
#endif

void readConf(void)
{
	/*
	 * read services.conf 
	 */

	FILE *fp;
	char buffer[256];
	char *args[256];
	char realargs[256][128];
	int i, x;

	for (i = 0; i < 256; i++)
		args[i] = realargs[i];

	fp = fopen(CFGFILE, "r");
	if (fp == NULL) {
		printf("Failed to open config file '%s': %s\n", CFGFILE, strerror(errno));
		exit(-1);
	}
	
	while (sfgets(buffer, 256, fp) != NULL) {
		if (*buffer == '#' || !*buffer) {
		} else {
			bzero(args[0], 128);
			bzero(args[1], 128);
			bzero(args[2], 128);
			
			i = breakString(255, buffer, args, " \t");
			
			if (!strcasecmp(args[0], "MyName"))
				strcpy(myname, args[1]);
			else if (!strcasecmp(args[0], "Port"))
				port[atoi(args[1])] = atoi(args[2]);
			else if (!strcasecmp(args[0], "LinkPass"))
				strcpy(mypass, args[1]);
			else if (!strcasecmp(args[0], "ConnectHost")) {
				strcpy(hostname[atoi(args[1])], args[2]);
				hostname[atoi(args[1]) + 1][0] = 0;
			} else if (!strcasecmp(args[0], "OSNick"))
				strcpy(services[0].name, args[1]);
			else if (!strcasecmp(args[0], "OSUserName"))
				strcpy(services[0].uname, args[1]);
			else if (!strcasecmp(args[0], "OSHostName"))
				strcpy(services[0].host, args[1]);
			else if (!strcasecmp(args[0], "OSRealName")) {
				x = 2;
				strcpy(services[0].rname, args[1]);
				
				while (x < i) {
					strcat(services[0].rname, " ");
					strcat(services[0].rname, args[x]);
					x++;
				}
			} else if (!strcasecmp(args[0], "OSUserMode"))
				strcpy(services[0].mode, args[1]);
			else if (!strcasecmp(args[0], "NSNick"))
				strcpy(services[1].name, args[1]);
			else if (!strcasecmp(args[0], "NSUserName"))
				strcpy(services[1].uname, args[1]);
			else if (!strcasecmp(args[0], "NSHostName"))
				strcpy(services[1].host, args[1]);
			else if (!strcasecmp(args[0], "NSRealName")) {
				x = 2;
				strcpy(services[1].rname, args[1]);
				
				while (x < i) {
					strcat(services[1].rname, " ");
					strcat(services[1].rname, args[x]);
					x++;
				}
			} else if (!strcasecmp(args[0], "NSUserMode"))
				strcpy(services[1].mode, args[1]);
			else if (!strcasecmp(args[0], "CSNick"))
				strcpy(services[2].name, args[1]);
			else if (!strcasecmp(args[0], "CSUserName"))
				strcpy(services[2].uname, args[1]);
			else if (!strcasecmp(args[0], "CSHostName"))
				strcpy(services[2].host, args[1]);
			else if (!strcasecmp(args[0], "CSRealName")) {
				x = 2;
				strcpy(services[2].rname, args[1]);
				
				while (x < i) {
					strcat(services[2].rname, " ");
					strcat(services[2].rname, args[x]);
					x++;
				}
			} else if (!strcasecmp(args[0], "CSUserMode"))
				strcpy(services[2].mode, args[1]);
			else if (!strcasecmp(args[0], "MSNick"))
				strcpy(services[3].name, args[1]);
			else if (!strcasecmp(args[0], "MSUserName"))
				strcpy(services[3].uname, args[1]);
			else if (!strcasecmp(args[0], "MSHostName"))
				strcpy(services[3].host, args[1]);
			else if (!strcasecmp(args[0], "MSRealName")) {
				x = 2;
				strcpy(services[3].rname, args[1]);
				
				while (x < i) {
					strcat(services[3].rname, " ");
					strcat(services[3].rname, args[x]);
					x++;
				}
			} else if (!strcasecmp(args[0], "MSUserMode"))
				strcpy(services[3].mode, args[1]);
			else if (!strcasecmp(args[0], "WSNick"))
				strcpy(services[4].name, args[1]);
			else if (!strcasecmp(args[0], "WSUserName"))
				strcpy(services[4].uname, args[1]);
			else if (!strcasecmp(args[0], "WSHostName"))
				strcpy(services[4].host, args[1]);
			else if (!strcasecmp(args[0], "WSRealName")) {
				x = 2;
				strcpy(services[4].rname, args[1]);
		
				while (x < i) {
					strcat(services[4].rname, " ");
					strcat(services[4].rname, args[x]);
					x++;
				}
			} else if (!strcasecmp(args[0], "WSUserMode"))
				strcpy(services[4].mode, args[1]);
			else if (!strcasecmp(args[0], "GMNick"))
				strcpy(services[5].name, args[1]);
			else if (!strcasecmp(args[0], "GMUserName"))
				strcpy(services[5].uname, args[1]);
			else if (!strcasecmp(args[0], "GMHostName"))
				strcpy(services[5].host, args[1]);
			else if (!strcasecmp(args[0], "GMRealName")) {
				x = 2;
				strcpy(services[5].rname, args[1]);
		
				while (x < i) {
					strcat(services[5].rname, " ");
					strcat(services[5].rname, args[x]);
					x++;
				}
			} else if (!strcasecmp(args[0], "GMUserMode"))
				strcpy(services[5].mode, args[1]);
			else if (!strcasecmp(args[0], "OpLimit"))
				OpLimit = atoi(args[1]);
			else if (!strcasecmp(args[0], "BanLimit"))
				AkickLimit = atoi(args[1]);
			else if (!strcasecmp(args[0], "AccessItemLimit"))
				AccessLimit = atoi(args[1]);
			else if (!strcasecmp(args[0], "ChanLimit"))
				ChanLimit = atoi(args[1]);
			else if (!strcasecmp(args[0], "NickLimit"))
				NickLimit = atoi(args[1]);
			else if (!strcasecmp(args[0], "RootAdmin")) {
				RegNickList *root;
				root = getRegNickData(args[1]);
				
				if (root) {
					root->flags |= NHOLD;
					root->flags |= NROOT;
				}
			} else if (!strcasecmp(args[0], "NetAdmin")) {
				RegNickList *root;
				root = getRegNickData(args[1]);
				
				if (root) {
					root->flags |= NHOLD;
					root->flags |= NROOT;
					root->flags |= NNETADM;
				}
			} else if (!strcasecmp(args[0], "NetFounder")) {
				RegNickList *root;
				root = getRegNickData(args[1]);
				
				if (root) {
					root->flags |= NHOLD;
					root->flags |= NROOT;
				}
			} else if (!strcasecmp(args[0], "ForceMask"))
				addForce(args[1]);
			else if (!strcasecmp(args[0], "BanChannel"))
				addBanChannel(args[1]);
                        else if (!strcasecmp(args[0], "BadNick"))
                                addBadNick(args[1]);
			}
	}
	fclose(fp);
}

#ifdef DEBUG
/* Don't call this function directly, use the macro Debug
 * as follows:
 * Debug((LOGLEVEL,"Mesg %s",param))
 * Don't forget the (( ))
 */
#ifdef DEBUGFILE
static int debugfd = -1;
#endif

void _debug(int level, char *msg, ...)
{
    va_list vl;
    static char debugbuf[2049];
    int len;

    if (level > DEBUG)
	return;

    va_start(vl, msg);
    vsprintf(debugbuf, msg, vl);
    va_end(vl);

    len = strlen(debugbuf);
    if (len == 0)
	return;
    if (debugbuf[len - 1] != '\n') {
	debugbuf[len++] = '\n';
	debugbuf[len] = 0;
    }

#ifdef DEBUGFILE
    if (debugfd == -1) {
	/* Open the debugging file */
	debugfd = open(DEBUGFILE, O_CREAT | O_WRONLY);
	if (debugfd == -1) {
	    fprintf(stderr,
		    "Panic! Can't open debugfile \"%s\"! Aborting!\n",
		    DEBUGFILE);
	    exit(200);
	}
    }
    write(debugfd, debugbuf, len);
#endif
    fputs(debugbuf, stderr);

    return;
}

#endif				/* DEBUG */
