#include "services.h"

/* NICKSERV.C -- Functions for handling both online nick control and
 *               maintenance of registered nick databases
 */

/*
 * ===DOC===
 * firstNick is the first nickname, and is static, lastNick is the last
 * allocated nickname slot. thus for allocation we use lastNick->next
 */

int countRegAddy(char *);

/* simple linked list with all the ghost stuff in them */
typedef struct ghost_struct {
    char ghost[NICKLEN];
    struct ghost_struct *next;
    struct ghost_struct *previous;
} GhostList;

int totalghosts = 0;

GhostList *firstGhost = NULL;
GhostList *lastGhost = NULL;

GhostList *getGhost(char *);
void addGhost(char *);
void delGhost(char *);

/* pretty simple, if someone desynchs doing something, yell about it,
 * then kill them for the evil desynch
 */

void nDesynch(char *nick, char *type)
{
    GhostList *killme;

    if (nick) {
	killme = getGhost(nick);

	if (killme)
	    return;

	sSend(":%s KILL %s :%s!%s (Desynch, type: [%s])", NickServ, nick,
	      services[1].host, NickServ, type);
    }
    return;
}


/*
 * ===DOC===
 * getNickData will find a nick and retreieve data for it
 * init*Data cleans up a user/nick ent
 * killide, uhm, killides someone, :)
 */

void initUserData(UserList *);
void annoyNickThief(UserList *);
void killide(char *);
void removeEnforcer(char *);

/*
 * ===DOC===
 * addNewUser simply adds a new user connection, :>
 */
void addNewUser(char **args, int numargs)
{
    int i;
    UserList *newnick;
    struct access_mod *ak;
    char realnamebuffer[129];
    newnick = smalloc(sizeof(UserList));
    initUserData(newnick);

    strcpy(newnick->nick, args[1]);
    if (args[4][0] == '~')
	args[4]++;
    strcpy(newnick->user, args[4]);
    strcpy(newnick->host, args[5]);
    strcpy(newnick->server, args[6]);

    *realnamebuffer = 0;
    if (numargs > 8) {
	args[8]++;
	strcpy(realnamebuffer, args[8]);
    }
    for (i = 9; i < numargs; i++) {
	strcat(realnamebuffer, " ");
	strcat(realnamebuffer, args[i]);
    }
    realnamebuffer[50] = 0;
    strcpy(newnick->realname, realnamebuffer);

    ak = isAKilled(newnick->nick, newnick->user, newnick->host);
    
    if (ak) {
	killUser(services[1], newnick, AKILLMSG);
	remUser(args[1]);
	if (!strcmp(ak->nick, "*")) {
            /* We need to send any matching exempts first. Rationale: The fact
             * that we've gotten to this point in the first place means that
             * one or more servers don't have an akill in place which they 
             * should have. If we only send the Akill again then exempted users
             * that try to connect to the server later on will no longer be 
             * able to use that server anymore either. Better to be save than
             * sorry and send any matching exempts out. -- Remmy 20020623 */
            send_matching_exempts(newnick->nick, newnick->user, newnick->host);
            /* Now send the akill out... */
	    sSend(":%s AKILL %s %s :Autokilled by %s [%s]", myname,
		  ak->host, ak->user, ak->setby, ak->reason);
        }
	addGhost(args[1]);
	timer(1800, delGhost, "delGhost", args[1]);
	return;
    }

    /* Check if user is supposed to be jinxed or not */
    ak = isHJinxed(newnick->nick, newnick->user, newnick->host);
    if (ak) {
	AddJinx(newnick->nick, "OperServ", "Host JINX");
        if (match(newnick->host, "*aol.com"))
            sSend(":%s GLOBOPS :Re-activating JINX for %s.", OperServ, 
                    newnick-> nick);
        sSend(":%s JINX %s", myname, newnick->nick);
    }

    newnick->reg = getRegNickData(args[1]);
    
    if (newnick->reg && (newnick->reg->flags & NACTIVE)) {
	newnick->reg->flags2 &= ~NSTICKYOP;	/* TURN OFF STICKY OPER UNTIL WE GET +o */
	if (newnick->reg->flags & NBANISH) {
	    killUser(services[1], newnick, NSBANISHMSG);
	    remUser(args[1]);
	    addGhost(args[1]);
	    timer(1800, delGhost, "delGhost", args[1]);
	    return;
	}
	if (!checkAccess(newnick->user, newnick->host, newnick->reg)) {
	    newnick->caccess = 1;
	    annoyNickThief(newnick);
	} else {
	    newnick->caccess = 2;
	    strcpy(newnick->reg->nick, newnick->nick);
	    strcpy(newnick->reg->user, newnick->user);
	    strcpy(newnick->reg->host, newnick->host);
	    strcpy(newnick->reg->realname, newnick->realname);
	    newnick->reg->timestamp = CTime;	/* MAKE SERVICES CLOCK MATTER MOST */
	    checkMemos(newnick);
	}
    }

    i = 0;

    i = addClone(newnick->nick, newnick->user, newnick->host);

    addNick(newnick);
    totalusers++;

    if (!i) {
	if (mostusers < totalusers) {
	    mostusers = totalusers;
	    if (mostusers % 5 == 0)
		sSend
		    (":%s GLOBOPS :New user-record established. There are now %i users online.",
		     Serenity, mostusers);
	}
    }

    /* Send out a message on connect */

    uSend(Serenity, newnick, GetReply(RPL_GLOBALNOTICE, newnick));
}

/*
 * ===DOC===
 * addNewUserSN adds a new user connection using SNICK protocol. -GZ
 */
void addNewUserSN(char **args, int numargs)
{
    int i, identified = 0;
    UserList *newnick;
    char *p;
    struct access_mod *ak;
    char realnamebuffer[129];
    char buffer[100];

    newnick = smalloc(sizeof(UserList));
    if (!newnick) {
		Debug((DEBUG_ERROR, "Failed to smalloc a newnick in addNewUserSN!"));
		return;
	 }
	 
	 initUserData(newnick);

    strcpy(newnick->nick, args[1]);
    if (args[4][0] == '~')
	args[4]++;
    strcpy(newnick->user, args[4]);
    strcpy(newnick->host, args[5]);
    strcpy(newnick->server, args[6]);

    *realnamebuffer = 0;
    if (numargs > 9) {
	args[9]++;
	strcpy(realnamebuffer, args[9]);
    }
    for (i = 10; i < numargs; i++) {
	strcat(realnamebuffer, " ");
	strcat(realnamebuffer, args[i]);
    }
    realnamebuffer[50] = 0;
    strcpy(newnick->realname, realnamebuffer);

    ak = isAKilled(newnick->nick, newnick->user, newnick->host);

    if (ak) {
	killUser(services[1], newnick, AKILLMSG);
	remUser(args[1]);
	if (!strcmp(ak->nick, "*"))
	    sSend(":%s AKILL %s %s :Autokilled by %s [%s]", myname,
		  ak->host, ak->user, ak->setby, ak->reason);
	addGhost(args[1]);
	timer(1800, delGhost, "delGhost", args[1]);
	return;
    }


    for (p = args[8]; *p; p++) {
	switch (*p) {
	case 'I':
	    identified = 1;
	    break;
        case 'w':
            newnick->webbie = 1;
            break;
	}
    }

    /* Check if user is supposed to be jinxed or not */
    /* don't hostjinx an identified user --InnerFIRE */
    if(!identified) { 
        ak = isHJinxed(newnick->nick, newnick->user, newnick->host);
        
        if (ak) {
	    AddJinx(newnick->nick, "OperServ", "Host JINX");
            sSend(":%s JINX %s", myname, newnick->nick);
        }
    }
    newnick->reg = getRegNickData(args[1]);
    
    if (newnick->reg && (newnick->reg->flags & NACTIVE)) {
	newnick->reg->flags2 &= ~NSTICKYOP;	/* TURN OFF STICKY OPER UNTIL WE GET +o */
	if (newnick->reg->flags & NBANISH) {
	    killUser(services[1], newnick, NSBANISHMSG);
	    remUser(args[1]);
	    addGhost(args[1]);
	    timer(1800, delGhost, "delGhost", args[1]);
	    return;
	}

	if (identified) {
	    newnick->caccess = 3;
	}

	if (!identified
	    && !checkAccess(newnick->user, newnick->host, newnick->reg)) {
	    newnick->caccess = 1;
	    annoyNickThief(newnick);
	} else if (!identified) {
	    newnick->caccess = 2;
	    newnick->reg->timestamp = CTime;	/* MAKE SERVICES CLOCK MATTER MOST */
	    if (!identified)
		checkMemos(newnick);
	}
    }

    i = 0;

    i = addClone(newnick->nick, newnick->user, newnick->host);

    addNick(newnick);
    totalusers++;

    if (!i)
	if (mostusers < totalusers)
	    mostusers = totalusers;

    /* Inject the modes back into the buffer -GZ */

    newnick->resynch = 1;

    bzero(buffer, 100);

    sprintf(buffer, ":%s MODE %s %s", args[1], args[1], args[8]);
    parseLine(buffer);

}

/* changeNick, called when someone uses the NICK command to change their
 * nicks obviously...changed from, to, and it hands off the timestamp
 * so that last seen time can be updated
 */
void changeNick(char *from, char *to, char *ts)
{
    UserList *changeme, *tmp;

    changeme = getNickData(from);
    if (changeme == NULL) {
	nDesynch(from, "NICK");
	return;
    }

    /* ALOT of changing around goes on here....it rather sucks...ohwell */
    tmp = smalloc(sizeof(UserList));
    strcpy(tmp->nick, to);
    strcpy(tmp->user, changeme->user);
    strcpy(tmp->host, changeme->host);
    strcpy(tmp->realname, changeme->realname);
    tmp->webbie = changeme->webbie;
    tmp->oflags = changeme->oflags;
    tmp->floodlevel = changeme->floodlevel;
    tmp->lastmsg = changeme->lastmsg;
    changeNickOnAllChans(changeme, tmp);

    /* If the user was jinxed (and nick thus got changed by SVSNICK) we should
     * remove the jinx for the old nick and re-add it for the new nick. Without
     * this change our jinx would get messed up. -- Remmy */
    if (CheckJinx(changeme->nick)) {
        DeleteJinx(changeme->nick);
        AddJinx(to, myname, "JINX following SVSNICK nick change");
    }
    
    delNick(changeme);
    addNick(tmp);

    if (isAKilled(tmp->nick, tmp->user, tmp->host)) {
	killUser(services[1], tmp, AKILLMSG);
	remUser(to);
	addGhost(to);
	timer(1800, delGhost, "delGhost", to);
	return;
    }

    tmp->reg = getRegNickData(to);

    if (tmp->reg && (tmp->reg->flags & NACTIVE)) {
	tmp->reg->flags2 &= ~NSTICKYOP;	/* TURN OFF STICKY OPER BIT */

	if ((tmp->reg->flags & NBANISH) && (!isOper(tmp))) {
	    killUser(services[1], tmp, NSBANISHMSG);
	    remUser(to);
	    addGhost(to);
	    timer(1800, delGhost, "delGhost", to);
	    return;
	}
	if (!checkAccess(tmp->user, tmp->host, tmp->reg)) {
	    tmp->caccess = 0;
	    annoyNickThief(tmp);
	}

	else {
	    tmp->caccess = 2;
	    strcpy(tmp->reg->nick, tmp->nick);
	    strcpy(tmp->reg->user, tmp->user);
	    strcpy(tmp->reg->host, tmp->host);
	    strcpy(tmp->reg->realname, tmp->realname);
	    if (isOper(tmp)) {
		tmp->reg->flags2 |= NSTICKYOP;	/* TURN ON STICKY OPER BIT */
	    }
	    if (ts[0] == ':')
		ts++;
	    tmp->reg->timestamp = CTime;	/* MAKE SERVICES TIME COUNT */

	    checkMemos(tmp);
	}
    }

    /* Check for HOSTJINX */
    if (isHJinxed(tmp->nick, tmp->user, tmp->host)) {
	AddJinx(tmp->nick, "OperServ", "Host JINX");
        sSend(":%s JINX %s", myname, tmp->nick);
    }

}


/* set mode, only deals with user modes, for things like away, or other
 * odd stuff, head to setFlags
 */
void setMode(char *nick, char *mode)
{
    UserList *changeme;
    int i, c = 1;
    changeme = getNickData(nick);

    if (changeme == NULL) {
	if (!isGhost(nick))
	    nDesynch(nick, "UMODE");
	return;
    }

    if (*mode == ':')
	mode++;			/* We don't need to bother with the : */

    for (i = 0; mode[i] != 0; i++) {
	switch (mode[i]) {
	case '-':
	    c = 0;
	    break;
	case '+':
	    c = 1;
	    break;
	case 'a':
	    if (c == 1) {
		if (!isServop(changeme)) {
		    sSend(":%s SVSMODE %s -a 0", OperServ, changeme->nick);
		    break;
		}
	    }
	case 'y':		/* NEW FLAG ADDITION FOR STAR-3.01+ LOCAL OPER DETECTION */
	    if (c == 1)
		changeme->oflags |= NISHELPOP;
	    else
		changeme->oflags &= ~NISHELPOP;
	    break;
	case 'x':
	    if (c == 1)
		changeme->oflags |= NISPLUSX;
	    else
		changeme->oflags &= ~NISPLUSX;
	    break;
	case 'i':
	    if (c == 1)
		changeme->oflags |= NISINVIS;
	    else
		changeme->oflags &= ~NISINVIS;
	    break;
	case 'O':
	case 'o':
	    if (c == 1) {
		if (changeme->reg && changeme->reg->flags & NNETADM
		    && changeme->caccess == 3 && !changeme->resynch) {
		    uSend(NickServ, changeme,
			  GetReply(RPL_YOUAREROOT, changeme));
		    sSend(":%s SVSMODE %s +aNn", myname, changeme->nick);
		}
		else if (changeme->reg && changeme->reg->flags & NROOT
		    && changeme->caccess == 3 && !changeme->resynch) {
		    uSend(NickServ, changeme,
			  GetReply(RPL_YOUAREROOT, changeme));
		    sSend(":%s SVSMODE %s +an", myname, changeme->nick);
		}
		else if (changeme->reg
		    && changeme->reg->flags & NSERVOP
		    && changeme->caccess == 3
		    && !changeme->resynch) {
		    uSend(NickServ, changeme,
			  GetReply(RPL_YOUARESERVOP, changeme));
		    sSend(":%s SVSMODE %s +a", myname, changeme->nick);
		}
		changeme->oflags |= NISOPER;
	    } else {
		changeme->oflags &= ~NISOPER;
	    }
	    break;

	case 'A':
	    if (c == 1)
		changeme->oflags |= NISADMIN;
	    else
		changeme->oflags &= ~NISADMIN;
	    break;

	}
    }

    mode--;

    if (changeme->reg) {
	if (isOper(changeme) && changeme->caccess > 1)
	    changeme->reg->flags2 |= NSTICKYOP;
	else
	    changeme->reg->flags2 &= ~NSTICKYOP;
    }

    changeme->resynch = 0;
    return;
}


/* set flags, like away, and you can use other things too I suppose...
 */
void setFlags(char *nick, int flag, char change)
{
    UserList *bah;
    bah = getNickData(nick);
    if (bah == NULL) {
	nDesynch(nick, "SETFLAG(internal source)");
	return;
    }

    else {
	if (change == '+')
	    bah->oflags |= flag;
	else
	    bah->oflags &= ~flag;
    }

}


/* simple, is the person an oper or not? */
int isOper(UserList * nick)
{
    if (nick == NULL)
	return 0;
    else {
	/* added helpop for little o's.... */
	if (nick->oflags & NISOPER || nick->oflags & NISHELPOP)
	    return 1;
	else
	    return 0;
    }
}

int isRoot(UserList * nick)
{
    if (!nick || !nick->reg)
	return 0;

    if ((nick->reg->flags & NROOT) && nick->caccess > 2)
	return isOper(nick);
    else
	return 0;
}

int isServop(UserList * nick)
{
    if (!nick || !nick->reg)
	return 0;

    if (
	(nick->reg->flags & NSERVOP || nick->reg->flags & NROOT) && nick->caccess > 2)
	return isOper(nick);
    else
	return 0;
}



/* remover a user when they signoff, save their info, and clean up
 * the nickname list
 */
void remUser(char *nick)
{
    UserList *killme;

    killme = getNickData(nick);
    if (killme == NULL)
	return;

    remFromAllChans(killme);
    if (killme->reg && killme->caccess > 1) {
	killme->reg->timestamp = CTime;	/* Adjust their last seen when they leave */
	cleanMemos(killme);
    }
    if (CheckJinx(killme->nick)) {
	DeleteJinx(killme->nick);
        if (match(killme->host, "*aol.com"))
            sSend(":%s GLOBOPS :%s left IRC. JINX has been removed", OperServ,
	          killme->nick);
    }
    delClone(killme->user, killme->host);
    delNick(killme);

    totalusers--;
}


/* INTERNAL FUNCTION -- ADD NICK RECORD TO USERLIST LIST */
void addNick(UserList * newnick)
{
    long hashEnt;

    if (!firstUser) {
	firstUser = newnick;
	newnick->previous = NULL;
    } else {
	lastUser->next = newnick;
	newnick->previous = lastUser;
    }

    lastUser = newnick;
    newnick->next = NULL;

    hashEnt = getHashKey(newnick->nick) % NICKHASHSIZE;
    if (!UserHash[hashEnt].nick) {
	UserHash[hashEnt].nick = newnick;
	newnick->hashprev = NULL;
    } else {
	UserHash[hashEnt].lastnick->hashnext = newnick;
	newnick->hashprev = UserHash[hashEnt].lastnick;
    }

    UserHash[hashEnt].lastnick = newnick;
    newnick->hashnext = NULL;

#ifdef HASHDEBUG
    sSend(":%s GNOTICE :%s hash entry is %i (ADD)", myname, newnick->nick,
	  hashEnt);
#endif
}

void addRegNick(RegNickList * newnick)
{
    long hashEnt;

    if (getRegNickData(newnick->nick)) {
	sfree(newnick);
	return;
    }

    if (!firstRegNick) {
	firstRegNick = newnick;
	newnick->previous = NULL;
    } else {
	lastRegNick->next = newnick;
	newnick->previous = lastRegNick;
    }

    lastRegNick = newnick;
    newnick->next = NULL;

    hashEnt = getHashKey(newnick->nick) % NICKHASHSIZE;
    if (!RegNickHash[hashEnt].nick) {
	RegNickHash[hashEnt].nick = newnick;
	newnick->hashprev = NULL;
    } else {
	RegNickHash[hashEnt].lastnick->hashnext = newnick;
	newnick->hashprev = RegNickHash[hashEnt].lastnick;
    }

    RegNickHash[hashEnt].lastnick = newnick;
    newnick->hashnext = NULL;

#ifdef HASHDEBUG
    sSend(":%s GNOTICE :%s hash entry is %i (ADD)", myname, newnick->nick,
	  hashEnt);
#endif
}


void delNick(UserList * killme)
{
    long hashEnt;
    UserList *tmpnick;

    if (!killme)
	return;

    if ((killme->next) || (killme->previous) || (killme == firstUser)) {	/* ARE WE IN THE TREE... IF NOT SKIP IT */
	if (killme->previous)
	    killme->previous->next = killme->next;
	else
	    firstUser = killme->next;
	if (killme->next)
	    killme->next->previous = killme->previous;
	else
	    lastUser = killme->previous;

	hashEnt = getHashKey(killme->nick) % NICKHASHSIZE;

	for (tmpnick = UserHash[hashEnt].nick; tmpnick;
	     tmpnick = tmpnick->hashnext) {
	    if (tmpnick == killme) {
		if (killme->hashprev)
		    killme->hashprev->hashnext = killme->hashnext;
		else
		    UserHash[hashEnt].nick = killme->hashnext;
		if (killme->hashnext)
		    killme->hashnext->hashprev = killme->hashprev;
		else
		    UserHash[hashEnt].lastnick = killme->hashprev;
		break;
	    }
	}
#ifdef HASHDEBUG
	sSend(":%s GNOTICE :%s hash entry is %i (DELETE)", myname,
	      killme->nick, hashEnt);
#endif
    }
    sfree(killme);
}

void delRegNick(RegNickList * killme)
{
    long hashEnt;
    RegNickList *tmpnick;
    UserList *online;

    MemoBox *memobox;
    MemoList *memo, *nextmemo;

    remfromallchanops(killme->nick);	/* REMOVE FROM ALL CHANNEL ACCESS LISTS - SYNCRONIZE CHANSERV */

    memobox = killme->memos;	/* WIPE OUT MEMO BOX -- SYNCH MEMOSERV */

    if (memobox) {
	for (memo = memobox->firstMemo; memo; memo = nextmemo) {
	    nextmemo = memo->next;
	    delMemo(memobox, memo);
	}
	sfree(memobox);
    }

    online = getNickData(killme->nick);
    if (online) {
	online->reg = NULL;
    }

    if (killme->previous)
	killme->previous->next = killme->next;
    else
	firstRegNick = killme->next;
    if (killme->next)
	killme->next->previous = killme->previous;
    else
	lastRegNick = killme->previous;

    hashEnt = getHashKey(killme->nick) % NICKHASHSIZE;
    for (tmpnick = RegNickHash[hashEnt].nick; tmpnick;
	 tmpnick = tmpnick->hashnext) {
	if (tmpnick == killme) {
	    if (killme->hashprev)
		killme->hashprev->hashnext = killme->hashnext;
	    else
		RegNickHash[hashEnt].nick = killme->hashnext;
	    if (killme->hashnext)
		killme->hashnext->hashprev = killme->hashprev;	/* AAHHHHHAAA!!!! */
	    else
		RegNickHash[hashEnt].lastnick = killme->hashprev;
	    break;
	}
    }
#ifdef HASHDEBUG
    sSend(":%s GNOTICE :%s hash entry is %i (DELETE)", myname,
	  killme->nick, hashEnt);
#endif
    sfree(killme);
}

/* get the nick item for the specified nick... */
UserList *getNickData(char *nick)
{
    long hashEnt;
    UserList *tmpnick;

    hashEnt = getHashKey(nick) % NICKHASHSIZE;
#ifdef HASHDEBUG
    sSend(":%s GNOTICE :%s has hash entry %lu (SEARCH)", myname, nick,
	  hashEnt);
#endif
    if (UserHash[hashEnt].nick) {
	for (tmpnick = UserHash[hashEnt].nick; tmpnick;
	     tmpnick = tmpnick->hashnext) {
	    if (!strcasecmp(tmpnick->nick, nick))
		return tmpnick;
	}
    }
    return NULL;
}

RegNickList *getRegNickData(char *nick)
{
    long hashEnt;
    RegNickList *tmpnick;

    hashEnt = getHashKey(nick) % NICKHASHSIZE;
#ifdef HASHDEBUG
    sSend(":%s GNOTICE :%s has hash entry %lu (SEARCH)", myname, nick,
	  hashEnt);
#endif

    if (RegNickHash[hashEnt].nick) {
	for (tmpnick = RegNickHash[hashEnt].nick; tmpnick;
	     tmpnick = tmpnick->hashnext) {
	    if (!strcasecmp(tmpnick->nick, nick))
		return tmpnick;
	}
    }

    /* DIDN'T FIND IT THERE, NOW LOOK THE HARD WAY 
       OH, WE PRAY WE DON'T NEED THIS ANYMORE

       for (tmpnick = firstRegNick; tmpnick; tmpnick = tmpnick->next)
       if (!strcasecmp(nick,tmpnick->nick))
       return tmpnick;
     */
    return NULL;
}


/* initNickData, this will clean up, or make it useable for new stuff
 * to be put into...
 */
void initUserData(UserList * nick)
{
    nick->nick[0] = 0;
    nick->user[0] = 0;
    nick->host[0] = 0;
    nick->server[0] = 0;
    nick->realname[0] = 0;
    nick->caccess = 0;
    nick->oflags = 0;
    nick->floodlevel = 0;
    nick->lastmsg = 0;
    nick->reg = NULL;
    nick->firstLink = nick->lastLink = NULL;
    nick->hashnext = nick->hashprev = NULL;
}

void initNickData(RegNickList * nick)
{
    nick->nick[0] = nick->user[0] = nick->host[0] = nick->realname[0] =
	nick->password[0] = nick->password2[0] = nick->url[0] =
	nick->email[0] = nick->info[0] = 0;
    nick->timestamp = nick->timereg = 0;
    nick->amasks = nick->flags = nick->chans = 0;
    nick->firstAcc = nick->lastAcc = NULL;
    nick->memos = NULL;
    nick->next = nick->previous = nick->hashnext = nick->hashprev = NULL;
}

/* IMPORTANT: make sure you understand this function, especially if you add
 * new commands to NickServ, it will tell you simply whether or not the
 * person has proper access to do something, along with the caccess flag
 */

int checkAccess(char *user, char *host, RegNickList * nick)
{
    nAccessList *tmp;
    char theirmask[HOSTLEN + USERLEN + 3];

    if (!nick)
	return 0;

    sprintf(theirmask, "%s@%s", user, host);

    if (!(nick->flags & NACTIVE))
	return -1;

    /* otherwise, check their access masks */
    for (tmp = nick->firstAcc; tmp; tmp = tmp->next) {
	if (!match(theirmask, tmp->mask))
	    return 1;		/* match! */
    }
    return 0;			/* NoMatch */
}

void sendToNickServ(UserList * tmp, char **args, int numargs)
{
    char from[NICKLEN];

    if (tmp == NULL) {
	nDesynch(from, "PRIVMSG");
	return;
    }

    strcpy(from, tmp->nick);

    if (!strcmp(args[0], "\001PING")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_PING1, tmp), args[1]);
	else
	    uSend(NickServ, tmp, GetReply(RPL_PING2, tmp), args[1],
		  args[2]);
	return;
    }

    else if (!strcasecmp(args[0], "help") && numargs == 1)
	sendHelp(from, NickServ, NULL, 0);
    else if (!strcasecmp(args[0], "help"))
	sendHelp(from, NickServ, args, numargs);
    else
	msgtabCheck(NickServ, tmp, args, numargs);

#ifdef SHOWSYNCHALL
    sSend(REPLY_DBSYNCH, NickServ, tmp->nick, ((nextNsync - CTime) / 60));
#endif
}

char LinkBuffer[512];

char *ListLinkId(UserList * nick)
{

    nLinkList *tmp;

    LinkBuffer[0] = '\0';

    for (tmp = nick->firstLink; tmp; tmp = tmp->next) {
	strcat(LinkBuffer, tmp->nick);
	strcat(LinkBuffer, " ");
    }
    return LinkBuffer;
}

int AddLinkId(UserList * nick, char *remote)
{

    nLinkList *tmp;

    for (tmp = nick->firstLink; tmp; tmp = tmp->next) {
	if (!strcasecmp(tmp->nick, remote)) {
	    uSend(NickServ, nick, GetReply(ERR_ALREADYLINKID, nick),
		  remote);
	    return 0;
	}
    }

    if (nick->linked >= MAXREMOTEID) {
	uSend(NickServ, nick, GetReply(ERR_MAXLINKID, nick), MAXREMOTEID);
	return 0;
    }

    tmp = smalloc(sizeof(nLinkList));
    strcpy(tmp->nick, remote);

    if (!nick->firstLink) {
	nick->firstLink = tmp;
	tmp->previous = NULL;
    } else {
	nick->lastLink->next = tmp;
	tmp->previous = nick->lastLink;
    }

    nick->lastLink = tmp;
    tmp->next = NULL;
    nick->linked++;

    return 1;
}

void NidentifyCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;
    int err;
    RegNickList *rid;
    time_t doot = time(NULL);
    char sendfrom[80];
    struct access_mod *ak;
    
    int old_access;

    old_access = tmp->caccess;

    if (!strcasecmp(args[0], "SIDENTIFY")) {
	if (!tmp->reg || !(tmp->reg->flags2 & NSIDENTIFY))
	    return;
    }

    /* Remote ID for nicknames. -GZ */

	if (numargs > 2)
		if (strcasecmp(args[2], from))
		{
			rid = getRegNickData(args[2]);
			if (!rid)
			{
				uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[2]);
				return;
			}

			if ((rid->flags & NBANISH) && (!isRoot(tmp)))
			{
				uSend(NickServ, tmp, GetReply(RPL_INFOBANISHED, tmp));
				return;
			}

	    if (!strncmp(args[1], rid->password, 15)) {
		if (AddLinkId(tmp, rid->nick)) {
		    uSend(NickServ, tmp, GetReply(RPL_PASSOKAY, tmp),
			  rid->nick);
		    strcpy(rid->user, tmp->user);
		    strcpy(rid->host, tmp->host);
		    strcpy(rid->realname, tmp->realname);
		    rid->timestamp = CTime;
		    tmp->remotefailid = 0;
		}
	    } else {
		tmp->remotefailid++;
		uSend(NickServ, tmp, GetReply(RPL_PASSNOTOKAY, tmp),
		      rid->nick);
	    }

	    /* Remote ID floodprotection to be, it's not pretty, but let's see how it will do. - GZ */

		if (tmp->remotefailid == 3)
		{
			uSend(NickServ, tmp, GetReply(RPL_EMAILHACK1, tmp));
			sSend(":%s GLOBOPS :\002Warning:\002 Possible remote-ID hacking from %s!%s@%s (%s)", NickServ, tmp->nick, tmp->user, tmp->host, rid->nick);
		}
		else if (tmp->remotefailid == 6)
		{
			sSend(":%s GLOBOPS :\002Warning:\002 Continued remote-ID hacking from %s!%s@%s (%s)", NickServ, tmp->nick, tmp->user, tmp->host, rid->nick);
			killUser(services[1], tmp, "Nickname hacking");
		}

	    return;
	}

    if (!tmp->reg || !tmp->reg->flags & NACTIVE) {
	uSend(NickServ, tmp, GetReply(ERR_NEEDACTNICK, tmp));
	return;
    }

    if (!strncmp(args[1], tmp->reg->password, 15)) {
	tmp->caccess = 3;
	tmp->failid = 0;	/* o O. it's the real person here better start all over. */
	strcpy(tmp->reg->nick, tmp->nick);
	strcpy(tmp->reg->user, tmp->user);
	strcpy(tmp->reg->host, tmp->host);
	strcpy(tmp->reg->realname, tmp->realname);
	tmp->reg->timestamp = CTime;

	/* Set user +I after succesful identification
         * or +Iw for webbies */
        if (tmp->reg->flags2 & NMSGMODE) {
            tmp->webbie = 1;
	    sSend(":%s SVSMODE %s +Iw", myname, from);
        } else {
            /* Person is no longer a webbie. Since IRC ops sometimes want
             * to test 'being' a webbie, we should reset the flags if
             * they ID back to their own non webbie nick */
            tmp->webbie = 0; 
            sSend(":%s SVSMODE %s +I-w", myname, from);
        }
        
        /* Tell them the password is ok */
	uSend(NickServ, tmp, GetReply(RPL_PASSOKAY, tmp), tmp->nick);
        
	tmp->oflags |= NISIDENT;

	if (isOper(tmp))
	    tmp->reg->flags2 |= NSTICKYOP;

	if ((tmp->reg->flags & NROOT) && isOper(tmp))
	    uSend(NickServ, tmp, GetReply(RPL_YOUAREROOT, tmp));
	else if (tmp->reg->flags & NSERVOP && isOper(tmp))
	    uSend(NickServ, tmp, GetReply(RPL_YOUARESERVOP, tmp));

	/* HELP FORGETFUL SERVOPS BY GIVING THEM THEIR +a BACK */
	if ((tmp->reg->flags & NSERVOP) && isOper(tmp))
	    sSend(":%s SVSMODE %s +a", myname, from);

	/* THIS IS FOR THE SERVICES ROOT, NETADMIN AND MASKS (+n/+N) */
	if ((tmp->reg->flags & NNETADM) && isOper(tmp))
	    sSend(":%s SVSMODE %s +Nna", myname, from);	/* Network Administrator */

	else if ((tmp->reg->flags & NROOT) && isOper(tmp))
	    sSend(":%s SVSMODE %s +na", myname, from);	/* Services Root */
	
	/* Remove a jinx only if it was a hostjinx */
        ak = isHJinxed(tmp->nick, tmp->user, tmp->host);
        
        if (ak) {
            DeleteJinx(tmp->nick) ;
            sSend(":%s UNJINX %s", myname, tmp->nick);
        }

	checkMemos(tmp);

	return;
    } else {
	if (tmp->caccess > 1) { /* Remove +I flag after failed ID */
	    sSend(":%s SVSMODE %s -I", myname, from);
	}

	tmp->caccess = 1;	/* failed identify takes away even the access list ID */
	uSend(NickServ, tmp, GetReply(RPL_PASSNOTOKAY, tmp), tmp->nick);

	tmp->failid++;		/* If you fail id, add one to it */
	if (tmp->failid == 3) {
	    uSend(NickServ, tmp, GetReply(RPL_EMAILHACK1, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_EMAILHACK2, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_EMAILHACK3, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_EMAILHACK4, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_EMAILHACK5, tmp));

	    sprintf(sendfrom, "Nickname Services <%s@%s>", NickServ,
		    NETHOST);
	    err =
		sendMail(tmp->reg->email, sendfrom,
			 "Incorrect password notification.",
			 "Hello %s,\n\n"
			 "%s has determined that someone entered an incorrect password\n"
                         "for your nick several times. The IP address of this individual\n"
			 "is %s@%s .\n\n"
			 "In case you forgot the correct password for %s, it is: %s\n\n"
			 "If you receive more than three (3) of these notices, please visit\n"
			 "#help to request an IRC operator to assist you.\n\n"
			 "This email was sent at: %s\n"
                         "Please do not reply to this email as your response will not be read.\n", 
                         tmp->nick, NickServ, tmp->user, tmp->host, tmp->nick,
			 tmp->reg->password, ctime(&(doot)));
	    logDump(nicklog,
		    "Automailed pass for the nick %s to %s due to hacking from %s@%s.",
		    from, tmp->reg->email, tmp->user, tmp->host);
	}

	if (tmp->failid > 5) {
	    sSend
		(":%s GLOBOPS :\002Warning:\002 Nickname-hacking attempt from %s!%s@%s (user has been killed).",
		 NickServ, from, tmp->user, tmp->host);
	    uSend(NickServ, tmp, GetReply(RPL_EMAILHACK6, tmp));
	    logDump(nicklog, "Persisted hacking for the nick %s by %s@%s.",
		    from, tmp->user, tmp->host);
	    killUser(services[1], tmp, "Nickname hacking");
	}

    }
}

void NsetCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;

    if (!strcasecmp(args[1], "kill") || !strcasecmp(args[1], "guard")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_NSETIS, tmp),
		  "Nickname Guard",
		  (tmp->reg->flags & NKILL) ? "ON" : "OFF");
	else {
	    if (!strcasecmp(args[2], "on")) {
		tmp->reg->flags |= NKILL;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Nickname Guard", "ON");
	    } else if (!strcasecmp(args[2], "off")) {
		tmp->reg->flags &= ~NKILL;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Nickname Guard", "OFF");
	    } else {
		uSend(NickServ, tmp, GetReply(ERR_BADSETTING, tmp),
		      args[2]);
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET GUARD");
	    }
	}
    }

    else if (!strcasecmp(args[1], "noop")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_NSETIS, tmp),
		  "Nickname No-Ops",
		  (tmp->reg->flags & NNOOP) ? "ON" : "OFF");
	else {
	    if (!strcasecmp(args[2], "on")) {
		tmp->reg->flags |= NNOOP;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Nickname No-Ops", "ON");
	    } else if (!strcasecmp(args[2], "off")) {
		tmp->reg->flags &= ~NNOOP;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Nickname No-Ops", "OFF");
	    } else {
		uSend(NickServ, tmp, GetReply(ERR_BADSETTING, tmp),
		      args[2]);
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET NOOP");
	    }
	}
    }

    else if (!strcasecmp(args[1], "sidentify")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_NSETIS, tmp),
		  "Use SIDENTIFY",
		  (tmp->reg->flags2 & NSIDENTIFY) ? "ON" : "OFF");
	else {
	    if (!strcasecmp(args[2], "on")) {
		tmp->reg->flags2 |= NSIDENTIFY;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Use SIDENTIFY", "ON");
	    } else if (!strcasecmp(args[2], "off")) {
		tmp->reg->flags2 &= ~NSIDENTIFY;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Use SIDENTIFY", "OFF");
	    } else {
		uSend(NickServ, tmp, GetReply(ERR_BADSETTING, tmp),
		      args[2]);
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET SIDENTIFY");
	    }
	}
    }

    else if (!strcasecmp(args[1], "language")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_NSETIS, tmp),
		  "Services Language", GetLanguage(tmp));
	else {
	    if (!strcasecmp(args[2], "english")) {
		tmp->reg->language = 0;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else if (!strcasecmp(args[2], "dutch")) {
		tmp->reg->language = 1;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else if (!strcasecmp(args[2], "french")) {
		tmp->reg->language = 2;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else if (!strcasecmp(args[2], "german")) {
		tmp->reg->language = 3;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else if (!strcasecmp(args[2], "lcars")) {
		tmp->reg->language = 4;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else if (!strcasecmp(args[2], "spanish")) {
		tmp->reg->language = 5;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else if (!strcasecmp(args[2], "cajun")) {
		tmp->reg->language = 6;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Services Language", GetLanguage(tmp));
	    } else {
		uSend(NickServ, tmp, GetReply(ERR_BADSETTING, tmp),
		      args[2]);
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET LANGUAGE");
	    }
	}
    }

    else if (!strcasecmp(args[1], "hide")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_NSETIS, tmp),
		  "Hostmask Hiding",
		  (tmp->reg->flags & NHIDE) ? "ON" : "OFF");
	else {
	    if (!strcasecmp(args[2], "on")) {
		tmp->reg->flags |= NHIDE;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Hostmask Hiding", "ON");
	    } else if (!strcasecmp(args[2], "off")) {
		tmp->reg->flags &= ~NHIDE;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Hostmask Hiding", "OFF");
	    } else {
		uSend(NickServ, tmp, GetReply(ERR_BADSETTING, tmp),
		      args[2]);
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET HIDE");
	    }
	}
    }

    else if (!match(args[1], "password")) {
	if (numargs < 3) {
	    uSend(NickServ, tmp, GetReply(ERR_BADPARAMS, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ,
		  "SET PASSWORD");
	} else {
	    if ((4 > strlen(args[2])) || (15 < strlen(args[2])))
		uSend(NickServ, tmp, GetReply(ERR_BADPASSWORD, tmp));
	    else {
		strncpy(tmp->reg->password, args[2], 15);
		uSend(NickServ, tmp, GetReply(RPL_PASSWORDNOW, tmp),
		      tmp->reg->password);
		tmp->reg->lastpwchange = CTime;
		logDump(nicklog, "%s!%s@%s: changed password.\n", from,
			tmp->user, tmp->host);

	    }
	}
    }

    else if (!strcasecmp(args[1], "email")) {
	if (numargs < 3)
	    uSend(NickServ, tmp, GetReply(RPL_NSETIS, tmp),
		  "Email Address",
		  (tmp->reg->flags & NEMAIL) ? "PUBLIC" : "PRIVATE");
	else {
	    if (!strcasecmp(args[2], "on")
		|| !strcasecmp(args[2], "public")) {
		tmp->reg->flags |= NEMAIL;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Email Address", "PUBLIC");
	    } else if (!strcasecmp(args[2], "off")
		       || !strcasecmp(args[2], "private")) {
		tmp->reg->flags &= ~NEMAIL;
		uSend(NickServ, tmp, GetReply(RPL_NSETNOW, tmp),
		      "Email Address", "PRIVATE");
	    } else {
		uSend(NickServ, tmp, GetReply(ERR_BADSETTING, tmp),
		      args[2]);
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET EMAIL");
	    }
	}
	} else if (!strcasecmp(args[1], "address")) {
	char sendfrom[80];
	int err;

	RegNickList *target;

	if (numargs > 3) {
	    if (!isServop(tmp)) {
		uSend(NickServ, tmp, GetReply(ERR_NEEDSERVOP, tmp));
		return;
	    }

	    target = getRegNickData(args[3]);

	    if (!target) {
		uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[3]);
		return;
	    }

	    if ((!strchr(args[2], '@')) || (strlen(args[2]) < 5)
		|| (strlen(args[2]) > EMAILLEN)) {
		uSend(NickServ, tmp, GetReply(ERR_VALIDEMAIL, tmp));
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET ADDRESS");
		return;
	    }

	    sSend(":%s GLOBOPS :%s changed email address for %s from %s to %s",
		  NickServ, from, args[3], target->email, args[2]);

		logDump(nicklog, "%s!%s@%s: changed address for %s from '%s' to '%s'\n",
		    from, tmp->user, tmp->host, args[3], target->email, args[2]);

		strcpy(target->email, args[2]);

	    return;
	}

	if (numargs < 3) {
	    uSend(NickServ, tmp, GetReply(ERR_BADPARAMS, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ,
		  "SET ADDRESS");
	} else {
	    srandom(time(NULL));

	    if ((!strchr(args[2], '@')) || (strlen(args[2]) < 5)
		|| (strlen(args[2]) > EMAILLEN)) {
		uSend(NickServ, tmp, GetReply(ERR_VALIDEMAIL, tmp));
		uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp),
		      NickServ, "SET ADDRESS");
		return;
	    }

	    if (countRegAddy(args[2]) >= NickLimit) {
		uSend(NickServ, tmp, GetReply(ERR_NICKLIMIT, tmp));
		return;
	    }

	    if (VerifyEmail(args[2])) {
		uSend(NickServ, tmp, GetReply(ERR_NOWEBMAIL, tmp));
		return;
	    }

	    sprintf(tmp->reg->password2, "%lu", (long) random());
	    strcpy(tmp->reg->email, args[2]);
	    sprintf(sendfrom, "Nickname services <%s@%s>", NickServ,
		    NETHOST);
	    err =
		sendMail(args[2], sendfrom, "Nickname activation key",
			 "\nYour nickname registration has been verified\n"
			 "To activate your nickname: switch to the nickname %s and\n"
			 "/msg %s ACTIVATE %s\n"
			 "%s is your activation key\n"
			 "Even when your nickname is registered, Do not lose this PLEASE.",
			 tmp->nick, NickServ, tmp->reg->password2,
			 tmp->reg->password2);
	    if (!err) {
		uSend(NickServ, tmp, GetReply(RPL_TEMPDEAC1, tmp));
		uSend(NickServ, tmp, GetReply(RPL_TEMPDEAC2, tmp));
		uSend(NickServ, tmp, GetReply(RPL_TEMPDEAC3, tmp));
		uSend(NickServ, tmp, GetReply(RPL_TEMPDEAC4, tmp),
		      args[2]);
		tmp->reg->flags &= ~NACTIVE;
		tmp->reg->flags |= NDEACC;
		logDump(nicklog, "%s!%s@%s: changed address to '%s'\n",
			from, tmp->user, tmp->host, args[2]);
	    } else
		uSend(NickServ, tmp, GetReply(ERR_NOMAILSENT, tmp));
	}
    }

    else
	uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ,
	      "SET");
}

void NregisterCmd(UserList * tmp, char **args, int numargs)
{
    char *from = tmp->nick;
    time_t doot = time(NULL);
    char scanhost[512];
    char sendfrom[80];
    int err;
    UserList *msgtmp=tmp;
    int donefudge=0;

    srandom(time(NULL));
    
	if(!match(args[1], "*webtv.net*"))
	{
		msgtmp = smalloc(sizeof(UserList));
		memcpy(msgtmp,tmp,sizeof(UserList));
		msgtmp->reg=smalloc(sizeof(RegNickList));
		msgtmp->reg->flags2 = NMSGMODE;
                msgtmp->webbie = 1;
		uSend(NickServ,msgtmp,"You have been detected as a WebTV user, if that is not true, please ask anyone in #help to change your message mode.");
		donefudge=1;
	}

	if (tmp->reg)
	{
		if (!tmp->reg->flags & NACTIVE)
		{
			uSend(NickServ, msgtmp, GetReply(ERR_NEEDTOACT, tmp));
			uSend(NickServ, msgtmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ, "ACTIVATE");
		}
		else uSend(NickServ, msgtmp, GetReply(ERR_ALREADYREG, tmp));

		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

	if ((!strncasecmp(from, "Guest", 5)) || (!strcasecmp(from, "done")) ||
            (!strncasecmp(from, "SereneMIRC", 10)) || (!strncasecmp(from, "SerenePIRCH", 11)))
	{
		uSend(NickServ, msgtmp, GetReply(ERR_CANNOTREGTHIS, tmp));
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

	if (numargs < 3)
	{
		uSend(NickServ, msgtmp, GetReply(ERR_BADPARAMS, tmp));
		uSend(NickServ, msgtmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ, "REGISTER");
		
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}
	
	if ((!strchr(args[1], '@')) || (strlen(args[1]) < 5) || (strlen(args[1]) > EMAILLEN))
	{
		uSend(NickServ, msgtmp, GetReply(ERR_VALIDEMAIL, tmp));
		uSend(NickServ, msgtmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ, "REGISTER");
		
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

	if (VerifyEmail(args[1]))
	{
		uSend(NickServ, msgtmp, GetReply(ERR_NOWEBMAIL, tmp));
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

    sprintf(scanhost, "*%s*", tmp->host);

	if (!match(args[1], scanhost))
	{
		uSend(NickServ, tmp, GetReply(ERR_BANNEDEMAIL, tmp));
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

	if ((15 < strlen(args[2])) || (4 > strlen(args[2])))
	{
		uSend(NickServ, msgtmp, GetReply(ERR_BADPASSWORD, tmp));
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

	if (countRegAddy(args[1]) >= NickLimit)
	{
		uSend(NickServ, msgtmp, GetReply(ERR_NICKLIMIT, tmp));
		sSend(":%s GLOBOPS :%s (%s) is trying to register too many nicks.", NickServ, from, args[1]);
		if(donefudge)
		{
			sfree(msgtmp->reg);
			sfree(msgtmp);
		}
		return;
	}

	if(donefudge)
	{
		sfree(msgtmp->reg);
		sfree(msgtmp);
	}
	
    tmp->reg = smalloc(sizeof(RegNickList));
    strcpy(tmp->reg->nick, tmp->nick);
    strcpy(tmp->reg->user, tmp->user);
    strcpy(tmp->reg->host, tmp->host);
    sprintf(tmp->reg->password2, "%lu", (long) random());
    tmp->reg->timestamp = CTime;
    tmp->reg->timereg = CTime;
    tmp->reg->lastpwchange = CTime;
    tmp->reg->flags |= NREG;
    tmp->reg->flags |= NHIDE;
    if(donefudge) {
        tmp->reg->flags2 = NMSGMODE;
        tmp->webbie = 1;
    }
    tmp->caccess = 3;
    strcpy(tmp->reg->email, args[1]);
    strncpy(tmp->reg->password, args[2], 15);
    sprintf(sendfrom, "Nickname Services <%s@%s>", NickServ, NETHOST);
    err = sendMail(args[1], sendfrom, "Nickname activation key",
		   "Hello %s, Welcome to the Serenity-IRC network!\n\n"
		   "Your nickname registration has been verified.\n\n"
		   "To activate your nickname:\n\n"
		   "1) switch to the nickname %s\n"
                   "2) /msg %s ACTIVATE %s\n\n"
		   "%s is your activation key.  %s is your password.\n\n"
		   "Please try to remember your password, you might need it\n"
		   "some day.\n\n"
		   "If you accidently lose your password, you can always\n"
		   "/join #help and ask one of our IRC Operators to recover\n"
		   "it for you. We will be more than glad to help you!\n\n"
		   "If you are new to IRC or Serenity-IRC, we would like to recommend\n"
		   "#help, as it is the place where you can obtain help or any\n"
		   "other kind of information that is IRC related.\n\n"
		   "Enjoy your stay on our safe and friendly network!\n\n"
		   "For additional help, try these commands:\n"
		   "     /msg %s HELP\n"
		   "     /msg %s HELP\n"
		   "     /msg %s HELP\n"
		   "\nNOTICE: This mail sent by: %s@%s at %s",
		   tmp->nick, tmp->nick, NickServ, tmp->reg->password2,
		   tmp->reg->password2, tmp->reg->password, NickServ,
		   ChanServ, MemoServ, tmp->user, tmp->host,
		   ctime(&(doot)));

    if (!err)
    {
	uSend(NickServ, tmp, GetReply(RPL_TEMPDEAC4, tmp), args[1]);
	logDump(nicklog, "%s!%s@%s: ACTIVATION key sent to '%s'\n", from,
		tmp->user, tmp->host, args[1]);
	addRegNick(tmp->reg);
    } else {
	uSend(NickServ, tmp, GetReply(ERR_NOMAILSENT, tmp));
	sfree(tmp->reg);
	tmp->reg = NULL;
    }
    return;
}

void NactivateCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;
    time_t doot;

    if (strcmp(args[1], tmp->reg->password2)) {
	uSend(NickServ, tmp, GetReply(ERR_ACCESSDENIED, tmp));
	return;
    }

    if (!(tmp->reg->flags & NACTIVE)) {
	char user2[USERLEN], host2[HOSTLEN];

	strcpy(user2, tmp->user);
	strcpy(host2, tmp->host);

	doot = time(NULL);

	if (!(tmp->reg->flags & NDEACC)) {
	    mostnicks++;
	    strcpy(tmp->reg->nick, tmp->nick);
	    tmp->reg->timereg = time(NULL);
		tmp->reg->firstAcc = tmp->reg->lastAcc = NULL;
/*	    tmp->reg->firstAcc = smalloc(sizeof(nAccessList));
	    mask(user2, host2, 0, tmp->reg->firstAcc->mask);
	    tmp->reg->firstAcc->index = 1;
	    tmp->reg->firstAcc->next = tmp->reg->firstAcc->previous = NULL;
	    tmp->reg->lastAcc = tmp->reg->firstAcc; */
	    tmp->reg->amasks = 0;
	    tmp->caccess = 3;
	    tmp->reg->flags |= NACTIVE;
	    tmp->reg->flags |= NHIDE;
	    tmp->reg->url[0] = 0;
	    tmp->reg->chans = 0;
	    strcpy(tmp->reg->user, tmp->user);
	    strcpy(tmp->reg->host, tmp->host);
	    strcpy(tmp->reg->realname, tmp->realname);
	    /* Set user +I after succesful identification
             * or +Iw for webbies */
            if (tmp->reg->flags2 & NMSGMODE) {
                tmp->webbie = 1;
	        sSend(":%s SVSMODE %s +Iw", myname, from);
            } else {
                /* Person is no longer a webbie. Since IRC ops sometimes want
                 * to test 'being' a webbie, we should reset the flags if
                 * they ID back to their own non webbie nick */
                tmp->webbie = 0; 
                sSend(":%s SVSMODE %s +I-w", myname, from);
            }

	    uSend(NickServ, tmp, GetReply(RPL_ACTIVATE1, tmp), from);
/*	    uSend(NickServ, tmp, GetReply(RPL_ACTIVATE2, tmp),
		  tmp->reg->firstAcc->mask);
	    uSend(NickServ, tmp, GetReply(RPL_ACTIVATE3, tmp));*/
	    uSend(NickServ, tmp, GetReply(RPL_ACTIVATE4, tmp),
		  ctime(&(doot)));
	    uSend(NickServ, tmp, GetReply(RPL_ACTIVATE5, tmp),
		  tmp->reg->password);
	    uSend(NickServ, tmp, GetReply(RPL_ACTIVATE6, tmp));
	    logDump(nicklog, "%s!%s@%s: REGISTER\n", from, tmp->user,
		    tmp->host);
	} else {
	    uSend(NickServ, tmp, GetReply(RPL_REACTIVATE, tmp));
	    tmp->reg->flags |= NACTIVE;
	    tmp->reg->flags &= ~NDEACC;
	}
    } else {
	uSend(NickServ, tmp, GetReply(RPL_ALREADYACT, tmp));
    }
}

char AccessString[128];
char OptionString[128];

char *GetAccessString(RegNickList * nick)
{

    AccessString[0] = '\0';

    if (nick->flags & NROOT)
	sprintf(AccessString, "Services Root Administrator");
    else if (nick->flags & NSERVOP)
	sprintf(AccessString, "Services Operator");
    else if (nick->flags2 & NSTICKYOP)
	sprintf(AccessString, "IRC Operator");

    return AccessString;
}

char *GetOptionString(RegNickList * nick)
{

    OptionString[0] = '\0';
    if (nick->flags & NKILL)
	strcat(OptionString, "guard ");
    if (nick->flags & NHOLD)
	strcat(OptionString, "held ");
    if (nick->flags & NMARKED)
	strcat(OptionString, "marked ");
    if (nick->flags & NNOOP)
	strcat(OptionString, "noop ");
    if (nick->flags & NHIDE)
	strcat(OptionString, "hide ");
    if (nick->flags2 & NMSGMODE)
	strcat(OptionString, "msgmode ");
    if (nick->flags & NVACATION)
	strcat(OptionString, "vacation ");

    return OptionString;
}

char *GetMemoString(RegNickList * nick)
{

    char redirect[128];
    OptionString[0] = redirect[0] = '\0';

    if (nick->memos) {
	if (nick->memos->flags & MNOMEMO)
	    strcat(OptionString, "nomemo ");
	if (nick->memos->flags & MNOFORWARD)
	    strcat(OptionString, "noforward ");
	if ((nick->memos->flags & MFORWARDED)
	    && (nick->memos->redirect[0])) {
	    sprintf(redirect, "redirect[%s] ", nick->memos->redirect);
	    strcat(OptionString, redirect);
	}
    }

    return OptionString;
}

char *GetAccess(UserList * nick)
{

    AccessString[0] = '\0';

    switch (nick->caccess) {
    case 1:
	strcpy(AccessString, "Not Identified");
	break;
    case 2:
	strcpy(AccessString, "Matching access list");
	break;
    case 3:
	strcpy(AccessString, "Identified");
	break;
    default:
	strcpy(AccessString, "Unknown");
	break;
    }

    return AccessString;
}

void NinfoCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;

    int i = 0;
    nMarkList *ml = NULL;
    RegNickList *err;
    UserList *online = NULL;

    if (numargs > 1)
	err = getRegNickData(args[1]);
    else
	err = tmp->reg;

    if (!err) {
	uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp),
	      numargs > 1 ? args[1] : from);
	return;
    }

    uSend(NickServ, tmp, GetReply(RPL_INFOHEADER, tmp), err->nick);

    if (err->flags & NBANISH) {
	uSend(NickServ, tmp, GetReply(RPL_INFOBANISHED, tmp));
	uSend(NickServ, tmp, GetReply(RPL_INFOFOOTER, tmp));
	return;
    }

    if ((!(err->flags & NREG)) && (err->flags & NACTIVE)) {
	uSend(NickServ, tmp, GetReply(RPL_INFODEREG, tmp), err->nick);
	uSend(NickServ, tmp, GetReply(RPL_INFOFOOTER, tmp));
	return;
    }

    if ((online = getNickData(err->nick)) && (err->flags & NREG))
	uSend(NickServ, tmp, GetReply(RPL_INFONOWON, tmp), err->nick);

    if ((err->flags & NACTIVE) && (err->flags & NREG)) {
	struct tm *time;
	char temp[80];

	if (isOper(tmp)
	    || (!strcasecmp(err->nick, from)
		&& tmp->caccess == 3)) uSend(NickServ, tmp,
					     GetReply(RPL_INFOLASTSEEN1,
						      tmp), err->user,
					     err->host, err->realname);
	else if (online)
	    uSend(NickServ, tmp, GetReply(RPL_INFOLASTSEEN1, tmp),
		  err->user, maskuser(online), err->realname);
	else
	    uSend(NickServ, tmp, GetReply(RPL_INFOLASTSEEN1, tmp),
		  err->user, maskreguser(err), err->realname);

	time = gmtime(&CTime);
	strftime(temp, 80, "%a %d/%m/%Y %T %Z", time);
	uSend(NickServ, tmp, GetReply(RPL_INFOTIMENOW, tmp), temp);

	time = gmtime(&err->timestamp);
	strftime(temp, 80, "%a %d/%m/%Y %T %Z", time);
	uSend(NickServ, tmp, GetReply(RPL_INFOLASTSEEN2, tmp), temp);

	time = gmtime(&err->timereg);
	strftime(temp, 80, "%a %d/%m/%Y %T %Z", time);
	uSend(NickServ, tmp, GetReply(RPL_INFOREGGEDON, tmp), temp);

	if (err->flags & NEMAIL || isOper(tmp)
	    || (!strcasecmp(err->nick, from) && tmp->caccess == 3))
	    uSend(NickServ, tmp, GetReply(RPL_INFOEMAIL, tmp), err->email);

	temp[0] = '\0';
	strcpy(temp, GetAccessString(err));
	if (temp[0])
	    uSend(NickServ, tmp, GetReply(RPL_INFOACCESS, tmp), temp);

	temp[0] = '\0';
	strcpy(temp, GetOptionString(err));
	if (temp[0] && isOper(tmp))
	    uSend(NickServ, tmp, GetReply(RPL_INFOOPTIONS, tmp), temp);

	temp[0] = '\0';
	strcpy(temp, GetMemoString(err));
	if (temp[0] && isOper(tmp))
	    uSend(NickServ, tmp, GetReply(RPL_INFOMEMO, tmp), temp);

	temp[0] = '\0';
	if (online)
	    strcpy(temp, ListLinkId(online));
	if (temp[0])
	    uSend(NickServ, tmp, GetReply(RPL_INFOLINKID, tmp), temp);

	if (online)
	    uSend(NickServ, tmp, GetReply(RPL_INFOCURRENTACC, tmp),
		  GetAccess(online));

	if (isOper(tmp)) {
	    err->chans = countRegChans(err->nick);
	    if (err->memos)
		uSend(NickServ, tmp, GetReply(RPL_INFODATA, tmp),
		      err->chans, err->memos->memos, err->amasks,
		      (online) ? online->floodbytes : 0);
	    else
		uSend(NickServ, tmp, GetReply(RPL_INFODATA, tmp),
		      err->chans, 0, err->amasks,
		      (online) ? online->floodbytes : 0);

	    if (err->firstMark != NULL) {
		for (ml = err->firstMark; ml != NULL; ml = ml->next) {
		time = gmtime(&ml->timestamp);
		strftime(temp, 80, "%d/%m/%Y", time);
		uSend(NickServ, tmp, GetReply(RPL_INFOREMARK, tmp),
		      ++i, ml->markedby, temp, ml->remark);
		}
	    }
	}

	uSend(NickServ, tmp, GetReply(RPL_INFOFOOTER, tmp));

	return;
    } else if (err->flags & NREG) {
	uSend(NickServ, tmp, GetReply(RPL_INFOSTATUS, tmp),
	      "Registered but not activated");
	if (isOper(tmp)) {
	    if ((err->flags & NMARKED) && isOper(tmp))
		uSend(NickServ, tmp, GetReply(RPL_INFOOPTIONS, tmp),
		      "marked");
	    uSend(NickServ, tmp, GetReply(RPL_INFOEMAIL, tmp), err->email);
	    uSend(NickServ, tmp, GetReply(RPL_INFOREGBY, tmp), err->user,
		  err->host);
	}

	uSend(NickServ, tmp, GetReply(RPL_INFOFOOTER, tmp));

	return;
    } else {
	uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp),
	      numargs > 1 ? args[1] : from);
	return;
    }
}


void NdropCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;

    if (!strncmp(args[1], tmp->reg->password, 15)) {
	mostnicks--;
	uSend(NickServ, tmp, GetReply(RPL_NICKDROPPED, tmp), from);
	logDump(nicklog, "%s!%s@%s: DROP\n", from, tmp->user, tmp->host);
	delRegNick(tmp->reg);
	return;
    }
}


void addAccessMask(char *mask, UserList * nick)
{

    if (USERLEN + HOSTLEN + 3 < strlen(mask)) {
	sSend
	    (":%s GLOBOPS :\002Warning:\002 %s tried to add a HUGE access mask",
	     NickServ, nick->nick);
	uSend(NickServ, nick, GetReply(ERR_BADACCMASK, nick));
	return;
    }

    if (nick->reg->amasks >= AccessLimit) {
	uSend(NickServ, nick, GetReply(ERR_TOOMANYACCMASKS, nick));
	return;
    }

    if (!addAccItem(nick->reg, mask))
	uSend(NickServ, nick, GetReply(ERR_ACCMASKNOTADDED, nick), mask);
    else
	uSend(NickServ, nick, GetReply(RPL_ACCMASKADDED, nick), mask);
}

void delAccessMask(char *mask, UserList * nick)
{
    nAccessList *tmp;
    char deleted[71];

    if (!nick->reg->firstAcc)
	return;
    if (*mask == '-') {
	for (tmp = nick->reg->firstAcc; tmp; tmp = tmp->next) {
	    delAccItem(nick->reg, tmp->mask, deleted);
	}
	uSend(NickServ, nick, GetReply(RPL_ACCLISTCLEARED, nick));
	nick->reg->amasks = 0;
	return;
    } else {
	if (delAccItem(nick->reg, mask, deleted)) {
	    uSend(NickServ, nick, GetReply(RPL_ACCMASKREMOVED, nick),
		  deleted);
	    nick->reg->amasks--;
	} else if (!strchr(mask, '@'))
	    uSend(NickServ, nick, GetReply(ERR_ACCENTRYNOTFOUND, nick),
		  mask);
	else
	    uSend(NickServ, nick, GetReply(ERR_ACCMASKNOTFOUND, nick),
		  mask);
    }
}

void NaddMaskCmd(UserList * nick, char **args, int numargs)
{

    char user[USERLEN], host[HOSTLEN], theirmask[USERLEN + HOSTLEN + 3];

    strcpy(user, nick->user);
    strcpy(host, nick->host);

    mask(user, host, 0, theirmask);

    addAccessMask(theirmask, nick);

    return;
}

void NmarkCmd(UserList * markedby, char **args, int numargs)
{
    int i, number = 0;
    char markreason[128];
    RegNickList *marktarget = NULL;

    if (!strcasecmp(args[1], "add")) {
	
	marktarget = getRegNickData(args[2]);

	if (!marktarget) {
	    uSend(NickServ, markedby, 
		  GetReply(ERR_NICKNOTREGGED, markedby), args[2]);
	    return;
	}

	bzero(markreason, 128);
	strncpy(markreason, args[3], 127);
	for (i = 4; i < numargs; i++) {
	    if ((strlen(markreason) + strlen(args[i])) > 126)
		break;
	    strcat(markreason, " ");
	    strcat(markreason, args[i]);
	}

	add_mark((time_t) 0, markedby->nick, marktarget, markreason);
	uSend(NickServ, markedby, GetReply(RPL_ADDREMARK, markedby), 
	      marktarget->nick);
	
	return;
	
    } else if (!strcasecmp(args[1], "del")) {

	marktarget = getRegNickData(args[2]);

	if (!marktarget) {
	    uSend(NickServ, markedby, 
		  GetReply(ERR_NICKNOTREGGED, markedby), args[2]);
	    return;
	}

	if (args[3][0] == '#')
	    number = atoi(&args[3][1]);
	else
	    number = atoi(&args[3][0]);

	if (!number) {
	    uSend(NickServ, markedby, GetReply(ERR_NOTAVALIDNUMBER, markedby), 
		  args[3]);
	    return;
	}

	del_mark(marktarget, number); 
	uSend(NickServ, markedby, GetReply(RPL_DELREMARK, markedby), 
	      marktarget->nick);
    } else {
	uSend(NickServ, markedby, GetReply(ERR_INCORRECTUSAGE, markedby),
	      "MARK");
	uSend(NickServ, markedby, GetReply(RPL_MOREHELPCMD, markedby),
	      NickServ, "MARK");
	return;
    }
}

void add_mark(time_t timestamp, char *by, RegNickList *target, char *reason)
{
    nMarkList *ml = NULL, *add = NULL;

    add = smalloc(sizeof(nMarkList));
    strcpy(add->markedby, by);
    strcpy(add->remark, reason);
    if (!timestamp)
	add->timestamp = CTime;
    else
	add->timestamp = timestamp;
    add->next = NULL;

    if (target->firstMark != NULL) {
	for (ml = target->firstMark; ml->next != NULL; ml = ml->next)
	    ; /* Find last marker in list */
	ml->next = add;
    } else
	target->firstMark = add;

    return;
}

void del_mark(RegNickList *target, int number)
{
    int i;
    nMarkList *ml, *tmp = NULL;

    tmp = ml = target->firstMark;
    for (i = 1; i < number && ml; i++) {
	tmp = ml;
	ml = ml->next;        /* Find remark to delete in list */
    }
    if (ml) {                 /* We have a list with remarks */
	if (tmp == ml)        /* First element in list */
	    if (ml->next)
		target->firstMark = ml->next;
	    else
		target->firstMark = NULL;
	else if (ml->next)    /* Element in middle of list */
	    tmp->next = ml->next;
	else
	    tmp->next = NULL; /* Last element in list */
	sfree(ml);
    }

    return;
}


void NaccessCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    if (numargs < 2) {
	uSend(NickServ, nick, GetReply(RPL_YOURACCESSIS, nick),
	      nick->caccess);
	return;
    }

    if (!strcasecmp(args[1], "list")) {
	nAccessList *tmp;
	if (numargs < 3) {
	    uSend(NickServ, nick, GetReply(RPL_ACCESSHEADER, nick), from);
	    for (tmp = nick->reg->firstAcc; tmp; tmp = tmp->next)
		uSend(NickServ, nick, GetReply(RPL_ACCESSLIST1, nick),
		      tmp->index, tmp->mask);
	    uSend(NickServ, nick, GetReply(RPL_ACCESSFOOTER, nick));
	} else {
	    RegNickList *tmpnick;
	    nAccessList *tmp;

	    tmpnick = getRegNickData(args[2]);

	    if (!tmpnick) {
		uSend(NickServ, nick, GetReply(ERR_NICKNOTREGGED, nick),
		      args[2]);
		return;
	    }

	    if (!checkAccess(nick->user, nick->host, tmpnick)
		&& !isOper(nick)) {
		uSend(NickServ, nick, GetReply(ERR_NOREMOTEACC, nick),
		      args[2]);
		return;
	    }

	    uSend(NickServ, nick, GetReply(RPL_ACCESSHEADER, nick),
		  tmpnick->nick);

	    for (tmp = tmpnick->firstAcc; tmp; tmp = tmp->next)
		uSend(NickServ, nick, GetReply(RPL_ACCESSLIST1, nick),
		      tmp->index, tmp->mask);
	    uSend(NickServ, nick, GetReply(RPL_ACCESSFOOTER, nick));
	}
    } else if (!strcasecmp(args[1], "add")) {
	if (numargs < 3) {
	    uSend(NickServ, nick, GetReply(ERR_BADPARAMS, nick));
	    uSend(NickServ, nick, GetReply(RPL_MOREHELPCMD, nick),
		  NickServ, "ACCESS");
	    return;
	}

	addAccessMask(args[2], nick);
	return;
    } else if (!strncasecmp(args[1], "del", 3)) {	/* also accepts delete */
	if (numargs < 3) {
	    uSend(NickServ, nick, GetReply(ERR_BADPARAMS, nick));
	    uSend(NickServ, nick, GetReply(RPL_MOREHELPCMD, nick),
		  NickServ, "ACCESS");
	    return;
	}

	delAccessMask(args[2], nick);
	return;
    } else {
	UserList *info;

	info = getNickData(args[1]);

	if (!info) {
	    uSend(NickServ, nick, GetReply(RPL_CURRENTACC, nick), args[1],
		  0);
	    return;
	}

	else
	    uSend(NickServ, nick, GetReply(RPL_CURRENTACC, nick), args[1],
		  info->caccess);
    }
}


void NghostCmd(UserList * nick, char **args, int numargs)
{
    RegNickList *tmp;
    char ghoststr[NICKLEN + USERLEN + HOSTLEN + 13]; // 13 = Message overhead
    int found_ghost = 0;

    if (!strcasecmp(nick->nick, args[1])) {
	uSend(NickServ, nick, GetReply(ERR_CANNOTGHOSTSELF, nick));
	return;
    }

    tmp = getRegNickData(args[1]);

    if (!tmp) {
	uSend(NickServ, nick, GetReply(ERR_NICKNOTREGGED, nick), args[1]);
	return;
    }

    found_ghost = isGhost(args[1]);

    if ((!found_ghost) && (!getNickData(args[1]))) {
	uSend(NickServ, nick, GetReply(ERR_NOTONLINE, nick), args[1]);
	return;
    }

    if (strcasecmp(tmp->password, args[2])) {
	uSend(NickServ, nick, GetReply(ERR_ACCESSDENIED, nick));
	if (strcmp(tmp->host, nick->host))
	    nick->remotefailid++;

	if (nick->remotefailid == 3) {
	    uSend(NickServ, nick, GetReply(RPL_EMAILHACK1, nick));
	    sSend
		(":%s GLOBOPS :\002Warning:\002 Possible ghost-ID hacking from %s!%s@%s (%s)",
		 NickServ, nick->nick, nick->user, nick->host, tmp->nick);
	} else if (nick->remotefailid == 6) {
	    sSend
		(":%s GLOBOPS :\002Warning:\002 Continued ghost-ID hacking from %s!%s@%s (%s)",
		 NickServ, nick->nick, nick->user, nick->host, tmp->nick);
	    killUser(services[1], nick, "Nickname hacking");
	}

	return;
    }

    uSend(NickServ, nick, GetReply(RPL_NICKGHOSTED, nick), tmp->nick);

    nick->remotefailid = 0;

    if (found_ghost) {
	delGhost(args[1]);
	return;
    }

    sprintf(ghoststr, "Ghost by: %s!%s@%s", nick->nick, nick->user,
	    maskuser(nick));
    killUser(services[1], getNickData(args[1]), ghoststr);
    remUser(args[1]);
}

void NsetflagsCmd(UserList * tmp, char **args, int numargs)
{
    char *from = tmp->nick;
    RegNickList *nick;
    int mode = 1, i;
    char plusstring[15];
    char minusstring[15];
    int posid;

    *plusstring = 0;
    *minusstring = 0;

    nick = getRegNickData(args[1]);
    if (!nick) {
	uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[1]);
	return;
    }

    for (i = 0; (unsigned int)i < strlen(args[2]); i++) {
	posid = 0;
	switch (args[2][i]) {
	case '+':
	    mode = 1;
	    break;
	case '-':
	    mode = 0;
	    break;
	case 'k':
	    posid = 1;
	    if (mode)
		nick->flags |= NKILL;
	    else
		nick->flags &= ~NKILL;
	    break;
	case 'v':
	    posid = 1;
	    if (mode) {
		nick->flags |= NVACATION;
		if (nick->flags & NHOLD)
		    strcat(args[2], "-h");
	    } else
		nick->flags &= ~NVACATION;
	    break;
	case 'e':
	    posid = 1;
	    if (mode)
		nick->flags |= NEMAIL;
	    else
		nick->flags &= ~NEMAIL;
	    break;
	case 'h':
	    posid = 1;
	    if (mode) {
		nick->flags |= NHOLD;
		if (nick->flags & NVACATION)
		    strcat(args[2], "-v");
	    } else
		nick->flags &= ~NHOLD;
	    break;
	case 'r':
	    posid = 1;
	    if (mode)
		nick->flags |= NREG;
	    else
		nick->flags &= ~NREG;
	    break;
	case 'a':
	    posid = 1;
	    if (mode) {
		nick->flags |= NACTIVE;
		if (nick->flags & NDEACC)
		    strcat(args[2], "-d");
	    } else
		nick->flags &= ~NACTIVE;
	    break;
	case 'd':
	    posid = 1;
	    if (mode)
		nick->flags |= NDEACC;
	    else
		nick->flags &= ~NDEACC;
	    break;
	case 's':
	    if (!isRoot(tmp))
		uSend(NickServ, tmp, GetReply(ERR_NEEDSRA, tmp));
	    else {
		posid = 1;
		if (mode)
		    nick->flags |= NSERVOP;
		else
		    nick->flags &= ~NSERVOP;
	    }
	    break;
	case 'l':
	    posid = 1;
	    if (mode)
	        nick->flags |= NOPER;
	    else
	        nick->flags &= ~NOPER;
	    break;
	case 'm':
	    posid = 1;
	    if (mode)
		nick->flags |= NMARKED;
	    else
		nick->flags &= ~NMARKED;
	    break;
	case 'w':
	    posid = 1;
	    if (mode)
		nick->flags2 |= NMSGMODE;
            else 
		nick->flags2 &= ~NMSGMODE;
	    break;
	default:
	    uSend(NickServ, tmp, GetReply(ERR_BADFLAG, tmp), args[2][i]);
	    break;
	}
	if (posid) {
	    char littlestring[2];
	    littlestring[0] = args[2][i];
	    littlestring[1] = 0;
	    if (mode)
		strcat(plusstring, littlestring);
	    else
		strcat(minusstring, littlestring);
	}
    }

    if ((*plusstring) || (*minusstring)) {
	uSend(NickServ, tmp, GetReply(RPL_CHANGEDFLAGS, tmp), nick->nick,
	      (*plusstring) ? "+" : "", plusstring,
	      (*minusstring) ? "-" : "", minusstring);
	sSend(":%s GLOBOPS :%s set mode %s%s%s%s for %s.", NickServ, from,
	      (*plusstring) ? "+" : "", plusstring,
	      (*minusstring) ? "-" : "", minusstring, nick->nick);
	logDump(operlog, "%s!%s@%s set flag(s): %s%s%s%s on %s", from,
		tmp->user, tmp->host, (*plusstring) ? "+" : "", plusstring,
		(*minusstring) ? "-" : "", minusstring, nick->nick);
    }
}

void createpasswd(char *passwd, int length)
{
	static char saltchars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
	int x;

	passwd[length] = '\0';

	for (x=0; x < length; x++)
	{
		passwd[x] = saltchars[random() % 64];
	}
	return;
}

void NbanishCmd(UserList * tmp, char **args, int numargs)
{
	char *from = tmp->nick;
	RegNickList *nick;

	nick = getRegNickData(args[1]);
	if (!nick)
	{
		nick = smalloc(sizeof(RegNickList));
		strcpy(nick->nick, args[1]);
		nick->flags |= NBANISH;
		nick->flags |= NACTIVE;
		nick->flags |= NREG;
		addRegNick(nick);
	}

    strcpy(nick->nick, args[1]);
    strcpy(nick->user, "banished");
    sprintf(nick->host, "nicks.%s", NETHOST);
    sprintf(nick->email, "nickserv@%s", NETHOST);
    sprintf(nick->url, "http://www.%s", NETHOST);
/*    strcpy(nick->password, "notrouble");*/
	createpasswd(nick->password,9);
    strcpy(nick->password2, "7654321");
    sprintf(nick->realname, "Banished by: %s", from);
    nick->amasks = 0;
    nick->timereg = 0;
    nick->timestamp = 0;
    nick->flags = 0;
    nick->flags |= NBANISH;
    nick->flags |= NACTIVE;
    nick->flags |= NREG;
    nick->firstAcc = nick->lastAcc = NULL;
    killUser(services[1], getNickData(args[1]), NSBANISHMSG);
    uSend(NickServ, tmp, GetReply(RPL_BANNEDNICK, tmp), args[1]);
    sSend(":%s GLOBOPS :%s has permanently banned the nickname \002%s\002",
	  NickServ, from, args[1]);
    logDump(operlog, "%s!%s@%s BANISHed %s", from, tmp->user, tmp->host,
	    args[1]);
}

void NgetkeyCmd(UserList * tmp, char **args, int numargs)
{
    char *from = tmp->nick;
    RegNickList *nick;

    nick = getRegNickData(args[1]);

    if (!nick || !(nick->flags & NREG))
	uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[1]);
    else {
	uSend(NickServ, tmp, GetReply(RPL_GETKEY, tmp), nick->nick,
	      nick->password2);
	sSend(":%s GLOBOPS :%s used getkey on %s", NickServ, from,
	      args[1]);
    }
    return;
}

void NgetrealpassCmd (UserList * tmp, char **args, int numargs)
{
    char *from = tmp->nick;
    RegNickList *nick;
    unsigned int flags;

    nick = getRegNickData(args[1]);

    if (!nick) uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[1]);
    else
    {
    	flags=0;
    	if (tmp->reg) flags=tmp->reg->flags;
			uSend (NickServ, tmp, GetReply(RPL_GETREALPASS, tmp), nick->nick, nick->password);
			if (!(flags&(NROOT)))
			{
				sSend (":%s GLOBOPS :%s used getrealpass on \002%s\002", NickServ, from, nick->nick);
				logDump (operlog, "%s!%s@%s used GETREALPASS on %s", from, tmp->user, tmp->host, nick->nick);
			}
    }
}

void NgetpassCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;
    RegNickList *nick;
    char sendfrom[80];
    time_t doot = time(NULL);

    nick = getRegNickData(args[1]);

    if (!nick)
	uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[1]);
    else {
	sSend(":%s GLOBOPS :%s used getpass on \002%s\002", NickServ, from,
	      nick->nick);
	uSend(NickServ, tmp, GetReply(RPL_GETPASS, tmp), nick->nick,
	      nick->email);
	sprintf(sendfrom, "%s Password Recovery <%s@%s>", NickServ,
		NickServ, NETHOST);
	sendMail(nick->email, sendfrom, "Password",
		 "\nThe password for %s is %s\n"
		 "The activation code for %s is %s\n"
		 "\nThis password recovered by %s at %s", nick->nick,
		 nick->password, nick->nick, nick->password2, tmp->nick,
		 ctime(&(doot)));
	logDump(operlog, "%s!%s@%s GETPASS %s", from, tmp->user, tmp->host,
		args[1]);
    }
}

void NdeleteCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;
    RegNickList *nick;

    nick = getRegNickData(args[1]);

    if (!nick) {
	uSend(NickServ, tmp, GetReply(ERR_NICKNOTREGGED, tmp), args[1]);
	return;
    }

    if ((nick->flags & NACTIVE) && !isServop(tmp)) {
	uSend(NickServ, tmp, GetReply(ERR_NEEDSERVOP, tmp));
	return;
    }

    if (nick) {
	uSend(NickServ, tmp, GetReply(RPL_DELETE, tmp), nick->nick);
	sSend
	    (":%s GLOBOPS :%s deleted the nickname \002%s\002 from the databases.",
	     NickServ, from, nick->nick);
	delRegNick(nick);
	logDump(nicklog, "%s DELETE %s", from, args[1]);
    }
}

void NsaveCmd(UserList * tmp, char **args, int numargs)
{
    char *from = tmp->nick;

    uSend(NickServ, tmp, GetReply(RPL_SAVE, tmp));
    sSend
	(":%s GLOBOPS :Synchronising \002NickServ database\002 to disk at the request of \002%s\002.",
	 NickServ, from);

    saveNickData(firstRegNick);
}

void NwebtvcountCmd(UserList * tmp, char **args, int numargs)
{
    UserList *u;
    int count = 0;

    for (u = firstUser; u; u = u->next) {
	if (u->webbie) {
	    sSend(":%s NOTICE %s :%s", NickServ, tmp->nick, u->nick);
	    count++;
	}
    }

    sSend(":%s NOTICE %s :%d WebTV user(s) online.",
	  NickServ, tmp->nick, count);

}

void NlistCmd(UserList * tmp, char **args, int numargs)
{
    char nick[NICKLEN], user[USERLEN], host[HOSTLEN];
    char searchargs[NICKLEN + USERLEN + HOSTLEN + 3];
    RegNickList *reg;
    UserList *online;
    int i, x;

    i = x = 0;

    if (!strcasecmp(args[1], "-email")) {
	char *s1;
	int goodmask = 0;

	if (numargs < 3) {
	    uSend(NickServ, tmp, GetReply(ERR_BADPARAMS, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ,
		  "LIST");
	    return;
	}

	for (s1 = args[2]; *s1; s1++) {
	    if ((*s1 != '@') && (*s1 != '.') && (*s1 != '*')
		&& (*s1 != '?')) goodmask = 1;
	}

	if (!goodmask) {
	    uSend(NickServ, tmp, GetReply(ERR_VALIDEMAIL, tmp));
	    return;
	}

	strcpy(searchargs, args[2]);

	uSend(NickServ, tmp, GetReply(RPL_NLISTHEADER, tmp), searchargs);

	for (reg = firstRegNick; reg; reg = reg->next) {
	    if (!match(reg->email, searchargs)) {
		i++;
		uSend(NickServ, tmp, GetReply(RPL_NLIST1, tmp), i,
		      reg->nick, reg->email);
	    }
	    x++;
	}
    } else if (!strcasecmp(args[1], "-flags")) {
	int showthis;
	char showoper[32];

	if (numargs < 3) {
	    uSend(NickServ, tmp, GetReply(ERR_BADPARAMS, tmp));
	    uSend(NickServ, tmp, GetReply(RPL_MOREHELPCMD, tmp), NickServ,
		  "LIST");
	    return;
	}

	showoper[0] = '\0';
	strcpy(searchargs, args[2]);

	uSend(NickServ, tmp, GetReply(RPL_NLISTHEADER, tmp), searchargs);

	for (reg = firstRegNick; reg; reg = reg->next) {
	    showthis = 0;
	    switch (args[2][0]) {
	    case 'b':
		if (reg->flags & NBANISH)
		    showthis = 1;
		break;
	    case 'h':
		if (reg->flags & NHOLD)
		    showthis = 1;
		break;
	    case 'v':
		if (reg->flags & NVACATION)
		    showthis = 1;
		break;
	    case 's':
		if (reg->flags & NNETADM) {
		    showthis = 1;
		    strcpy(showoper, " [Network Administrator]");
		} else if (reg->flags & NROOT) {
		    showthis = 1;
		    strcpy(showoper, " [Services Root Administrator]");
		} else if (reg->flags & NSERVOP) {
		    showthis = 1;
		    strcpy(showoper, " [Services Operator]");
		}
		break;
	    case 'l':
		if (reg->flags & NOPER)
		    showthis = 1;
		break;
	    case 'm':
		if (reg->flags & NMARKED)
		    showthis = 1;
		break;
	    case 'w':
		if (reg->flags2 & NMSGMODE)
		    showthis = 1;
		break;
	    }
	    if (showthis) {
		i++;
		uSend(NickServ, tmp, GetReply(RPL_NLIST2, tmp), i,
		      reg->nick, reg->user, reg->host, showoper);
	    }
	    x++;
	}
    } else {
	int matchflag;

	if (index(args[1], '!') && index(args[1], '@'))
	    sscanf(args[1], "%[^!]!%[^@]@%s", nick, user, host);
	else {
	    uSend(NickServ, tmp, GetReply(ERR_PROPERHOST1, tmp));
	    return;
	}

	if (!strcmp(nick, "*") && !strcmp(user, "*") && !strcmp(host, "*")
	    && !isServop(tmp)) {
	    uSend(NickServ, tmp, GetReply(ERR_WIDEHOST, tmp));
	    return;
	}

	strcpy(searchargs, args[1]);

	uSend(NickServ, tmp, GetReply(RPL_NLISTHEADER, tmp), searchargs);

	for (reg = firstRegNick; reg; reg = reg->next) {

	    matchflag = 0;
	    if ((reg->flags2 & NSTICKYOP) && (!isOper(tmp))) {
		if (!match(reg->nick, nick) && !match(reg->user, user)
		    && !match("IRCop.Serenity-IRC.Net", host))
		    matchflag = 1;
	    } else {
		online = getNickData(reg->nick);
		if (!isOper(tmp) && (online)
		    && (online->oflags & NISPLUSX)) {
		    if (!match(reg->nick, nick) && !match(reg->user, user)
			&& !match(return_user_mask(reg->host), host))
			matchflag = 1;
		} else {
		    if (!match(reg->nick, nick) && !match(reg->user, user)
			&& !match(reg->host, host))
			matchflag = 1;
		}
	    }

	    if (matchflag) {
		i++;
		uSend(NickServ, tmp, GetReply(RPL_NLIST3, tmp), i,
		      reg->nick, reg->user, reg->host);
	    }
	    x++;
	}
    }

    if (i == 0)
	uSend(NickServ, tmp, GetReply(ERR_NLISTEMPTY, tmp));

    uSend(NickServ, tmp, GetReply(RPL_NLISTFOOTER, tmp), i, x, searchargs);
}

/*
 * ===DOC===
 * This function will basically bother a nick owner into giving a nick back
 */
void annoyNickThief(UserList * nick)
{
    char tmp[NICKLEN+1];
    int assume_webbie = 0;

    /* If someone is using a WebTV user nick but is not set as +w we'd
     * still like them to be assumed a WebTv user and thus given a chance to 
     * identify to their nick. In this case we use message instead of notice 
     * (default). We set them as webbie for a short moment, then remove that
     * mode after sending our message out.  -- Remmy 20020616 */
    if ((nick->reg->flags2 & NMSGMODE) && !nick->webbie)
        nick->webbie = assume_webbie = 1;
    
    uSend(NickServ, nick, GetReply(RPL_NICKWARN1, nick));
    if ((nick->reg->flags & NKILL) && !isServop(nick)) {
	uSend(NickServ, nick, GetReply(RPL_NICKWARN3, nick));
	uSend(NickServ, nick, GetReply(RPL_NICKWARN2, nick));
	sprintf(tmp, "6%s", nick->nick);
	timer(15, killide, "killide", tmp);	/* killide func, 60 secs */
    }

    /* Reset 'assumed webbie' */
    if (assume_webbie)
        nick->webbie = 0;
}

/* general ghost like stuff */
void killide(char *info)
{
	char tmp[NICKLEN+1];
	char newnick[16];
	UserList *check;
	char killcount;

	killcount = info[0];
	info++;

	check = getNickData(info);

	if (!check)			/* THEY'VE ALREADY GONE */
		return;

	if (check->caccess > 1)	/* AHA!  THEY'VE ID'D... NEVER MIND. */
		return;

	switch (killcount) {
		case '6':
			uSend(NickServ, check, GetReply(RPL_NICKWARN4, check));
			sprintf(tmp, "3%s", info);
			timer(15, killide, "killide", tmp);	/* 15 MORE SECONDS */
			break;
		case '3':
			sSend(":%s 433 %s %s :Belongs to another user", myname, info, info);
			uSend(NickServ, check, GetReply(RPL_NICKWARN5, check));
			sprintf(newnick, "Guest%d", (((unsigned int) rand()) % 9998) + 1);
			sSend(":%s SVSNICK %s %s 0", myname, info, newnick);
			break;
	}
}

void addGhost(char *ghost)
{

    GhostList *newghost = smalloc(sizeof(GhostList));
    char host[HOSTLEN];

    strcpy(newghost->ghost, ghost);

    if (!firstGhost) {
	firstGhost = newghost;
	newghost->previous = NULL;
    } else {
	lastGhost->next = newghost;
	newghost->previous = lastGhost;
    }

    lastGhost = newghost;
    newghost->next = NULL;

    sprintf(host, "nicks.%s", NETHOST);
    addUser(ghost, "owned", host, "Nick Enforcement", "i");
    totalusers++;
    totalghosts++;
}

void delGhost(char *ghost)
{
    GhostList *killme;

    totalusers--;
    killme = getGhost(ghost);
    if (!killme)
	return;
    sSend(":%s QUIT :Ahhh, they finally let me off work!", ghost);

#ifdef DEBUG2
    sSend(":%s GNOTICE :Removed ghost %s", myname, ghost);
#endif

    if (killme->previous)
	killme->previous->next = killme->next;
    else
	firstGhost = killme->next;
    if (killme->next)
	killme->next->previous = killme->previous;
    else
	lastGhost = killme->previous;

    sfree(killme);
    totalghosts--;

    return;
}

GhostList *getGhost(char *ghost)
{
    GhostList *tmp;
    for (tmp = firstGhost; tmp; tmp = tmp->next) {
	if (!strcasecmp(tmp->ghost, ghost))
	    return tmp;
    }
    return NULL;
}

int isGhost(char *ghost)
{
    if (getGhost(ghost))
	return 1;
    else
	return 0;
}

int countRegAddy(char *email)
{
    RegNickList *tmp;
    int count = 0;

    for (tmp = firstRegNick; tmp; tmp = tmp->next) {
	if (!strcasecmp(tmp->email, email))
	    count++;
    }
    return count;
}

void expireNicks(char *placeholder)
{
    time_t timestart, timeend;
    RegNickList *tmp, *next;
#ifdef BACKUPDATABASE
    char backup[30];
    char cmd[500];
#endif
    int i = 0;
    int zapped;

    MemoBox *memobox;
    MemoList *memo, *nextmemo;
    long memocnt = 0;

    mostnicks = 0;
    timestart = time(NULL);
#ifdef BACKUPDATABASE
    strftime(backup, 30, "nickserv.db%d%m%Y", localtime(&timestart));
    sprintf(cmd, "cp %s/nickserv.db %s/%s",
	    DBDIR, DBDIR, backup);
    system(cmd);
#endif
    for (tmp = firstRegNick; tmp; tmp = next) {
	next = tmp->next;
	zapped = 0;

	if (tmp->timestamp > timestart)	/* HOLD UP THERE -- NEGATIVE TIME? */
	    tmp->timestamp = timestart;	/* FIX THAT RIGHT NOW */

	if (!(tmp->flags & NREG) || tmp->flags & NHOLD
	    || tmp->flags & NBANISH) {
	} else if ((tmp->flags & NVACATION)
		   && ((timestart - tmp->timestamp) >= (6 * NICKDROPTIME)))
	    zapped++;
	else if (!(tmp->flags & NACTIVE)
		 && ((timestart - tmp->timestamp) >= (24 * 3600)))
	    zapped++;
	else if (((timestart - tmp->timestamp) >= NICKDROPTIME) & !(tmp->flags & NVACATION))
	    zapped++;

	if (zapped) {
	    logDump(nicklog, "%s EXPIRE", tmp->nick);
	    delRegNick(tmp);
	    i++;
	} else {
	    mostnicks++;

	    memobox = tmp->memos;


	    if (memobox) {
		for (memo = memobox->firstMemo; memo; memo = nextmemo) {
		    nextmemo = memo->next;
		    if ((timestart - memo->sent) >= MEMODROPTIME)
			delMemo(memobox, memo);
		    else
			memocnt++;
		}
	    }
	}
    }

    mostmemos = memocnt;
    timeend = time(NULL);
#ifdef SHOWSYNCH
    sSend(":%s GLOBOPS :NickServ EXPIRE(%i/%lu) %lu seconds", NickServ, i,
	  mostnicks, (timeend - timestart));
#endif
    timer(2 * 3600, expireNicks, "expireNicks", "void");
}

void indexAccItems(RegNickList * nick);

int addAccItem(RegNickList * nick, char *mask)
{
    nAccessList *tmp;

    if (!nick)
	return 0;
    if (!strcmp(mask, "*@*") || !strcmp(mask, "*@*.*") ||
	!strcmp(mask, "*@*.*.*") || !strcmp(mask, "*@*.*.*.*") ||
	!strcmp(mask, "*@*.com") || !strcmp(mask, "*@*.org") ||
	!strcmp(mask, "*@*.net") || !strcmp(mask, "*@*.edu") ||
	index(mask, '!') || !index(mask, '@'))
	return 0;
    for (tmp = nick->firstAcc; tmp; tmp = tmp->next) {
	if (!strcasecmp(mask, tmp->mask))
	    return 0;
    }

    tmp = smalloc(sizeof(nAccessList));
    strcpy(tmp->mask, mask);

    if (!nick->firstAcc) {
	nick->firstAcc = tmp;
	tmp->previous = NULL;
    } else {
	nick->lastAcc->next = tmp;
	tmp->previous = nick->lastAcc;
    }

    nick->lastAcc = tmp;
    tmp->next = NULL;
    nick->amasks++;
    indexAccItems(nick);
    return 1;
}

int delAccItem(RegNickList * nick, char *mask, char *deletedmask)
{
    nAccessList *tmp;
    int found = 0;
    int x = 0;
    if (!nick)
	return 0;

    x = atoi(mask);

    for (tmp = nick->firstAcc; tmp; tmp = tmp->next) {
	if (!strcasecmp(mask, tmp->mask) || (x && tmp->index == x)) {
	    found = 1;
	    break;
	}
    }
    if (found == 0) {
	return 0;
    }
    strcpy(deletedmask, tmp->mask);
    if (tmp->previous)
	tmp->previous->next = tmp->next;
    else
	nick->firstAcc = tmp->next;
    if (tmp->next)
	tmp->next->previous = tmp->previous;
    else
	nick->lastAcc = tmp->previous;

    sfree(tmp);
    indexAccItems(nick);
    return 1;
}

void indexAccItems(RegNickList * nick)
{
    nAccessList *tmp;
    int i = 1;
    for (tmp = nick->firstAcc; tmp; tmp = tmp->next) {
	tmp->index = i;
	i++;
    }
}

void syncNickData(time_t next)
{
    nextNsync = next;
    saveNickData(firstRegNick);
}

void syncMemoData(time_t next)
{
    nextMsync = next;
    saveMemoData(firstRegNick);
}

/* LISTOPS is now moved here - also redone so it sorts by order and shows away -GZ */

void NlistopsCmd(UserList * nick, char **args, int numargs)
{
    int i = 0;
    int ucnt = 0;
    UserList *tmp;

    uSend(NickServ, nick, GetReply(RPL_LISTOPHEADER, nick));

	for(tmp = firstUser; tmp; tmp = tmp->next)
	{
		ucnt++;
		if(isOper(tmp) && isRoot(tmp))
		{
			i++;
			uSend(NickServ, nick, GetReply(RPL_LISTOP1, nick), i,
				tmp->nick, "[Services Root Administrator]",
				(tmp->oflags & NISAWAY) ? " (AWAY)" : "");
		}
	}

	for (tmp = firstUser; tmp; tmp = tmp->next)
	{
		if (isOper(tmp) && isServop(tmp) && !isRoot(tmp))
		{
			i++;
			uSend(NickServ, nick, GetReply(RPL_LISTOP1, nick), i,
				tmp->nick, "[Services Operator]",
				(tmp->oflags & NISAWAY) ? " (AWAY)" : "");
		}
	}

	for (tmp = firstUser; tmp; tmp = tmp->next)
	{
		if (isOper(tmp) && (tmp->oflags & NISADMIN) && !isServop(tmp))
		{
			i++;
			uSend(NickServ, nick, GetReply(RPL_LISTOP1, nick), i,
				tmp->nick, "[Server Administrator]",
				(tmp->oflags & NISAWAY) ? " (AWAY)" : "");
		}
	}

	for (tmp = firstUser; tmp; tmp = tmp->next)
	{
		if (isOper(tmp) && !isServop(tmp) && !(tmp->oflags & NISADMIN) &&
			!(strstr(tmp->nick, "Opm")))
		{
			i++;
			uSend(NickServ, nick, GetReply(RPL_LISTOP1, nick), i,
				tmp->nick, "[IRC Operator]",
				(tmp->oflags & NISAWAY) ? " (AWAY)" : "");
		}
	}

    uSend(NickServ, nick, GetReply(RPL_LISTOPFOOTER, nick));
}

void NresendCmd (UserList * nick, char **args, int numargs)
{
		char *py,*email;
		int err;
		long int doot;
		RegNickList *tmp;
		char sendfrom[512];
		
		py=nick->nick;
		tmp = getRegNickData (args[1]);
		if (!tmp)
		{
			sSend (":%s NOTICE %s :Nick %s is not registered.",NickServ,py,args[1]);
			return;
		}
		email=tmp->email;
		time (&doot);

    sprintf (sendfrom, "Nickname Services <%s@%s>", NickServ, NETHOST);
    err = sendMail (email, sendfrom, "Nickname activation key",
		   "Hello %s, Welcome to the Serenity-IRC network!\n\n"
		   "Your nickname registration has been verified.\n\n"
		   "To activate your nickname:\n\n"
		   "1) switch to the nickname %s\n"
		   "2) /msg %s ACTIVATE %s\n\n"
		   "%s is your activation key.  %s is your password.\n\n"
		   "Please try to remember your password, you might need it\n"
		   "some day.\n\n"
		   "If you accidently lose your password, you can always\n"
		   "/join #help and ask one of our IRC Operators to recover\n"
		   "it for you. We will be more than glad to help you!\n\n"
		   "If you are new to IRC or Serenity-IRC, we would like to recommend\n"
		   "#help, as it is the place where you can obtain help or any\n"
		   "other kind of information that is IRC related.\n\n"
		   "Enjoy your stay on our safe and friendly network!\n\n"
		   "For additional help, try these commands:\n"
		   "     /msg %s HELP\n"
		   "     /msg %s HELP\n"
		   "     /msg %s HELP\n",
		   tmp->nick, tmp->nick, NickServ, tmp->password2,
		   tmp->password2, tmp->password,
		   NickServ, ChanServ, MemoServ
		   );

    if (!err)
    {
			logDump (nicklog, "%s!%s@%s: ACTIVATION key sent to '%s'\n", tmp->nick, tmp->user, tmp->host, args[1]);
			sSend (":%s NOTICE %s :*** ACTIVATION key sent to '%s'\n", NickServ, py, args[1]);
			sSend (":%s GLOBOPS :%s resent activation key to for nick %s",NickServ,py,args[1]);
    }
    else
    {
			sSend (":%s NOTICE %s :Error: No email sent to %s",NickServ,py,args[1]);
    }
}

























