/* 542 */

#include "services.h"

int HJCheck(char *);

long getkilltime(char *s)
{
/* THIS FUNCTION RETURNS A TIME IN SECONDS BASED ON INPUT STRING
   FORMATS ALLOWED:
		<value>h = hours
		<value>d = days
		<value>w = weeks
   MISSING LETTER IS ASSUMED TO MEAN MINUTES (BACKWARDS COMPATIBLE)
   MISSING VALUE WILL BE ASSUMED AS 1.

   THIS FUNCTION IS USED BY Oautokill and Oignore FOR TIME COMPUTATION */

    long base_mult;
    char mult_char[2];
    int value_width;
    int base_value;

    base_mult = 60L;		/* INITIALIZE TO CONVERT MINS TO SECS */
    strcpy(mult_char, "m");
    base_value = 1;

    if ((!strcasecmp(s, "add")) || (!(*s)))	/* COMPATIBILITY FOR OLD FUNCTION */
	return 0;

    value_width = strlen(s) - 1;

    if (isalpha(s[value_width])) {	/* IF THERE'S A LETTER ON THE END, PARSE IT */
	mult_char[0] = s[value_width];

	if (!strcasecmp(mult_char, "h"))
	    base_mult = 3600L;
	else if (!strcasecmp(mult_char, "d"))
	    base_mult = 86400L;
	else if (!strcasecmp(mult_char, "w"))
	    base_mult = 604800L;

	if (value_width > 0) {
	    base_value = atoi(s);
	    if (base_value < 1)
		base_value = 1;
	}
    } else {
	base_value = atoi(s);
	if (base_value < 0)
	    base_value = 0;
    }

    return ((long) (base_value) * base_mult);
}

char *expand_time(long t)
{
/* EXPANDS TIME IN SECONDS TO FULLY QUALIFIED STRING 
   USED BY AUTOKILL FUNCTIONS FOR DISPLAY */

    static char retstr[128];
    int temp1;

    *retstr = 0;

    if (t == 0) {
	strcpy(retstr, "forever");
    } else {
	if (t >= 604800L) {
	    temp1 = (int) (t / 604800L);
	    t = t % 604800L;
	    sprintf(retstr, "%d week%s", temp1, (temp1 == 1 ? "" : "s"));
	}
	if (t >= 86400L) {
	    temp1 = (int) (t / 86400L);
	    t = t % 86400L;
	    if (*retstr)
		strcat(retstr, ", ");
	    sprintf(retstr, "%s%d day%s", retstr, temp1,
		    (temp1 == 1 ? "" : "s"));
	}
	if (t >= 3600L) {
	    temp1 = (int) (t / 3600L);
	    t = t % 3600L;
	    if (*retstr)
		strcat(retstr, ", ");
	    sprintf(retstr, "%s%d hour%s", retstr, temp1,
		    (temp1 == 1 ? "" : "s"));
	}

	temp1 = (int) (t / 60L);

	if ((*retstr) && (temp1 > 0))
	    strcat(retstr, ", ");

	if ((!(*retstr)) && (temp1 == 0))
	    temp1 = 1;

	if (temp1)
	    sprintf(retstr, "%s%d minute%s", retstr, temp1,
		    (temp1 == 1 ? "" : "s"));
    }

    return ((char *) retstr);
}

void sendToOperServ(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    if (!strcmp(args[0], "\001PING")) {
	if (numargs < 3)
	    uSend(OperServ, nick, GetReply(RPL_PING1, nick), args[1]);
	else
	    uSend(OperServ, nick, GetReply(RPL_PING2, nick), args[1],
		  args[2]);
	return;
    }
    if (!isOper(nick)) {
	uSend(OperServ, nick, GetReply(ERR_NEEDOPER, nick));
	return;
    }

    if (!strcasecmp(args[0], "help") && numargs > 1) {
	sendHelp(from, OperServ, args, numargs);
	return;
    } else if (!strcasecmp(args[0], "help")) {
	sendHelp(from, OperServ, NULL, 0);
	return;
    }

    msgtabCheck(OperServ, nick, args, numargs);
}

void OrollupCmd(UserList * nick, char **args, int numargs)
{
    int akills_placed = 0;
    ChanList *chan;
    cNickList *inspectnick, *nextnick = NULL;
    char mask[NICKLEN + USERLEN + HOSTLEN + 2];

    chan = getChanData (args[1]);
    if (!chan)
    {
        uSend (OperServ, nick, GetReply (ERR_NOSUCHCHAN, nick), args[1]);
        return;
    }

    for (inspectnick = chan->firstUser; inspectnick; inspectnick = nextnick)
    {
        nextnick = inspectnick->next;

        if (!(inspectnick->person->reg) || 
            !(inspectnick->person->reg->flags & NACTIVE))
        { /* Only set an akill if the person is not a registered nick,
             we should also look at ID status, but won't because we
             don't want to hurt too many 'innocents' */
            if (!isAKilled(inspectnick->person->nick, inspectnick->person->user,
                           inspectnick->person->host)) {
                strcpy(mask, "*!*@");
                strcat(mask, inspectnick->person->host);
                add_access_mod((3 * 3600), mask, OperServ, A_AKILL, 
                               "Excessive abuse");
                akills_placed++;
            }
        }
    }
    sSend(":%s GLOBOPS :%s rolled up %s. (%d akills added)", OperServ, 
            nick->nick, args[1], akills_placed);
    logDump(operlog, "%s!%s@%s rolled %s up.", nick->nick, nick->user, 
            nick->host, args[1]);
}

void OaccessmodCmd(UserList * nick, char **args, int numargs)
{
    char type;
    char *from = nick->nick;
    long t;
    int i;
    char amreason[80];
    char amtype[9];

    if (!match(args[0], "a*kill")) {
	strcpy(amtype, "Autokill");
    	type = A_AKILL;
    } else if (!strcasecmp(args[0], "ignore")) {
	strcpy(amtype, "Ignore");
	type = A_IGNORE;
    } else if (!strcasecmp(args[0], "exempt")) {
	strcpy(amtype, "Exempt");
    	type = A_EXEMPT;
    } else if (!strcasecmp(args[0], "hostjinx")) {
	strcpy(amtype, "Hostjinx");
    	type = A_HOSTJINX;
    } else
	return;
    
    if (numargs < 2) {
	list_access_mod(from, type, NULL);
	return;
    }
    if (!strcasecmp(args[1], "list")) {
	if (numargs < 3)
	    list_access_mod(from, type, NULL);
	else
	    list_access_mod(from, type, args[2]);
	return;
    }

    if (!isServop(nick) && !isRoot(nick)) {
	uSend(OperServ, nick, GetReply(ERR_NEEDSERVOP, nick));
	return;
    }

    if (numargs < 3) {
	uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
	uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ,
	      amtype);
	return;
    }

    if (args[1][0] == '-' || !strcasecmp(args[1], "del")
	|| !strcasecmp(args[1], "delete")) {
        if (args[2][0] != '#')
	    if (!check_mask(args[2], 0)) {
	        uSend(OperServ, nick, GetReply(ERR_BADMASK, nick), args[2]);
	        uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick),
		      OperServ, amtype);
	        return;
	    }
	remove_access_mod(from, type, args[2]);
	return;
    }
    if (numargs < 4) {
	uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
	uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ,
	      amtype);
	return;
    } else {
	t = getkilltime(args[1]);
	if (!check_mask(args[2], 0)) {
	    uSend(OperServ, nick, GetReply(ERR_BADMASK, nick), args[2]);
	    return;
	}

	if (find_access_mod(type, args[2])) {
	    uSend(OperServ, nick, GetReply(ERR_MASKEXISTS, nick), args[2],
		  amtype);
	    return;
	}

	bzero(amreason, 80);
	strncpy(amreason, args[3], 79);
	for (i = 4; i < numargs; i++) {
	    if ((strlen(amreason) + strlen(args[i])) > 78)
		break;
	    strcat(amreason, " ");
	    strcat(amreason, args[i]);
	}

	uSend(OperServ, nick, GetReply(RPL_AKILLADDED, nick), amtype,
	      args[2], expand_time(t), amreason);
	logDump(operlog, "%s!%s@%s added %s for %s for %s [%s]", from,
		nick->user, nick->host, amtype, args[2], expand_time(t),
		amreason);
	add_access_mod(t, args[2], from, type, amreason);
	return;
    }
}

void OmodeCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    int on = 1, onarg = 3, i = 0, valid = 1;
    ChanList *tmp = getChanData(args[1]);
    cBanList *tmpban;
    cNickList *tmpnick;
    char string[512];

    if (!tmp) {
	uSend(OperServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
	return;
    }

    if ((args[2][0] != '+') && (args[2][0] != '-')) {
	uSend(OperServ, nick, GetReply(ERR_INVALIDMODES, nick), args[2]);
	return;
    }

    for (i = 0; args[2][i]; i++) {
	if (!valid)
	    break;

	switch (args[2][i]) {
	case '+':
	    on = 1;
	    break;
	case '-':
	    on = 0;
	    break;
	case 'b':
	    if (numargs > onarg) {
		if (on) {
		    tmpban = smalloc(sizeof(cBanList));
		    strcpy(tmpban->ban, args[onarg]);
		    addChanBan(tmp, tmpban);
#ifdef CDEBUG
		    sSend(":%s GNOTICE :Added ban %s on %s", myname,
			  args[onarg], tmp->name);
#endif
		} else {
		    tmpban = getChanBan(tmp, args[onarg]);
		    if (tmpban) {
#ifdef CDEBUG
			sSend(":%s GNOTICE :Removed ban %s on %s", myname,
			      args[onarg], tmp->name);
#endif
			delChanBan(tmp, tmpban);
		    }
		}
		onarg++;
	    } else
		valid = 0;
	    break;

	case 'i':
	    if (on) {
		tmp->modes |= PM_I;
#ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s +i", myname, tmp->name);
#endif
	    } else {
		tmp->modes -= PM_I;
#ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -i", myname, tmp->name);
#endif
	    }
	    break;

	case 'l':
	    if (on) {
		if (numargs > onarg) {
		    int chan_limit = atoi(args[onarg]);
		    if (chan_limit > 0) {
			tmp->modes |= PM_L;
			if (tmp->reg)
			    tmp->reg->limit = chan_limit;
#ifdef CDEBUG
			sSend(":%s GNOTICE :Set %s +l %lu", myname,
			      tmp->name, tmp->reg->limit);
#endif
		    } else
			valid = 0;
		    onarg++;
		}
	    } else {
		tmp->modes -= PM_L;
		if (tmp->reg)
		    tmp->reg->limit = 0;
#ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -l", myname, tmp->name,
		      tmp->reg->limit);
#endif
	    }
	    break;

	case 'k':
	    if (numargs > onarg) {
		if (on) {
		    tmp->modes |= PM_K;
		    if (tmp->reg)
			strcpy(tmp->reg->key, args[onarg]);
# ifdef CDEBUG
		    sSend(":%s GNOTICE :Set %s +k %s", myname, tmp->name,
			  tmp->reg->key);
# endif
		} else {
		    tmp->modes -= PM_K;
		    if (tmp->reg)
			tmp->reg->key[0] = '\0';
# ifdef CDEBUG
		    sSend(":%s GNOTICE :Set %s -k %s", myname,
			  tmp->reg->name, args[onarg]);
# endif
		}
		onarg++;
	    } else
		valid = 0;
	    break;

	case 'm':
	    if (on) {
		tmp->modes |= PM_M;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s +m", myname, tmp->name);
# endif
	    } else {
		tmp->modes -= PM_M;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -m", myname, tmp->name);
# endif
	    }
	    break;

	case 'n':
	    if (on) {
		tmp->modes |= PM_N;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s +n", myname, tmp->name);
# endif
	    } else {
		tmp->modes -= PM_N;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -m", myname, tmp->name);
# endif
	    }
	    break;

	case 'o':
	    if (numargs > onarg) {
		if (on) {
		    tmpnick =
			getChanUserData(tmp, getNickData(args[onarg]));
		    if (tmpnick) {
			tmpnick->op |= CHANOP;
# ifdef CDEBUG
			sSend(":%s GNOTICE :Oped %s in %s", myname,
			      args[onarg], tmp->name);
# endif
		    } else
			valid = 0;
		} else {
		    tmpnick =
			getChanUserData(tmp, getNickData(args[onarg]));
		    if (tmpnick) {
			tmpnick->op &= ~CHANOP;
# ifdef CDEBUG
			sSend(":%s GNOTICE :DeOped %s in %s", myname,
			      args[onarg], tmp->name);
# endif
		    } else
			valid = 0;
		}
		onarg++;
	    } else
		valid = 0;
	    break;

	case 'p':
	    if (on) {
		tmp->modes |= PM_P;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s +p", myname, tmp->name);
# endif
	    } else {
		tmp->modes -= PM_P;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -p", myname, tmp->name);
# endif
	    }
	    break;

	case 's':
	    if (on) {
		if ((tmp->reg) && isServop(nick))
		    tmp->reg->flags |= CPRIVATE;
		tmp->modes |= PM_S;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s +s", myname, tmp->name);
# endif
	    } else {
		if ((tmp->reg) && isServop(nick))
		    tmp->reg->flags &= ~CPRIVATE;

		tmp->modes -= PM_S;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -s", myname, tmp->name);
# endif
	    }
	    break;

	case 't':
	    if (on) {
		tmp->modes |= PM_T;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s +t", myname, tmp->name);
# endif
	    } else {
		tmp->modes -= PM_T;
# ifdef CDEBUG
		sSend(":%s GNOTICE :Set %s -t", myname, tmp->name);
# endif
	    }
	    break;

	case 'v':
	    if (numargs > onarg) {
		if (on) {
		    tmpnick =
			getChanUserData(tmp, getNickData(args[onarg]));
		    if (tmpnick) {
			tmpnick->op |= CHANVOICE;
# ifdef CDEBUG
			sSend(":%s GNOTICE :Voiced %s in %s", myname,
			      args[onarg], tmp->name);
# endif
		    } else
			valid = 0;
		} else {
		    tmpnick =
			getChanUserData(tmp, getNickData(args[onarg]));
		    if (tmpnick) {
			tmpnick->op &= ~CHANVOICE;
# ifdef CDEBUG
			sSend(":%s GNOTICE :DeVoiced %s in %s", myname,
			      args[onarg], tmp->name);
# endif
		    } else
			valid = 0;
		}
		onarg++;
	    } else
		valid = 0;
	    break;
	}
    }

    if (!valid) {
	uSend(OperServ, nick, GetReply(ERR_INVALIDMODES, nick), args[2]);
	return;
    }

    string[0] = 0;
    for (i = 2; i < onarg && args[i]; i++) {
	strcat(string, args[i]);
	strcat(string, " ");
    }
    string[strlen(string) - 1] = 0;

    sSend(":%s MODE %s %s", OperServ, args[1], string);
    sSend(":%s GLOBOPS :%s MODE %s (%s)", OperServ, from, args[1], string);
    logDump(operlog, "%s!%s@%s MODE %s (%s)", from, nick->user, nick->host,
	    args[1], string);
}

void OrawCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;
    char msg[512];
    int i = 2;

    if (!args[1])
	return;
    strcpy(msg, args[1]);
    while (i < numargs) {
	strcat(msg, " ");
	strcat(msg, args[i]);
	i++;
    }
    sSend("%s", msg);
    logDump(rootlog, "%s!%s@%s RAW (%s)", from, nick->user, nick->host,
	    msg);
    return;
}

void OinjectCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;
    char msg[512];
    int i = 2;

    if (!args[1])
	return;
    strcpy(msg, args[1]);
    while (i < numargs) {
	strcat(msg, " ");
	strcat(msg, args[i]);
	i++;
    }
    parseLine(msg);
    logDump(rootlog, "%s!%s@%s INJECT (%s)", from, nick->user, nick->host, msg);
    return;
}

void OrootOpsCmd (UserList *nick, char **args, int num)
{
	int i,l,s;
	UserList *tmp;
	char msg[4096];
	
	*msg=0;s=0;
	for (i=1;i<num;i++)
	{
		l=strlen (args[i]);
		if ((s+l)>4095) break;;
		strcat (msg," ");
		strcat (msg,args[i]);
		s+=l;
	}
	for (tmp=firstUser;tmp;tmp=tmp->next)
	{
		if (!tmp->reg) continue;
		if (tmp->reg->flags&(NROOT))
		{
			sSend (":%s NOTICE %s :*** RootOps from %s: %s",OperServ,tmp->nick,nick->nick,msg+1);
		}
	}
	return;
}

void OrestartCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;
    char msg[256];

    *msg = 0;

    if (numargs > 1) {
	int i = 2;

	strcpy(msg, " (");
	strcat(msg, args[1]);

	while (i < numargs) {
	    strcat(msg, " ");
	    strcat(msg, args[i]);
	    i++;
	}
	strcat(msg, ")");
    }

    sSend(":%s GLOBOPS :RESTART From %s%s", myname, from, msg);
    logDump(operlog, "%s!%s@%s RESTART%s", from, nick->user, nick->host,
	    msg);
    uSend(OperServ, nick, GetReply(RPL_RESTARTING, nick));
    srestart();
}

void OshutdownCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;
    char msg[256];

    *msg = 0;

    if (numargs > 1) {
	int i = 2;

	strcpy(msg, " (");
	strcat(msg, args[1]);

	while (i < numargs) {
	    strcat(msg, " ");
	    strcat(msg, args[i]);
	    i++;
	}
	strcat(msg, ")");
    }

    sSend(":%s GLOBOPS :SHUTDOWN From %s%s", myname, from, msg);
    logDump(operlog, "%s!%s@%s SHUTDOWN%s", from, nick->user, nick->host,
	    msg);

    uSend(OperServ, nick, GetReply(RPL_SHUTTINGDOWN, nick));
    sshutdown(0);
}

void OjupeCmd(UserList * nick, char **args, int numargs)
{
	char *from = nick->nick;
	char jpreason[80];
	aJupe *jupeptr, *jupeptr2;
	int i;

	if (!strcasecmp(args[1], "add")) {
		if (!isServop(nick) && !isRoot(nick)) {
			uSend(OperServ, nick, GetReply(ERR_NEEDSERVOP, nick));
			return;
		}
		
		if (numargs < 4) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JUPE");
			return;
		}

		if (CheckJupe(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_THISEXISTS, nick), args[2], "jupe");
			return;
		}
		
		if (CheckServer(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_THISEXISTS, nick), args[2], "server");
			return;
		}
  
		strncpy(jpreason, args[3], sizeof(jpreason));
		for (i = 4; i < numargs; i++) {
			if ((strlen(jpreason) + strlen(args[i])) > 78)
				break;

			strcat(jpreason, " ");
			strcat(jpreason, args[i]);
		}
		
		AddJupe(args[2], from, CTime, jpreason, 1);
						  
	 	sSend(":%s GLOBOPS :Server %s Jupitered by %s: %s", myname, args[2], from, jpreason);
		uSend(OperServ, nick, GetReply(RPL_JINXADDED, nick), "JUPE", args[2]);
		logDump(operlog, "%s!%s@%s Jupitered Server %s: %s", from, nick->user, nick->host, args[2], jpreason);
		return;
	} else if (!strcasecmp(args[1], "del")) {
      if (!isServop(nick) && !isRoot(nick)) {
			uSend(OperServ, nick, GetReply(ERR_NEEDSERVOP, nick));
			return;
		}
									
		if (numargs < 3) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JUPE");
			return;
		}
		if (!CheckJupe(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHJINX, nick), "JUPE", args[2]);
			return;
		}

		DeleteJupe(args[2]);

		sSend(":%s SQUIT %s", myname, args[2]);
		sSend(":%s GLOBOPS :%s Un-jupitered by %s", myname, args[2], from);
		uSend(OperServ, nick, GetReply(RPL_JINXREMOVED, nick), "JUPE", args[2]);
		logDump(operlog, "%s!%s@%s removed JUPE for %s", from, nick->user, nick->host, args[2]);

		return;
	} else if (!strcasecmp(args[1], "list")) {
		if (JupeList == NULL) {
			uSend(OperServ, nick, GetReply(RPL_JINXLISTEMPTY, nick), "JUPE");
			return;
		}
		uSend(OperServ, nick, GetReply(RPL_JINXLISTHEADER, nick), "JUPE");
		for (jupeptr = JupeList; jupeptr; jupeptr = jupeptr2) {
			jupeptr2 = jupeptr->next;
			uSend(OperServ, nick, GetReply(RPL_JINXLIST1, nick), jupeptr->server, jupeptr->juper, jupeptr->reason, ctime(&(jupeptr->date)));
		}
		uSend(OperServ, nick, GetReply(RPL_JINXLISTFOOTER, nick), "JUPE");
	} else {
		uSend(OperServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[1]);
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JUPE");
		return;
	}
}

void ObroadcastCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    int i;

    char broadcast[256];
    memset(broadcast, '\0', 255);

    for (i = 1; i < numargs && 
         strlen(broadcast) + strlen(args[i]) <= 255; i++) {
	strcat(broadcast, args[i]);
	strcat(broadcast, " ");
    }

    sSend
	(":%s NOTICE $*.Serenity-IRC.Net :\002[GLOBAL NOTICE]\002 %s (Please do not reply.)",
	 Serenity, broadcast);
    sSend(":%s GLOBOPS :%s used global broadcast.", myname, from);
    logDump(operlog, "%s used global broadcast (%s).", from, broadcast);
    return;
}

void OsvsjoinCmd(UserList * nick, char **args, int numargs)
{
    if (!getNickData(args[1])) {
	uSend(OperServ, nick, GetReply(ERR_NOSUCHNICK, nick), args[1]);
	return;
    }

    if (args[2][0] != '#') {
	uSend(OperServ, nick, GetReply(ERR_BADCHANNELNAME, nick), args[2]);
	return;
    }

    sSend(":%s GLOBOPS :%s pushed %s into %s.",
	  OperServ, nick->nick, args[1], args[2]);
    sSend(":%s SVSJOIN %s %s", myname, args[1], args[2]);
    
    logDump(operlog, "%s pushed %s into %s.", nick->nick, args[1], args[2]);
    return;
}

void OsvspartCmd(UserList * nick, char **args, int numargs)
{
    if (!getChanData(args[2])) {
	uSend(OperServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[2]);
	return;
    }

    if (!getNickData(args[1])) {
	uSend(OperServ, nick, GetReply(ERR_NOSUCHNICK, nick), args[1]);
	return;
    }

    sSend(":%s GLOBOPS :%s pulled %s out of %s.",
	  myname, nick->nick, args[1], args[2]);
    sSend(":%s SVSPART %s %s", myname, args[1], args[2]);
    
    logDump(operlog, "%s pulled %s out of %s.", nick->nick, args[1], args[2]);
    return;
}

void OsvsbeamCmd(UserList * nick, char **args, int numargs)
{
    if (!getChanData(args[2])) {
	uSend(OperServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[2]);
	return;
    }

    if (!getNickData(args[1])) {
	uSend(OperServ, nick, GetReply(ERR_NOSUCHNICK, nick), args[1]);
	return;
    }
	
    if (args[3][0] != '#') {
	uSend(OperServ, nick, GetReply(ERR_BADCHANNELNAME, nick), args[3]);
	return;
    }

    sSend(":%s GLOBOPS :%s beamed %s from %s to %s.",
	  myname, nick->nick, args[1], args[2], args[3]);
    sSend(":%s SVSPART %s %s", OperServ, args[1], args[2]);
    sSend(":%s SVSJOIN %s %s", OperServ, args[1], args[3]);

    logDump(operlog, "%s beamed %s from %s to %s.", nick->nick, args[1], args[2], args[3]);
    return;
}

void OsvsnoopCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    aServer *server;
    UserList *tmp;

    if (!strcasecmp(args[1], "ADD")) {

	server = GetServer(args[2]);
	if (!server) {
	    uSend(OperServ, nick, GetReply(ERR_NOSUCHSERVER, nick),
		  args[2]);
	    return;
	}

	if (server->svsnoop == 1) {
	    uSend(OperServ, nick, GetReply(ERR_ALREADYSVSNOOP, nick),
		  args[2]);
	    return;
	}

	server->svsnoop = 1;
	sSend(":%s GLOBOPS :%s activated SVSNOOP for %s.", myname, from,
	      args[2]);
	sSend(":%s SVSNOOP %s +", myname, server->name);

	for (tmp = firstUser; tmp; tmp = tmp->next) {
	    if (isOper(tmp) && !strcasecmp(tmp->server, server->name)) {
		setMode(tmp->nick, "-NnAaOo");
	    }
	}
    }

    else if (!strcasecmp(args[1], "DEL")) {

	server = GetServer(args[2]);
	if (!server) {
	    uSend(OperServ, nick, GetReply(ERR_NOSUCHSERVER, nick),
		  args[2]);
	    return;
	}

	if (server->svsnoop == 0) {
	    uSend(OperServ, nick, GetReply(ERR_NOTSVNOOP, nick), args[2]);
	    return;
	}

	server->svsnoop = 0;
	sSend(":%s GLOBOPS :%s deactivated SVSNOOP for %s.", myname, from,
	      args[2]);
	sSend(":%s SVSNOOP %s -", myname, server->name);
    }

    logDump(operlog, "%s used SVSNOOP (%s).", from, args[2]);
    return;
}

void OekillCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    anEmail *emailptr;
    anEmail *emailptr2;

    char ereason[80];

    int i;

    if (numargs < 3 || !strcasecmp(args[1], "LIST")) {

	if (EmailList == NULL) {
	    uSend(OperServ, nick, GetReply(RPL_EKILLEMPTY, nick));
	    return;
	}

	uSend(OperServ, nick, GetReply(RPL_EKILLHEAD, nick));
	for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	    emailptr2 = emailptr->next;
	    uSend(OperServ, nick, GetReply(RPL_EKILLDATA, nick),
		  emailptr->email, emailptr->by, emailptr->reason,
		  ctime(&(emailptr->time)));
	}
	uSend(OperServ, nick, GetReply(RPL_EKILLEND, nick));
	return;

    }

    if ((!strchr(args[2], '@')) || (strlen(args[2]) < 5)
	|| (strlen(args[2]) > EMAILLEN)) {
	uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
	return;
    }

    if (numargs < 3) {
	uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
	return;
    }

    if (!strcasecmp(args[1], "DEL")) {
	if (!CheckEmail(args[2])) {
	    uSend(OperServ, nick, GetReply(ERR_NOSUCHEKILL, nick),
		  args[2]);
	    return;
	}

	uSend(OperServ, nick, GetReply(RPL_EKILLDELED, nick), args[2]);
	sSend(":%s GLOBOPS :%s removed dynamic EKILL for %s", OperServ,
	      from, args[2]);
	DeleteEmail(args[2]);
	return;
    }

    if (numargs < 4) {
	uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
	return;
    }

    if (!strcasecmp(args[1], "ADD")) {

	if (CheckEmail(args[2])) {
	    uSend(OperServ, nick, GetReply(ERR_EKILLEXISTS, nick),
		  args[2]);
	    return;
	}

	bzero(ereason, 80);
	strncpy(ereason, args[3], 79);

	for (i = 4; i < numargs; i++) {
	    if ((strlen(ereason) + strlen(args[i])) > 78)
		break;
	    strcat(ereason, " ");
	    strcat(ereason, args[i]);
	}

	uSend(OperServ, nick, GetReply(RPL_EKILLADDED, nick), args[2]);
	sSend(":%s GLOBOPS :%s added dynamic EKILL for %s (%s)", OperServ,
	      from, args[2], ereason);
	AddEmail(args[2], from, ereason);
    }

    return;
}

void OlistCmd(UserList * nick, char **args, int numargs)
{
    HostCloneList *clonelist;
    int i = 0;

    if (!strcasecmp(args[1], "PERMIGNORES")) {
	uSend(OperServ, nick, GetReply(RPL_PERMCLONEIGNHEAD, nick));
	for (clonelist = firstPermClone; clonelist;
	     clonelist = clonelist->next) {
	    i++;
	    uSend(OperServ, nick, GetReply(RPL_PERMCLONEIGNDATA, nick), i,
		  clonelist->host);
	}
	uSend(OperServ, nick, GetReply(RPL_PERMCLONEIGNEND, nick));
    }
    return;
}

void OuptimeCmd(UserList * nick, char **args, int numargs)
{
    long blah;
    int days, hours, mins, seconds;

    blah = time(NULL);
    blah -= startup;
    days = (int) (blah / (24 * 3600));
    blah %= (24 * 3600);
    hours = (int) (blah / (3600));
    blah %= (3600);
    mins = (int) (blah / (60));
    blah %= (60);
    seconds = blah;
    uSend(OperServ, nick, GetReply(RPL_UPTIME1, nick));
    uSend(OperServ, nick, GetReply(RPL_UPTIME2, nick), days, hours, mins,
	  seconds);
    return;
}

void OallocstatCmd(UserList * nick, char **args, int numargs)
{
#ifndef WIN32
    char *from = nick->nick;
    FILE *fp;

    char psline[128];

    uSend(OperServ, nick, GetReply(RPL_ALLOCSTAT1, nick), AllocCalled);
    uSend(OperServ, nick, GetReply(RPL_ALLOCSTAT2, nick), FreeCalled);

    /* Ax0n: Added -w for wrap, incase we are invoked /very/long/path/services */
    sprintf(psline, "ps -uaxw | grep services");
    fp = popen(psline, "r");
    if (fp) {
	while (sfgets(psline, 128, fp)) {
	    if (match(psline, "*grep*"))
		sSend(":%s NOTICE %s :%s", OperServ, from, psline);
	}
	fclose(fp);
    }
    return;
#endif
}

void OsyncCmd(UserList * nick, char **args, int numargs)
{
    save_access_mod();
    strcpy(args[0], "save");
    sendToNickServ(nick, args, numargs);
    sendToChanServ(nick, args, numargs);
    sSend(":%s GLOBOPS :Synchronizing all remaining databases to disk...",
	  OperServ);
    savedynamicekills();
}

void OtriggerCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;

    if (index(args[1], '@')) {
	char user[USERLEN];
	char host[HOSTLEN];
	HostClone *hc;
	UserClone *uc;
	sscanf(args[1], "%[^@]@%s", user, host);
	hc = getCloneData(host);
	if (hc)
	    uc = getUserCloneData(hc, user);
	else
	    uc = NULL;
	if (uc && hc) {
	    if (numargs < 3)
		uSend(OperServ, nick, GetReply(RPL_CLONESTATUS1, nick),
		      user, host, uc->clones, uc->trigger);
	    else {
		int i = atoi(args[2]);
		if (i > 0) {
		    uc->trigger = i;
		    sSend(":%s GLOBOPS :%s@%s retriggered to %i by %s",
			  OperServ, user, host, i, from);
		} else
		    uSend(OperServ, nick,
			  GetReply(ERR_INVALIDTRIGLEVEL, nick));
	    }
	} else
	    uSend(OperServ, nick, GetReply(ERR_NOSUCHTRIGGER, nick), user,
		  host);
    } else {
	char host[HOSTLEN];
	HostClone *hc;
	strcpy(host, args[1]);
	hc = getCloneData(host);
	if (hc) {
	    if (numargs < 3)
		uSend(OperServ, nick, GetReply(RPL_CLONESTATUS2, nick),
		      host, hc->clones, hc->trigger);
	    else {
		int i = atoi(args[2]);
		if (i > 0) {
		    hc->trigger = i;
		    sSend(":%s GLOBOPS :%s retriggered to %i by %s",
			  OperServ, host, i, from);
		} else
		    uSend(OperServ, nick,
			  GetReply(ERR_INVALIDTRIGLEVEL, nick));
	    }
	} else {
	    uSend(OperServ, nick, GetReply(ERR_NOSUCHCHAN, nick), host);
	}
    }
}

void OclonesetCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;
    char date[20];
    struct tm *t;
    HostClone *hc;
    HostCloneList *permignore;
    int tempflags = 0;
    int goodhost = 0;

    if (!strcasecmp(args[1], "list")) {
        sSend(":%s NOTICE %s :Listing PERMIGNORES only:", OperServ, from);
        for (permignore = firstPermClone; permignore; permignore = permignore->next) {
            t = localtime(&(permignore->setwhen));
            strftime(date, 20, "%m/%d/%y-%I:%M%p", t);
                                            
            sSend(":%s NOTICE %s :%s %-40s (%s)",
                  OperServ, from, date, permignore->host, 
                  permignore->setby);
        }
        sSend(":%s NOTICE %s :End of list", OperServ, from);
        return;
    }

    hc = getCloneData(args[1]);

    if (!hc) {
	for (permignore = firstPermClone; permignore;
	     permignore = permignore->next) {
	    if (!strcasecmp(permignore->host, args[1])) {
		tempflags = CLONE_PERMIGNORE;
		break;
	    }
	}
    } else
	tempflags = hc->flags;

    if (numargs < 3) {
	if (tempflags)
	    uSend(OperServ, nick, GetReply(RPL_HASCLONEFLAGS, nick),
		  args[1],
		  (tempflags & CLONE_KILLFLAG) ? " Auto Remove" : "",
		  (tempflags & CLONE_IGNOREFLAG) ? " Ignore All" : "",
		  (tempflags & CLONE_PERMIGNORE) ? " Perm Ignore" : "");
	else
	    uSend(OperServ, nick, GetReply(RPL_NOCLONEFLAGS, nick),
		  args[1]);
    } else {
	if (!strcasecmp(args[2], "kill")) {
	    if (hc) {
		hc->flags |= CLONE_KILLFLAG;
		tempflags = hc->flags;
		goodhost = 1;
	    }
	} else if (!strcasecmp(args[2], "-kill")) {
	    if (hc) {
		hc->flags &= ~CLONE_KILLFLAG;
		tempflags = hc->flags;
		goodhost = 1;
	    }
	} else if (!strcasecmp(args[2], "ignore")) {
	    if (hc) {
		hc->flags |= CLONE_IGNOREFLAG;
		tempflags = hc->flags;
		goodhost = 1;
	    }
	} else if (!strcasecmp(args[2], "-ignore")) {
	    if (hc) {
		hc->flags &= ~CLONE_IGNOREFLAG;
		tempflags = hc->flags;
		goodhost = 1;
	    }
	} else if (!strcasecmp(args[2], "permignore")) {

	    if (!hc && tempflags) {
		uSend(OperServ, nick, GetReply(ERR_PERMIGNEXISTS, nick),
		      args[1]);
		goodhost = 2;
	    }

	    if (hc) {
		if (hc->flags & CLONE_PERMIGNORE) {
		    uSend(OperServ, nick,
			  GetReply(ERR_PERMIGNEXISTS, nick), args[1]);
		    goodhost = 2;
		} else {
		    hc->flags |= CLONE_PERMIGNORE;
		    permignore =
			smalloc(sizeof(struct hostcloneliststruct));
		    strcpy(permignore->host, hc->host);
		    strcpy(permignore->setby, from);
		    permignore->setwhen = CTime;
		    permignore->next = firstPermClone;
		    permignore->prev = NULL;
		    if (permignore->next)
			permignore->next->prev = permignore;
		    firstPermClone = permignore;
		    tempflags = hc->flags;
		    goodhost = 1;
		}
	    }
	} else if (!strcasecmp(args[2], "-permignore")) {
	    if (hc) {
		hc->flags &= ~CLONE_PERMIGNORE;
		tempflags = hc->flags;
	    }

	    for (permignore = firstPermClone; permignore;
		 permignore = permignore->next) {
		if (!strcasecmp(args[1], permignore->host)) {
		    if (permignore->prev)
			permignore->prev->next = permignore->next;
		    if (permignore->next)
			permignore->next->prev = permignore->prev;
		    if (permignore == firstPermClone)
			firstPermClone = permignore->next;
		    sfree(permignore);
		    goodhost = 1;
		    break;
		}
	    }

	    if (!goodhost) {
		uSend(OperServ, nick, GetReply(ERR_NOSUCHPERMIGN, nick),
		      args[1]);
		return;
	    }
	}

	else {
	    uSend(OperServ, nick, GetReply(ERR_UNKNOWNFLAG, nick),
		  args[1]);
	    return;
	}

	if (goodhost == 1) {
	    if (tempflags)
		sSend(":%s GLOBOPS :%s changed flags for %s (now%s%s%s).",
		      OperServ, from, args[1],
		      (tempflags & CLONE_KILLFLAG) ? " Auto Remove" : "",
		      (tempflags & CLONE_IGNOREFLAG) ? " Ignore All" : "",
		      (tempflags & CLONE_PERMIGNORE) ? " Perm Ignore" :
		      "");
	    else
		sSend(":%s GLOBOPS :%s changed flags for %s (now NONE).",
		      OperServ, from, args[1]);
	} else if (goodhost != 2)
	    uSend(OperServ, nick, GetReply(ERR_UNKNOWHOST, nick), args[1]);
    }
}

void OmkillCmd(UserList * nick, char **args, int numargs)
{
    char *from = nick->nick;
    char *s1;
    char *at; // Position on '@' in the mask 
    char *start, *end;
    int i = 0;
    char user[USERLEN];
    char host[HOSTLEN];
    int goodmask = 1; // Innocent until proven guilty
    
    /* Check the input received for validity */
    start = s1 = args[1];
    end = start + strlen(args[1]) + 1;
    at = index(args[1], '@');

    while (*s1 && goodmask) {
        /* Wildcards are not allowed, nor is !, nor is a dot in the userpart.
         * Only one @ is allowed in the mask */
        if ((*s1 == '?') || (*s1 == '*') || (*s1 == '!') || 
            ((*s1 == '.') && (s1 < at)) || ((*s1 == '@') && (s1 > at)))
            goodmask = 0;
        s1++;
    }
    
    /* Check that we won't overflow our buffer if we'd split the string up */
    if (at) {
        if ((end - at) > HOSTLEN || ((at - start) + 1) > USERLEN)
            goodmask = 0;
    } else {
        if ((end - start) + 1 > HOSTLEN)
            goodmask = 0;
    }
    
    /* Did we find an error in the mask? Then we give up here */
    if (!goodmask) {
	uSend(OperServ, nick, GetReply(ERR_BADMASK, nick), args[1]);
	uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ,
	      args[0]);
	return;
    }

    if (at)
	sscanf(args[1], "%[^@]@%s", user, host);
    else {
	strcpy(host, args[1]);
	strcpy(user, "*");
    }
    
    i = killMassClones(user, host);

    if (i) {
	sSend(":%s GLOBOPS :%s masskilled %s@%s (%i matches)", OperServ,
	      from, user, host, i);
	logDump(operlog, "%s!%s@%s masskilled %s@%s (%i matches)", from,
		nick->user, nick->host, user, host, i);
    } else
	uSend(OperServ, nick, GetReply(ERR_NOMATCH, nick), user, host);
}

void OmapCmd(UserList * nick, char **args, int numargs)
{

    aServer *srvptr;
    aServer *srvptr2;

//      printf("%s", 3408932); /* <- manual core if uncommented, useful for core-debugging -GZ */

    if (ServerList == NULL)
	return;

	uSend(OperServ, nick, GetReply(RPL_MAPHEAD, nick));
	for (srvptr = ServerList; srvptr; srvptr = srvptr2) {
		srvptr2 = srvptr->next;
		uSend(OperServ, nick, GetReply(RPL_MAPDATA, nick), srvptr->name, srvptr->version);
	}
	uSend(OperServ, nick, GetReply(RPL_MAPDATA, nick), myname, PACKAGE " v" VERSION);
	uSend(OperServ, nick, GetReply(RPL_MAPEND, nick));
	return;
}

void OjinxCmd(UserList * nick, char **args, int numargs)
{

	char *from = nick->nick;
	UserList *target;
	aJinx *jinxptr, *jinxptr2;
	int i;
	char reason[1000];

	bzero(reason, 1000);

	if (!strcasecmp(args[1], "add")) {
		if (numargs < 4) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JINX");
			return;
		}

		if (!(target = getNickData(args[2]))) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHNICK, nick), args[2]);
			return;
		}

		if (CheckJinx(target->nick)) {
			uSend(OperServ, nick, GetReply(ERR_JINXEXIST, nick), "JINX",
			target->nick);
			return;
		}

		if (isRoot(target)) {
			uSend(OperServ, nick, GetReply(ERR_CANTJINXSRA, nick));
			sSend(":%s GLOBOPS :\002Warning:\002 %s attempted to JINX \002%s\002.", OperServ, from, args[2]);
			return;
		}

		sprintf(reason, args[3]);
		for (i = 4; i < numargs; i++) {
			strcat(reason, " ");
			strcat(reason, args[i]);
		}

		AddJinx(target->nick, from, reason);

		sSend(":%s GLOBOPS :%s added a JINX for \002%s\002 (%s)", OperServ, from, target->nick, reason);
		sSend(":%s JINX %s", myname, target->nick);
		uSend(OperServ, nick, GetReply(RPL_JINXADDED, nick), "JINX", target->nick);
		logDump(operlog, "%s!%s@%s Added a JINX for %s", from, nick->user,
		nick->host, target->nick);

		return;
    } else if (!strcasecmp(args[1], "del")) {

		if (numargs < 3) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JINX");
			return;
		}

		if (!CheckJinx(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHJINX, nick), "JINX", args[2]);
			return;
		}

		DeleteJinx(args[2]);
		sSend(":%s GLOBOPS :%s removed the JINX for \002%s\002.", OperServ, from, args[2]);
		uSend(OperServ, nick, GetReply(RPL_JINXREMOVED, nick), "JINX", args[2]);
		sSend(":%s UNJINX %s", myname, args[2]);
		logDump(operlog, "%s!%s@%s removed a JINX for %s", from,
		nick->user, nick->host, args[2]);

		return;
	} else if (!strcasecmp(args[1], "list")) {

		if (JinxList == NULL) {
			uSend(OperServ, nick, GetReply(RPL_JINXLISTEMPTY, nick), "JINXES");
			return;
		}

		uSend(OperServ, nick, GetReply(RPL_JINXLISTHEADER, nick), "JINXES");
		for (jinxptr = JinxList; jinxptr; jinxptr = jinxptr2) {
			jinxptr2 = jinxptr->next;
			uSend(OperServ, nick, GetReply(RPL_JINXLIST1, nick), jinxptr->nick, jinxptr->by, jinxptr->reason, ctime(&(jinxptr->time)));
		}
		uSend(OperServ, nick, GetReply(RPL_JINXLISTFOOTER, nick), "JINXES");
	} else {
		uSend(OperServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[1]);
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JINX");
	}

	return;
}

void OenforceCmd(UserList * tmp, char **args, int numargs)
{

    char *from = tmp->nick;
    RegNickList *reg;
    int i;

    anEmail *emailptr;
    anEmail *emailptr2;

    i = 0;

    if (EmailList == NULL)
	return;

    sSend(":%s GLOBOPS :Enforcing EKILLS. (Requested by %s)", OperServ,
	  from);

    for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	emailptr2 = emailptr->next;

	for (reg = firstRegNick; reg; reg = reg->next) {
	    if (!match(reg->email, emailptr->email)) {
		i++;
		logDump(nicklog, "%s DELETE %s (EKILL ENFORCE for %s)",
			from, reg->nick, reg->email);
		delRegNick(reg);
	    }
	}
    }

    logDump(nicklog, "EKILL ENFORCE requested by %s", from);
    sSend
	(":%s GLOBOPS :Deleted %i nicks from database. (EKILL ENFORCE by %s)",
	 NickServ, i, from);
}

void OlistOpsCmd (UserList *nick, char **args, int numargs)
{
	char *psn;
	UserList *tn;
	aServer *svr;
	unsigned int flags;
	
	psn=nick->server;
	if (numargs>1) psn=args[1];
	svr=GetServer (psn);
	if (!svr)
	{
		uSend (OperServ,nick,"No such server: %s",psn);
		return;
	}
	uSend (OperServ,nick,"*** Online opers for server %s:",svr->name);
	for (tn=firstUser;tn;tn=tn->next) if (!match (tn->server,psn))
	{
		flags=tn->oflags;
		printf ("%s %08X\n",tn->nick,flags);
		if (flags&NISADMIN) uSend (OperServ,nick,"Admin: %s",tn->nick);
		else if (flags&NISOPER) uSend (OperServ,nick,"Oper: %s",tn->nick);
	}
	uSend (OperServ,nick,"*** End of Oper list.");
	return;
}
