/*
 * various functions that do useful crap. 
 */

#include "services.h"

/*
 * this will assign an integer # of arguments and place them each in a 
 * * sequential order in a character array of any number of args.
 */

/*
 * ===DOC===
 * This function returns data much like an argc/argv setup.
 * send it an integer, the string of data you want broken up, an array for
 * storage, and a delimiter character to seperate sets of data by.
 * Ex: I call fixUpData like this:
 * int blah;
 * char *holdme[];
 * ...
 * breakString(blah, "WD is a good programmer", holdme[], ' ');
 * now:
 * blah == 5
 * holdme[0] == "WD", holdme[2] == "a", holdme[4] == "programmer", etc.
 * (All strings end in \0)
 * the idea for this was given to me by JoelKatz (David Schwartz) of DALnet
 * 
 */

#define MAXVIRTSIZE     (3 + 5 + 1)

int checkDelim(char, char *);

int checkDelim(char it, char *delims)
{
    while (*delims) {
	if (it == *delims)
	    return 1;
	else
	    delims++;
    }
    return 0;
}


int breakString(int maxargs, char *string, char **args, char *delimiter)
{
    int x, a;

    x = a = 0;

    while ((*string) && (x < maxargs)) {
	while (*string && !checkDelim(*string, delimiter)) {
	    args[x][a] = *string;
	    a++;
	    string++;
	}
	args[x][a] = 0;
	x++;
	while (*string && checkDelim(*string, delimiter))
	    string++;
	a = 0;
    }
    return x;
}


/*
 * ===DOC===
 * * Simple, really.
 * * int i;
 * * ...
 * * i=match("IamAfooAren'ti", "*foo*")
 * * ...
 * * i==0.
 * * i=match("IamAfooAren'ti", "*bar*")
 * * ..
 * * i==1;
 */

/*
 * Note, thanx to RogerY and Redtrio for this wonderful code.  Please
 * keep this notice intact.  -- TazQ 
 */

/*
 * My wildcard matching functions. I think it works well :) Seems to be 
 * working stablely, but not too sure.                                  
 */
int match(char *str, char *wld)
{
    int i = check_match(wld);
    switch (i) {
    case 0:
	return strcasecmp(str, wld);
	break;
    case 1:
	return 0;
	break;
    }

    if (!(*wld) || !(*str))
	return 1;
    while (*wld || *str) {
	if (tolower(*wld) == tolower(*str)) {
	    wld++;
	    str++;
	} else if (tolower(*wld) != tolower(*str) && *wld == '*') {
	    int i = 0;
	    char *st;
            st = (char *) smalloc(strlen(wld) + 1);
	    while (*wld == '*' && *wld) {
		wld++;
		if (!*wld) {
		    sfree(st);
		    return 0;
		}
	    }
	    while (*wld && *wld != '*' && *wld != '?') {
		st[i] = *wld;
		i++;
		wld++;
	    }
	    st[i] = '\0';
	    while (*str && strncasecmp(st, str, strlen(st)))
		str++;
	    if (!strncasecmp(st, str, strlen(st)))
		str += strlen(st);
	    else {
		sfree(st);
		return 1;
	    }
	    sfree(st);
	} else if (tolower(*wld) != tolower(*str) && *wld == '?' && *wld
		   && *str) {
	    wld++;
	    str++;
	} else if (tolower(*wld) != tolower(*str))
	    return 1;
    }
    return 0;
}

/*
 * experimental matching code (WD) I hope this runs faster, who knows...
 */
int check_match(char *mask)
{
    if (!index(mask, '*') && !index(mask, '?'))
	return 0;
    else if (!strcmp(mask, "*"))
	return 1;
    else
	return 2;
}

int exp_match(char *string, char *mask)
{
    int i = check_match(mask);
    if (!i)
	return strcasecmp(string, mask);
    else if (i == 1 && *string)
	return 0;
    else {
	char *str = string;
	char *wld = mask;

	while (*str || *wld) {
	    if (*wld == '*') {
		wld++;
		if (*wld) {
		    str = index(str, *wld);
		    if (!*str && *wld)
			return 1;
		} else
		    return 0;
	    } else if (*wld == '?') {
		str++;
		wld++;
	    } else if (*str == *wld) {
		str++;
		wld++;
	    } else if (!*str && !*wld)
		return 0;
	    else
		return 1;
	}
	return 1;
    }
}

#define mklower(c)      (((c) >= 'A' && (c) <= 'Z') ? ((c) - 'A' + 'a') : c)

void strtolower(char *str)
{
    char tmp;
    int i;

    for (i = 0; str[i]; i++) {
	tmp = mklower(str[i]);
	str[i] = tmp;
    }
}

char *sfgets(char *str, int len, FILE * fp)
{
    if (!fgets(str, len, fp))
	return NULL;
    else {
	if (str[0])
	    str[strlen(str) - 1] = 0;
	return str;
    }
}

int killMassClones(char *usermask, char *hostmask)
{
    UserList *tmp;
    char clonestr[15];
    int i = 0;

    for (tmp = firstUser; tmp; tmp = tmp->next) {
	if (!match(tmp->host, hostmask) && !match(tmp->user, usermask)) {
	    i++;
	    sprintf(clonestr, "clone (%d)", i);
	    killUser(services[0], tmp, clonestr);
	}
    }
    return i;
}

HostClone *firstClone = NULL;
HostClone *lastClone = NULL;

void initCloneData(HostClone * hc)
{
    hc->host[0] = 0;
    hc->firstUser = hc->lastUser = NULL;
    hc->trigger = DEFHOSTCLONETRIGGER;
    hc->clones = 0;
    hc->flags = DEFCLONEFLAGS;
    hc->next = hc->previous = hc->hashnext = hc->hashprev = NULL;
}

int addClone(char *nick, char *user, char *host)
{
    HostClone *hc;
    HostCloneList *clonelist;
    long hashEnt;
    int i, killcnt;
    char tempstr[USERLEN + HOSTLEN + 4];

    hc = getCloneData(host);
    if (!hc) {
	hc = smalloc(sizeof(HostClone));
	initCloneData(hc);
	strcpy(hc->host, host);
	hc->clones = 1;
	if (!firstClone) {
	    firstClone = hc;
	    hc->previous = NULL;
	} else {
	    lastClone->next = hc;
	    hc->previous = lastClone;
	}
	lastClone = hc;
	hc->next = NULL;

	hashEnt = getHashKey(host) % CLONEHASHSIZE;
	if (!CloneHash[hashEnt].clone) {
	    CloneHash[hashEnt].clone = hc;
	    hc->hashprev = NULL;
	} else {
	    CloneHash[hashEnt].lastclone->hashnext = hc;
	    hc->hashprev = CloneHash[hashEnt].lastclone;
	}
	CloneHash[hashEnt].lastclone = hc;
	hc->hashnext = NULL;

	for (clonelist = firstPermClone; clonelist;
	     clonelist = clonelist->next) {
	    if (!strcasecmp(host, clonelist->host)) {
		hc->flags = CLONE_PERMIGNORE;
		break;
	    }
	}

    } else
	hc->clones++;

    i = addUserClone(hc, user);

    if (hc->flags & CLONE_IGNOREFLAG || hc->flags & CLONE_PERMIGNORE)
	return 0;

    if (i) {
	sSend(":%s GLOBOPS :%i possible clones from %s@%s", OperServ, i,
	      user, host);
	return 0;
    }

    if (hc->flags & CLONE_PERMIGNORE)
	return 0;

    if (hc->clones >= hc->trigger) {
	sSend(":%s GLOBOPS :%i possible clones from *@%s", OperServ,
	      hc->clones, host);
	if (hc->clones >= hc->trigger) {
	    if (hc->flags & CLONE_KILLFLAG) {
		sSend(":%s KILL %s :%s!%s (No Clones Allowed.)", NickServ,
		      nick, services[1].host, NickServ);
		return 1;
	    } else if (hc->clones > DEFHOSTILEHOSTCLONES) {
		killcnt = killMassClones("*", host);
		sSend(":%s GLOBOPS :%s masskilled *@%s (%i matches)",
		      OperServ, OperServ, host, killcnt);
		logDump(operlog,
			"%s masskilled *@%s [AUTO-TRIGGER] (%i matches)",
			OperServ, host, killcnt);
		strcpy(tempstr, "*!*@");
		strcat(tempstr, host);
		add_access_mod((3 * 3600), tempstr, OperServ, A_AKILL,
			 "Excessive Cloning Not Permitted.");
		return 1;
	    }
	}
    }
    return 0;
}

int addUserClone(HostClone * hc, char *user)
{
    UserClone *uc = getUserCloneData(hc, user);

    if (!uc) {
	uc = smalloc(sizeof(UserClone));
	uc->user[0] = 0;
	uc->clones = 1;
	uc->trigger = DEFUSERCLONETRIGGER;

	strcpy(uc->user, user);

	if (!hc->firstUser) {
	    hc->firstUser = uc;
	    uc->previous = NULL;
	} else {
	    hc->lastUser->next = uc;
	    uc->previous = hc->lastUser;
	}
	hc->lastUser = uc;
	uc->next = NULL;
    } else
	uc->clones++;

    return (((uc->clones >= (uc->trigger)) ? uc->clones : 0));
}

void delClone(char *user, char *host)
{
    HostClone *hc = getCloneData(host);
    if (hc) {
	UserClone *uc = getUserCloneData(hc, user);
	if (uc) {
	    hc->clones--;
	    uc->clones--;
	    if (!uc->clones)
		delUserClone(hc, uc);
	    if (!hc->clones) {
		long hashEnt;
		HostClone *tmp;
		if (hc->previous)
		    hc->previous->next = hc->next;
		else
		    firstClone = hc->next;
		if (hc->next)
		    hc->next->previous = hc->previous;
		else
		    lastClone = hc->previous;
		hashEnt = getHashKey(hc->host) % CLONEHASHSIZE;
		for (tmp = CloneHash[hashEnt].clone; tmp;
		     tmp = tmp->hashnext) {
		    if (tmp == hc) {
			if (hc->hashprev)
			    hc->hashprev->hashnext = hc->hashnext;
			else
			    CloneHash[hashEnt].clone = hc->hashnext;
			if (hc->hashnext)
			    hc->hashnext->hashprev = hc->hashprev;
			else
			    CloneHash[hashEnt].lastclone = hc->hashprev;
		    }
		}
		sfree(hc);
	    }
	}
    }
}

void delUserClone(HostClone * hc, UserClone * uc)
{
    if (uc->previous)
	uc->previous->next = uc->next;
    else
	hc->firstUser = uc->next;
    if (uc->next)
	uc->next->previous = uc->previous;
    else
	hc->lastUser = uc->previous;

    free(uc);
}

HostClone *getCloneData(char *host)
{
    long hashEnt;
    hashEnt = getHashKey(host) % CLONEHASHSIZE;
    if (!CloneHash[hashEnt].clone)
	return NULL;
    else {
	HostClone *tmp;
	for (tmp = CloneHash[hashEnt].clone; tmp; tmp = tmp->hashnext) {
	    if (!strcasecmp(host, tmp->host))
		return tmp;
	}
	return NULL;
    }
}

UserClone *getUserCloneData(HostClone * hc, char *user)
{
    UserClone *tmp;

    for (tmp = hc->firstUser; tmp; tmp = tmp->next) {
	if (!strcasecmp(tmp->user, user))
	    return tmp;
    }
    return NULL;
}

char *Maskchecksum(char *data, char *salt)
{
  char static tmp[HOSTLEN + 1] ;
  
  strncpy(tmp,crypt(data, salt),HOSTLEN) ;
  return (tmp) ;
}

int str2array(char **pparv, char *string, char *delim)
{
    char *tok;
    int pparc = 0;

    tok = (char *) strtok(string, delim);
    while (tok != NULL) {
	pparv[pparc++] = tok;
	tok = (char *) strtok(NULL, delim);
    }
    return pparc;
}

char *return_user_mask(char *s)
{
    static char mask[HOSTLEN + 1];
    char *csum;
    char *dot, *ptr ;
    static char destroy[HOSTLEN + 1], *parv[HOSTLEN + 1];
    char salt[12] = "$1$\0\0\0\0\0\0\0\0\0";
    int len = 0, overflow = 0, parc = 0;
    
    strncpy(destroy, s, HOSTLEN);
    len = strlen(destroy);

    if ((len + MAXVIRTSIZE) > HOSTLEN) {
        overflow = (len + MAXVIRTSIZE) - HOSTLEN;
        ptr = &destroy[overflow];
    } else {
        ptr = &destroy[0];
    }
    memset(mask, 0, HOSTLEN);
    
    parc = str2array(parv, ptr, ".");

    if (strlen(s) > HOSTLEN) {
        s[HOSTLEN] = 0;
    }
    
    if (parc == 4) {
        len = strlen(parv[3]);
        if (isdigit(parv[3][len - 1])) {
            /* Numeric IP, lets use the last octets of the IP address
             * as salt */
            strcat(salt, parv[3]);
            strcat(salt, parv[2]);
            csum = Maskchecksum(s, salt);
            sprintf(mask, "%s.%s.%s.%i", parv[0], parv[1],parv[2],csum[15]+256+csum[19]);
            return mask;
        }
    }

    /* Hostname.... lets use it as our salt... */
    strncat(salt, s, 8);
    csum = Maskchecksum(s, salt);

    dot = (char *) strchr(s, '.');
    
    if (dot == NULL) {
        sprintf(mask, "Serene%i%i%i%i%i%i.%s",csum[14]%10,csum[15]%10,csum[16]%10,csum[17]%10,csum[18]%10,csum[19]%10, s);
        return mask;
    } else {
        sprintf(mask, "Serene%i%i%i%i%i%i.%s",csum[14]%10,csum[15]%10,csum[16]%10,csum[17]%10,csum[18]%10,csum[19]%10, dot + 1);
        return mask;
    }
}

char *maskuser(UserList * tmp)
{
    static char newmask[128];

    if (tmp->reg && tmp->reg->flags & NNETADM)
	strcpy(newmask, "NetAdmin.Serenity-IRC.Net");
    else if (tmp->reg && tmp->reg->flags & NROOT)
	strcpy(newmask, "SRA.Serenity-IRC.Net");
    else if (tmp->reg && tmp->reg->flags & NSERVOP)
	strcpy(newmask, "ServOp.Serenity-IRC.Net");
    else if (tmp->oflags & NISADMIN)
	strcpy(newmask, "Admin.Serenity-IRC.Net");
    else if (tmp->oflags & NISOPER)
	strcpy(newmask, "IRCop.Serenity-IRC.Net");
    else if (tmp->oflags & NISPLUSX)
        strcpy(newmask, return_user_mask(tmp->host));    
    else
	strcpy(newmask, tmp->host);

    return newmask;
}

char *maskreguser(RegNickList * tmp)
{
	static char newmask[128];

	if (tmp->flags & NNETADM) strcpy(newmask, "NetAdmin.Serenity-IRC.Net");
	else if (tmp->flags & NROOT) strcpy(newmask, "SRA.Serenity-IRC.Net");
	else if (tmp->flags & NSERVOP) strcpy(newmask, "Servop.Serenity-IRC.Net");
	else if (tmp->flags2 & NSTICKYOP) strcpy(newmask, "IRCop.Serenity-IRC.Net");
	else if (tmp->flags & NHIDE) strcpy(newmask, return_user_mask(tmp->host));
/*	{
		strcpy(vol_mask, tmp->host);

		parc = str2array(tempmask, vol_mask, ".");

		masksum = maskchecksum(tmp->host);

		if (parc == 4)
		{
			len = strlen(tempmask[3]);
			if (strchr("0123456789", tempmask[3][len - 1]))
			{
				sprintf(newmask, "%s.%s.%s.%i", tempmask[0], tempmask[1],
					tempmask[2], masksum[9] + 256 + masksum[10]);
				return newmask;
			}
		}

		firstdot = strchr(tmp->host, '.');

		if (firstdot == NULL) 
		    sprintf(newmask, "Serene%i%i.%s", masksum[7], masksum[8], tmp->host);
		else 
		    sprintf(newmask, "Serene%i%i.%s", masksum[7], masksum[8], firstdot + 1);
		    }*/
    else strcpy(newmask, tmp->host);

    return newmask;
}

void loadpermignores()
{
    FILE *ignore_file;
    HostCloneList *permclone;
    char buffer[256];
    int count = 0;

    firstPermClone = NULL;

    ignore_file = fopen(DBDIR "/permignore.db", "r");

    if (!ignore_file) {
	return;
    }

    while (sfgets(buffer, 256, ignore_file)) {

	if (strcmp(buffer, "-") != 0) {
	    sSend(":%s GLOBOPS :Sync error in permignore.db", OperServ);
	    fprintf(corelog, "Sync error in permignore.db");
	    fclose(ignore_file);
	    return;
	}

	permclone = smalloc(sizeof(struct hostcloneliststruct));

	fscanf(ignore_file, "%s\n", permclone->host);
	fscanf(ignore_file, "%s\n", permclone->setby);
	fscanf(ignore_file, "%lu\n", (long *) &(permclone->setwhen));
	count++;

	permclone->next = firstPermClone;
	permclone->prev = NULL;
	if (permclone->next)
	    permclone->next->prev = permclone;
	firstPermClone = permclone;
    }
    fclose(ignore_file);
}

void savepermignores()
{
    FILE *ignore_file;
    HostCloneList *permclone;

    ignore_file = fopen(DBDIR "/permignore.db", "w");
    if (ignore_file == NULL) {
	Debug((DEBUG_ERROR, "Failed to open ./permignore.db"));
	return;
    }

    for (permclone = firstPermClone; permclone;
	 permclone = permclone->next) {
	fprintf(ignore_file, "-\n");
	fprintf(ignore_file, "%s\n", permclone->host);
	fprintf(ignore_file, "%s\n", permclone->setby);
	fprintf(ignore_file, "%lu\n", (unsigned long) permclone->setwhen);
    }
    fclose(ignore_file);
}

/* This will add a server to the internal list of servers. */

void AddServer(char *name, char *version)
{

    aServer *Serverptr;

    Serverptr = make_server();
    strncpy(Serverptr->name, name, sizeof(Serverptr->name));
    strncpy(Serverptr->version, version, sizeof(Serverptr->version));

    if (ServerList == NULL)
	ServerList = Serverptr;
    else {
	ServerList->prev = Serverptr;
	Serverptr->next = ServerList;
	ServerList = Serverptr;
    }

    return;
}

/* This function will make an empty server structure */

aServer *make_server()
{
    aServer *Serverptr;

    Serverptr = (aServer *) smalloc(sizeof(aServer));

    Serverptr->name[0] = '\0';
    Serverptr->version[0] = '\0';
    Serverptr->svsnoop = 0;
    Serverptr->prev = Serverptr->next = NULL;

    return Serverptr;
}

void free_server(srvptr)
aServer *srvptr;
{
    free((char *) srvptr);
}

void DeleteServer(char *name)
{

    aServer *srvptr;
    aServer *srvptr2;

    if (ServerList == NULL)
	return;

    for (srvptr = ServerList; srvptr; srvptr = srvptr2) {
	srvptr2 = srvptr->next;
	if (!strcmp(srvptr->name, name)) {
	    if (srvptr->prev)
		srvptr->prev->next = srvptr->next;
	    else
		ServerList = srvptr->next;
	    if (srvptr->next)
		srvptr->next->prev = srvptr->prev;
	    free_server(srvptr);
	    return;
	}
    }

    return;
}

int CheckServer(char *name)
{

    aServer *srvptr;
    aServer *srvptr2;

    if (ServerList == NULL)
	return 0;

    for (srvptr = ServerList; srvptr; srvptr = srvptr2) {
	srvptr2 = srvptr->next;
	if (!strcasecmp(srvptr->name, name)) {
	    return 1;
	}
    }

    return 0;
}

void AddJupe(char *server, char *juper, time_t date, char *reason, int now)
{
	aJupe *jupeptr;

	jupeptr = make_jupe();

	strncpy(jupeptr->server, server, sizeof(jupeptr->server));
	strncpy(jupeptr->juper, juper, sizeof(jupeptr->juper));
	jupeptr->date = date;
	strncpy(jupeptr->reason, reason, sizeof(jupeptr->reason));

	if (JupeList == NULL)
		JupeList = jupeptr;
	else {
		JupeList->prev = jupeptr;
		jupeptr->next = JupeList;
		JupeList = jupeptr;
	}
	
	if (now == 1) {
		jupeptr->synced = 1;
		if (CheckServer(jupeptr->server))
		{
			sSend(":OperServ SQUIT %s :Enforcing Jupe set by %s on %s", jupeptr->server, jupeptr->juper, ctime(&(jupeptr->date)));
			DeleteServer(jupeptr->server);
		}

		sSend("SERVER %s 2 JupiteredServer :Jupitered by %s %s", jupeptr->server, jupeptr->juper, ctime(&jupeptr->date));
	} else
		jupeptr->synced = 0;
}

void sync_jupes(char *doot)
{
	aJupe *jupeptr;
	aJupe *jupeptr2;

	for (jupeptr = JupeList; jupeptr; jupeptr = jupeptr2) {
		jupeptr2 = jupeptr->next;
	
		if (jupeptr->synced != 1) {
			if (CheckServer(jupeptr->server))
			{
				sSend(":OperServ SQUIT %s :Enforcing Jupe set by %s on %s", jupeptr->server, jupeptr->juper, ctime(&(jupeptr->date)));
				DeleteServer(jupeptr->server);
			}

			sSend("SERVER %s 2 JupiteredServer :Jupitered by %s %s", jupeptr->server, jupeptr->juper, ctime(&jupeptr->date));
			jupeptr->synced = 1;
		}				
	}
	
	return;
}

aJupe *make_jupe()
{
	aJupe *jupeptr;		  

	jupeptr = (aJupe *) smalloc(sizeof(aJupe));
	
	jupeptr->server[0] = '\0';
	jupeptr->juper[0] = '\0';
	jupeptr->date = 0;
	jupeptr->reason[0] = '\0';
	
	jupeptr->prev = jupeptr->next = NULL;
			  
	return jupeptr;
}

void free_jupe(aJupe *jupeptr)
{
	free((char *) jupeptr);
}

void DeleteJupe(char *server)
{
	aJupe *jupeptr;
	aJupe *jupeptr2;

	if (JupeList == NULL)
		return;
	for (jupeptr = JupeList; jupeptr; jupeptr = jupeptr2) {
		jupeptr2 = jupeptr->next;
		if (strcmp(jupeptr->server, server) == 0) {
			if (jupeptr->prev)
				jupeptr->prev->next = jupeptr->next;
			else
				JupeList = jupeptr->next;
			if (jupeptr->next)
				jupeptr->next->prev = jupeptr->prev;
			free_jupe(jupeptr);
			return;
		}
	}
	
	return;
}

int CheckJupe(char *server)
{
	aJupe *jupeptr;
	aJupe *jupeptr2;

	if (JupeList == NULL)
		return 0;

	for (jupeptr = JupeList; jupeptr; jupeptr = jupeptr2) {
		jupeptr2 = jupeptr->next;
		if (!strcasecmp(jupeptr->server, server)) {
			return 1;
		}
	}
	return 0;
}


/* This will add an email address to the internal list of banned email addresses. */

void AddEmail(char *email, char *by, char *reason)
{

    anEmail *Emailptr;

    Emailptr = make_email();
    strncpy(Emailptr->email, email, sizeof(Emailptr->email));
    strncpy(Emailptr->by, by, sizeof(Emailptr->by));
    Emailptr->time = CTime;
    strncpy(Emailptr->reason, reason, sizeof(Emailptr->reason));

    if (EmailList == NULL)
	EmailList = Emailptr;
    else {
	EmailList->prev = Emailptr;
	Emailptr->next = EmailList;
	EmailList = Emailptr;
    }

    return;
}

/* This function will make an empty email structure */

anEmail *make_email()
{
    anEmail *Emailptr;

    Emailptr = (anEmail *) smalloc(sizeof(anEmail));

    Emailptr->email[0] = '\0';
    Emailptr->by[0] = '\0';
    Emailptr->time = 0;
    Emailptr->reason[0] = '\0';
    Emailptr->prev = Emailptr->next = NULL;

    return Emailptr;
}

void free_email(emailptr)
anEmail *emailptr;
{
    free((char *) emailptr);
}

void DeleteEmail(char *email)
{

    anEmail *emailptr;
    anEmail *emailptr2;

    if (EmailList == NULL)
	return;

    for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	emailptr2 = emailptr->next;
	if (!strcmp(emailptr->email, email)) {
	    if (emailptr->prev)
		emailptr->prev->next = emailptr->next;
	    else
		EmailList = emailptr->next;
	    if (emailptr->next)
		emailptr->next->prev = emailptr->prev;
	    free_email(emailptr);
	    return;
	}
    }

    return;
}

int CheckEmail(char *email)
{

    anEmail *emailptr;
    anEmail *emailptr2;

    if (EmailList == NULL)
	return 0;

    for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	emailptr2 = emailptr->next;
	if (!match(emailptr->email, email)) {
	    return 1;
	}
    }

    return 0;
}

int VerifyEmail(char *email)
{

    anEmail *emailptr;
    anEmail *emailptr2;

    if (EmailList == NULL)
	return 0;

    for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	emailptr2 = emailptr->next;
	if (!match(email, emailptr->email)) {
	    return 1;
	}
    }

    return 0;
}

void savedynamicekills()
{
    FILE *dekill_file;
    anEmail *emailptr;
    anEmail *emailptr2;

    if (EmailList == NULL)
	return;

    dekill_file = fopen(DBDIR "/dekill.db", "w");

    if (dekill_file == NULL) {
	Debug((DEBUG_ERROR, "Failed to open ./dekill.db"));
	return;
    }

    for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	emailptr2 = emailptr->next;
	fprintf(dekill_file, "-\n");
	fprintf(dekill_file, "%s\n", emailptr->email);
	fprintf(dekill_file, "%s\n", emailptr->by);
	fprintf(dekill_file, "%s\n", emailptr->reason);
	fprintf(dekill_file, "%lu\n", (unsigned long) emailptr->time);
    }

    fclose(dekill_file);
    return;
}

void loaddynamicekills()
{
    FILE *dekill_file;
    char buffer[256];
    int count = 0;

    anEmail *emailptr;

    EmailList = NULL;

    dekill_file = fopen(DBDIR "/dekill.db", "r");

    if (!dekill_file) {
	return;
    }

    while (sfgets(buffer, 256, dekill_file)) {

	if (strcmp(buffer, "-") != 0) {
	    sSend(":%s GLOBOPS :Sync error in dekill.db", OperServ);
	    fprintf(corelog, "Sync error in dekill.db");
	    fclose(dekill_file);
	    return;
	}

	emailptr = make_email();

	fscanf(dekill_file, "%s\n", emailptr->email);
	fscanf(dekill_file, "%s\n", emailptr->by);
	fscanf(dekill_file, "%s\n", emailptr->reason);
	fscanf(dekill_file, "%lu\n", (long *) &(emailptr->time));

	if (EmailList == NULL)
	    EmailList = emailptr;
	else {
	    EmailList->prev = emailptr;
	    emailptr->next = EmailList;
	    EmailList = emailptr;
	}
	count++;
    }

    fclose(dekill_file);

    return;
}

aServer *GetServer(char *server)
{

    aServer *srvptr;
    aServer *srvptr2;

    if (ServerList == NULL)
	return NULL;
    for (srvptr = ServerList; srvptr; srvptr = srvptr2) {
	srvptr2 = srvptr->next;
	if (!match(srvptr->name, server))
	    return srvptr;
    }

    return NULL;
}

#ifdef ANALYSE_FUNCTIONS
void AnalyseFunction(char *function)
{
    GetClock();
    return;
}
#endif
