#include "services.h"

typedef struct timer_struct Timer;

struct timer_struct {
    time_t run;
    unsigned int id;
    void (*func) (char *);
    char *desc;
    char *args;
    Timer *next, *prev;
};

unsigned int timers = 0;
unsigned int timersused = 0;

time_t lastcheck;

void delTimer(Timer *);

Timer *firstTimer = NULL, *lastTimer = NULL;

unsigned int timer(long seconds, void (*func) (char *), char *desc, char *args)
{
    Timer *newtimer;

    int goodvalue = 0;

    CTime = time(NULL);

    if (seconds <= 0) {
	func(args);
	return 0;
    }

    timersused++;
    timers++;

    if (!timers)		/* WHOOPS -- WE ROLLED OVER */
	timers++;

    while (!goodvalue) {
	goodvalue = 1;
	for (newtimer = firstTimer; newtimer; newtimer = newtimer->next) {
	    if (newtimer->id == timers) {
		goodvalue = 0;
		break;
	    }
	}
	if (!goodvalue) {
	    timers++;
	    if (!timers)	/* STILL ROLLING OVER */
		timers++;
	}
    }

    newtimer = smalloc(sizeof(Timer));
    newtimer->id = timers;
    newtimer->func = func;
    newtimer->desc = desc;
    newtimer->args = smalloc(strlen(args) + 1);
    strcpy(newtimer->args, args);
    newtimer->run = CTime + seconds;

    if (!firstTimer) {
	firstTimer = newtimer;
	newtimer->prev = NULL;
    } else {
	lastTimer->next = newtimer;
	newtimer->prev = lastTimer;
    }
    lastTimer = newtimer;
    newtimer->next = NULL;

    return timers;
}

void checktimers(void)
{

    Timer *tmp, *next;
    CTime = time(NULL);

    if (lastcheck == CTime)
	return;

    for (tmp = firstTimer; tmp; tmp = next) {
	next = tmp->next;
	if (tmp->run <= CTime) {
	    sprintf(timerbuff, "%s(%s)", tmp->desc, tmp->args);
	    timertime = CTime;
	    tmp->func(tmp->args);
	    delTimer(tmp);
	}
    }
}

void delTimer(Timer * deltimer)
{
    timersused--;

    if (deltimer->prev)
	deltimer->prev->next = deltimer->next;
    else
	firstTimer = deltimer->next;

    if (deltimer->next)
	deltimer->next->prev = deltimer->prev;
    else
	lastTimer = deltimer->prev;

    sfree(deltimer->args);
    sfree(deltimer);
}

int cancel_timer(unsigned int id)
{
    Timer *tmp;
    for (tmp = firstTimer; tmp; tmp = tmp->next) {
	if (tmp->id == id) {
	    delTimer(tmp);
	    return 1;
	}
    }
    return 0;
}
